-- ============================================================================
-- Project:	rbench - Logic Programming and Database Benchmarks
-- Filename:	db/create_db.sql
-- Purpose:	Database for Benchmark Runs: Create Tables and Views
-- Last Change:	11.10.2018
-- Language:	SQL (Tested with PostgreSQL)
-- Author:	Stefan Brass
-- EMail:	brass@informatik.uni-halle.de
-- WWW:		http://www.informatik.uni-halle.de/~brass/
-- Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
-- Copyright:	(c) 2016-2018 by Stefan Brass
-- License:	See file "LICENSE" for copying conditions.
-- Note:	There is no warranty at all - this code may contain bugs.
-- ============================================================================


-- This file contains the CREATE TABLE statements for the benchmark database,
-- and also many CREATE VIEW statements to store interesting queries
-- that can be used to analyze the data and to generate various output formats
-- (HTML, LaTeX, csv, tsv, Markdown, pieces of bash-scripts).

-- One should first delete old versions of the tables with the script
--	drop_db.sql

-- The table data are in various files which should be run after this script.
-- See the load_data script.
--	MACHINE:	machines.sql
--	INPUT_TYPE:	input_types.sql
--	BENCHMARK:	benchmarks.sql
--	SYS_TYPE:	sys_types.sql
--	SYSTEM, SYS_VER, SYS_INSTALL, BENCH_IMPL:
--			One file per system, e.g. xsb.sql
--	INPUT_FILE:	input_files.sql
--	INPUT_SET:	input_sets.sql
--	INPUT_GRAPH:	input_graphs.sql
--	BENCH_RUN:	results/*.sql
-- These SQL scripts should be run in the given sequence after this file
-- was executed.


-- ============================================================================
-- List of Machines on which Benchmarks are Executed:
-- ============================================================================

CREATE TABLE MACHINE(
	MACHINE     VARCHAR(20)  NOT NULL,
	SEQ_NO      NUMERIC(2)   NOT NULL,
	STATUS      CHAR(1)      NOT NULL,
        MODEL       VARCHAR(60)  NULL,
	YEAR        NUMERIC(4)   NULL,
        MONTH       NUMERIC(2)   NULL,
        PRICE       NUMERIC(7)   NULL,
	NUM_CPUS    NUMERIC(4)   NOT NULL,
	CPU         VARCHAR(60)  NOT NULL,
	CPU_CORES   NUMERIC(4)   NULL, -- Number of cores per CPU
        CPU_THREADS NUMERIC(4)   NULL, -- Number of hardware threads per CPU
	GHZ         NUMERIC(5,2) NULL,
	ARCH        VARCHAR(20)  NULL,
	RAM         NUMERIC(8,3) NOT NULL,
	NUM_DISKS   NUMERIC(4)   NULL,
        DISK        VARCHAR(40)  NULL,
	SSD         CHAR(1)      NULL,
	OS          VARCHAR(40)  NULL,
	OS_CORE     VARCHAR(40)  NULL,
	CONSTRAINT MACHINE_PK
		PRIMARY KEY(MACHINE),
	CONSTRAINT MACHINE_KEY_SEQ
		UNIQUE (SEQ_NO),
	CONSTRAINT MACHINE_SEQ_NO_POSITIVE
		CHECK(SEQ_NO > 0),
        CONSTRAINT MACHINE_YEAR_VALID
		CHECK(YEAR >= 2000),
        CONSTRAINT MACHINE_MONTH_VALID
                CHECK(MONTH >= 1 AND MONTH <= 12),
        CONSTRAINT MACHINE_PRICE_NON_NEGATIVE
                CHECK(PRICE >= 0),
	CONSTRAINT MACHINE_STATUS_S_OR_SPACE
		CHECK(STATUS IN ('S', ' ')),
		-- The machine with STATUS='S' is the standard machine.
		-- Only one row of the table should have STATUS='S'.
		-- Some views select only benchmark results on the standard
		-- machine.
		-- In future, there many be more STATUS values, e.g. for
		-- comparing the standard machine to an alternative machine.
	CONSTRAINT MACHINE_NUM_CPUS_POSITIVE
		CHECK(NUM_CPUS > 0),
	CONSTRAINT MACHINE_CPU_CORES_POSITIVE
		CHECK(CPU_CORES > 0),
	CONSTRAINT MACHINE_CPU_THREADS_POSITIVE
		CHECK(CPU_THREADS > 0),
	CONSTRAINT MACHINE_RAM_POSITIVE
		CHECK(RAM > 0),
	CONSTRAINT MACHINE_NUM_DISKS_POSITIVE
		CHECK(NUM_DISKS > 0),
	CONSTRAINT MACHINE_SSD_YN
		CHECK(SSD IN ('Y','N')));

-- Note that the CORE might not be current. It changes with every update.
-- It was current when we started the benchmarks in May 2018.
-- It gives some impression of the machine and its operating system.

-- The table data are in the file "machines.sql".


-- ============================================================================
-- List of Input Data (Problem Instance) Types:
-- ============================================================================


CREATE TABLE INPUT_TYPE(
	ITYPE       VARCHAR(8)  NOT NULL,
	SEQ_NO      NUMERIC(2)  NOT NULL,
	HEADLINE    VARCHAR(80) NOT NULL,
	DESCRIPTION VARCHAR(80) NOT NULL,
	CONSTRAINT INPUT_TYPE_SEQ_NO_NONNEGATIVE
		CHECK(SEQ_NO >= 0),
	CONSTRAINT INPUT_TYPE_PK
		PRIMARY KEY(ITYPE),
	CONSTRAINT INPUT_TYPE_SEQ_NO_UNIQUE
		UNIQUE(SEQ_NO));

-- The table data are in the file "input_types.sql".


-- ============================================================================
-- List of Benchmarks:
-- ============================================================================


CREATE TABLE BENCHMARK(
	BENCH      VARCHAR(20)  NOT NULL,
	SEQ_NO     NUMERIC(2)   NOT NULL,
	ITYPE      VARCHAR(8)   NULL,
	HEADLINE   VARCHAR(80)  NOT NULL,
	CONSTRAINT BENCHMARK_SEQ_NONNEGATIVE
		CHECK(SEQ_NO >= 0),
	CONSTRAINT BENCHMARK_PK
		PRIMARY KEY (BENCH),
	CONSTRAINT BENCHMARK_KEY_SEQ
		UNIQUE (SEQ_NO),
	CONSTRAINT BENCHMARK_REF_INPUT_TYPE
		FOREIGN KEY (ITYPE) REFERENCES INPUT_TYPE);

-- The table data are in the file "benchmarks.sql".


-- ============================================================================
-- System Types (for Subheadlines in the list of systems to be tested):
-- ============================================================================

CREATE TABLE SYS_TYPE(
	STYPE      VARCHAR(10)  NOT NULL,
	SEQ_NO     NUMERIC(2)   NOT NULL,
	HEADLINE   VARCHAR(80)  NOT NULL,
	CONSTRAINT SYS_TYPE_SEQ_NONNEGATIVE
		CHECK(SEQ_NO >= 0),
	CONSTRAINT SYS_TYPE_PK
		PRIMARY KEY (STYPE),
	CONSTRAINT SYS_TYPE_KEY_SEQ
		UNIQUE (SEQ_NO));

-- The table data are in the file "sys_types.sql".


-- ============================================================================
-- List of Systems to be Tested/Compared:
-- ============================================================================

-- SYSTEM is a reserved word in SQL:2003 and in DB2.

CREATE TABLE TEST_SYS(
	SYS         VARCHAR(20)  NOT NULL,
	SEQ_NO      NUMERIC(2)   NOT NULL,
	FULL_NAME   VARCHAR(40)  NOT NULL,
	STYPE       VARCHAR(10)  NOT NULL,
	URL         VARCHAR(80)  NULL,
	REMARK      VARCHAR(800) NULL,
	CONSTRAINT TEST_SYS_SEQ_NONNEGATIVE
		CHECK(SEQ_NO >= 0),
	CONSTRAINT TEST_SYS_PK
		PRIMARY KEY (SYS),
	CONSTRAINT TEST_SYS_KEY_SEQ
		UNIQUE (SEQ_NO),
	CONSTRAINT TEST_SYS_REF_SYS_TYPE
		FOREIGN KEY(STYPE) REFERENCES SYS_TYPE);

-- The table data are in the files for the single systems, e.g. "xsb.sql".


-- ============================================================================
-- Table for Versions of the Test Systems:
-- ============================================================================

-- This table contains information about program versions.
--
-- VER_ADDTEXT should be concatenated with VER to get the full name
-- of the version.
--
-- This table also permits to store additional information about a version
-- in the column REMARK (e.g., the major changes).
--
-- Note that this table describes a version of the source code.
-- SYS_INSTALL (below) describes a version of the program binary,
-- which also depends on the machine, the compiler, compiler settings,
-- and configuration settings.

CREATE TABLE SYS_VER (
	SYS         VARCHAR(20)  NOT NULL,
	VER         VARCHAR(20)  NOT NULL,
	VER_ADDTEXT VARCHAR(80)  NOT NULL,
	VER_DATE    DATE         NULL,
	REMARK      VARCHAR(800) NULL,
	CONSTRAINT SYS_VER_PK
		PRIMARY KEY (SYS, VER),
	CONSTRAINT SYS_VER_REF_TEST_SYS
		FOREIGN KEY (SYS) REFERENCES TEST_SYS);

-- The table data are in the files for the single systems, e.g. "xsb.sql".


-- ============================================================================
-- Table for Installed Program Versions on Machines:
-- ============================================================================

-- This table lists program versions that are or were installed
-- on a test machine.
-- They are identified here by a number INST_NO that is unique together with
-- the program name and the machine.
-- These three elements (program, installation ID, and machine)
-- are also encoded in the file names with the test results.
--
-- Whereas SYS_VER describes a version of the source code,
-- SYS_INSTALL describes a binary program.
-- For instance, if the same source code is compiled with g++ and with clang,
-- these are two different entries in SYS_INSTALL, both referring to the
-- same entry in SYS_VER.
-- Also different configuration settings may give different binaries,
-- and therefore different entries in SYS_INSTALL.
--
-- I try to make sure that the same version of a program
-- gets the same INST_NO number even on different machines,
-- but the database schema does not require this
-- (and it is impossible when there are really different installations
--  of the same program version).
--
-- Only program installations that are listed here can appear in the test runs.
--
-- This table also define a current version (more precisely, installation)
-- of a program on a machine (by STATUS='C').
-- Some views show only results for the current version.
-- It is also possible to define a previous version (STATUS='P') as reference,
-- if the result of a program update should be evaluated.
-- Finally, one can mark an alternative version (STATUS='A'),
-- e.g. compiled with a different compiler or compiler options,
-- which can also be used for comparison.

CREATE TABLE SYS_INSTALL (
	SYS         VARCHAR(20)  NOT NULL,
	VER         VARCHAR(20)  NOT NULL,
	MACHINE     VARCHAR(20)  NOT NULL,
	INST_NO     NUMERIC(4)   NOT NULL,
	COMPILER    VARCHAR(80)  NOT NULL,
	STATUS      CHAR(1)      NOT NULL,
	REMARK      VARCHAR(800) NULL,
	CONSTRAINT SYS_INSTALL_PK
		PRIMARY KEY (SYS, INST_NO, MACHINE),
	CONSTRAINT SYS_INSTALL_REF_SYS_VER
		FOREIGN KEY (SYS, VER) REFERENCES SYS_VER,
	CONSTRAINT SYS_INSTALL_REF_MACHINE
		FOREIGN KEY (MACHINE) REFERENCES MACHINE,
	CONSTRAINT SYS_INSTALL_STATUS_VALID
		CHECK(STATUS IN ('C','P','A',' ')));
	-- 'C': Current version
	-- 'P': Previous version
	-- 'A': Alternate version
	-- ' ': Any other version
	-- Views may select only benchmark results for the current version,
	-- or help to compare the current with the previous version.

-- The table data are in the files for the single systems, e.g. "xsb.sql".


-- ============================================================================
-- Table for Implementation Variants of Benchmarks:
-- ============================================================================

CREATE TABLE BENCH_IMPL(
	SYS         VARCHAR(20)  NOT NULL,
	BENCH       VARCHAR(20)  NOT NULL,
	IMPL        VARCHAR(20)  NOT NULL,
	STATUS      CHAR(1)      NOT NULL,
	DESCRIPTION VARCHAR(80)  NULL,
	CONSTRAINT BENCH_IMPL_PK
		PRIMARY KEY (SYS, BENCH, IMPL),
	CONSTRAINT BENCH_IMPL_REF_BENCHMARK
		FOREIGN KEY (BENCH) REFERENCES BENCHMARK,
	CONSTRAINT BENCH_IMPL_REF_TEST_SYS
		FOREIGN KEY (SYS) REFERENCES TEST_SYS,
	CONSTRAINT BENCH_IMPL_STATUS_VALID
		CHECK(STATUS IN ('B','A',' ')));

-- STATUS='B' marks the implementation variant (option settings)
-- that is considered best and is used in comparisons between systems.
-- One can use the views BEST_IMPL_REAL and BEST_IMPL_CPU to check
-- whether these implementation variants are really best.
-- Sometimes the best implementation depends on the specific data file,
-- but it would somehow be unfair to adapt the options to the data file.

-- STATUS='A' marks an interesting alternative variants
-- that appear in some result tables.
-- There should be only one "best" version,
-- but there can be several alterative versions.

-- STATUS=' ' means that the variant was tested
-- in order to prove that the best version is really good,
-- but it will not be listed in summary tables.

-- The table data are in the files for the single systems, e.g. "xsb.sql".


-- ============================================================================
-- Table for Data Files (Input Files to Benchmarks):
-- ============================================================================

CREATE TABLE INPUT_FILE(
	FILE_ID     VARCHAR(10)  NOT NULL,
	SEQ_NO      NUMERIC(3)   NOT NULL,
	ITYPE       VARCHAR(8)   NOT NULL,
	LINES       NUMERIC(12)  NOT NULL, -- should be the number of facts
	BYTES       NUMERIC(12)  NOT NULL, -- refers to Datalog facts version
	FILENAME    VARCHAR(40)  NOT NULL, -- without extension
--	DESCRIPTION VARCHAR(80)  NULL,
	CONSTRAINT INPUT_FILE_PK
		PRIMARY KEY (FILE_ID),
	CONSTRAINT INPUT_FILE_KEY2
		UNIQUE (SEQ_NO),
	CONSTRAINT INPUT_FILE_SEQ_NO_NOT_NEGATIVE
		CHECK(SEQ_NO >= 0),
	CONSTRAINT INPUT_FILE_REF_INPUT_TYPE
		FOREIGN KEY (ITYPE) REFERENCES INPUT_TYPE,
	CONSTRAINT INPUT_FILE_LINES_NOT_NEGATIVE
		CHECK(LINES >= 0),
	CONSTRAINT INPUT_FILE_BYTES_NOT_NEGATIVE
		CHECK(BYTES >= 0));

-- wc was used to compute lines and bytes

-- The table data are in file "input_files.sql"


-- ============================================================================
-- Table of Collections of Files Used for Publications:
-- ============================================================================

-- This can be used to put arbitrary tags at input files.

CREATE TABLE INPUT_SET(
	SET_ID      VARCHAR(10)  NOT NULL,
	FILE_ID     VARCHAR(10)  NOT NULL,
	CONSTRAINT INPUT_SET_PK
		PRIMARY KEY (SET_ID, FILE_ID),
	CONSTRAINT INPUT_SET_REFERENCES_INPUT_FILE
		FOREIGN KEY (FILE_ID) REFERENCES INPUT_FILE);

-- The table data are in file "input_sets.sql"


-- ============================================================================
-- Graph Types / Parameterized Graphs (For Input Data Files that are Graphs):
-- ============================================================================

CREATE TABLE GRAPH_TYPE(
	GTYPE      CHAR(1)      NOT NULL,
	SEQ_NO     NUMERIC(2)   NOT NULL,
	NAME       VARCHAR(80)  NOT NULL,
	CONSTRAINT GRAPH_TYPE_SEQ_NONNEGATIVE
		CHECK(SEQ_NO >= 0),
	CONSTRAINT GRAPH_TYPE_PK
		PRIMARY KEY (GTYPE),
	CONSTRAINT GRAPH_TYPE_KEY_SEQ
		UNIQUE (SEQ_NO));

-- The table data are in file "graph_types.sql".


-- ============================================================================
-- More Data on Data Files That are Directed Graphs (Subclass of Input Files):
-- ============================================================================

CREATE TABLE INPUT_GRAPH (
	FILE_ID        VARCHAR(10) NOT NULL,
	GTYPE          CHAR(1)     NOT NULL,
	PARAM_1        NUMERIC(12) NULL,
	PARAM_2        NUMERIC(12) NULL,
	NUM_NODES      NUMERIC(12) NOT NULL,
	NUM_EDGES      NUMERIC(12) NOT NULL,
	DUP_EDGES      NUMERIC(12) NOT NULL,
	LOOPS          CHAR(1)     NOT NULL,
	MIN_IN_DEGREE  NUMERIC(12) NOT NULL,
	MAX_IN_DEGREE  NUMERIC(12) NOT NULL,
	MIN_OUT_DEGREE NUMERIC(12) NOT NULL,
	MAX_OUT_DEGREE NUMERIC(12) NOT NULL,
	CYCLES         CHAR(1)     NOT NULL,
	TC_SIZE        NUMERIC(12) NOT NULL,
	NUM_ITER       NUMERIC(12) NOT NULL,
	COST           NUMERIC(12) NOT NULL,
	CONSTRAINT INPUT_GRAPH_PK
		PRIMARY KEY(FILE_ID),
	CONSTRAINT INPUT_GRAPH_REF_INPUT_FILE
		FOREIGN KEY(FILE_ID) REFERENCES INPUT_FILE,
	CONSTRAINT INPUT_GRAPH_REF_GRAPH_TYPE
		FOREIGN KEY (GTYPE) REFERENCES GRAPH_TYPE,
	CONSTRAINT INPUT_GRAPH_PARAM_1_NONNEGATIVE
		CHECK(PARAM_1 >= 0),
	CONSTRAINT INPUT_GRAPH_PARAM_2_NONNEGATIVE
		CHECK(PARAM_2 >= 0),
	CONSTRAINT INPUT_GRAPH_NUM_NODES_NONNEGATIVE
		CHECK(NUM_NODES >= 0),
	CONSTRAINT INPUT_GRAPH_NUM_EDGES_NONNEGATIVE
		CHECK(NUM_EDGES >= 0),
	CONSTRAINT INPUT_GRAPH_DUP_EDGES_NONNEGATIVE
		CHECK(DUP_EDGES >= 0),
	CONSTRAINT INPUT_GRAPH_LOOPS_YES_NO
		CHECK(LOOPS  in ('Y','N')),
	CONSTRAINT INPUT_GRAPH_MIN_IN_DEGREE_NONNEGATIVE
		CHECK(MIN_IN_DEGREE >= 0),
	CONSTRAINT INPUT_GRAPH_MAX_IN_DEGREE_NONNEGATIVE
		CHECK(MAX_IN_DEGREE >= 0),
	CONSTRAINT INPUT_GRAPH_MIN_OUT_DEGREE_NONNEGATIVE
		CHECK(MIN_OUT_DEGREE >= 0),
	CONSTRAINT INPUT_GRAPH_MAX_OUT_DEGREE_NONNEGATIVE
		CHECK(MAX_OUT_DEGREE >= 0),
	CONSTRAINT INPUT_GRAPH_CYCLES_YES_NO
		CHECK(CYCLES IN ('Y','N')),
	CONSTRAINT INPUT_GRAPH_TC_SIZE_NONNEGATIVE
		CHECK(TC_SIZE >= 0),
	CONSTRAINT INPUT_GRAPH_NUM_ITER_NONNEGATIVE
		CHECK(NUM_ITER >= 0),
	CONSTRAINT INPUT_GRAPH_COST_NONNEGATIVE
		CHECK(COST >= 0));


-- PARAM_1 and PARAM_2 are the parameters for the Data Generation Program

-- DUP_EDGES is the number of duplicate edges,
-- i.e. the number of node pairs (x,y) such that there are at least two edges
-- between these nodes.
-- Normally, it should be 0.
-- However, the OpenRuleBench graph generation program generates such
-- duplicate edges for acyclic graphs.
-- Therefore, our graph analysis checks the input files for this case.

-- LOOPS='Y' means that there are edges of the form (x,x).

-- COST is the number of applicable rule instances in the standard
-- tail-recursive Prolog program for computing the transitive closure:
--	tc(X, Y) :- par(X, Y).
--	tc(X, Z) :- par(X, Y), tc(Y, Z).
-- COST is the number of rule instances that semi-naive evaluation would
-- consider.
-- It can be computed as the sum of the number of edges
-- (i.e. the size of the par relation) and
-- the size of the join of par and tc.

-- The data is in file input_graphs.sql


-- ============================================================================
-- Table for Results of Benchmark Runs (Test Runs):
-- ============================================================================


CREATE TABLE BENCH_RUN(
	RUN_DATE    DATE          NOT NULL,
	RUN_NO      NUMERIC(4)    NOT NULL,
	SYS         VARCHAR(20)   NOT NULL,
	INST_NO     NUMERIC(4)    NOT NULL,
	MACHINE     VARCHAR(20)   NOT NULL,
	BENCH       VARCHAR(20)   NOT NULL,
	IMPL        VARCHAR(20)   NOT NULL,
	FILE_ID     VARCHAR(10)   NOT NULL,
	WITH_OUTPUT CHAR(1)       NOT NULL,
	LOAD_TIME   NUMERIC(10,3) NULL,
	EXEC_TIME   NUMERIC(10,3) NULL,
	TOTAL_TIME  NUMERIC(10,3) NULL,
	USER_TIME   NUMERIC(10,3) NULL,
	SYS_TIME    NUMERIC(10,3) NULL,
	REAL_TIME   NUMERIC(10,3) NOT NULL,
	MEM         NUMERIC(10,3) NULL,
	STATUS      VARCHAR(80)   NOT NULL,
	REMARK      VARCHAR(800)  NULL,
        CONSTRAINT  BENCH_RUN_RUN_NO_POSITIVE
		CHECK(RUN_NO > 0),
        CONSTRAINT  BENCH_RUN_WITH_OUTPUT_BOOL
		CHECK(WITH_OUTPUT IN ('Y','N')),
        CONSTRAINT  BENCH_RUN_LOAD_TIME_NONNEGATIVE
		CHECK(LOAD_TIME >= 0),
        CONSTRAINT  BENCH_RUN_EXEC_TIME_NONNEGATIVE
		CHECK(EXEC_TIME >= 0),
        CONSTRAINT  BENCH_RUN_TOTAL_TIME_NONNEGATIVE
		CHECK(TOTAL_TIME >= 0),
        CONSTRAINT  BENCH_RUN_USER_TIME_NONNEGATIVE
		CHECK(USER_TIME >= 0),
        CONSTRAINT  BENCH_RUN_SYS_TIME_NONNEGATIVE
		CHECK(SYS_TIME >= 0),
        CONSTRAINT  BENCH_RUN_REAL_TIME_NONNEGATIVE
		CHECK(REAL_TIME >= 0),
        CONSTRAINT  BENCH_RUN_MEM_NONNEGATIVE
		CHECK(MEM >= 0),
        CONSTRAINT  BENCH_RUN_STATUS_VALID
		CHECK(STATUS IN ('OK', 'FAILED', 'OUTL',
				'Warn', 'Warn1', 'Warn2', 'Warn3', 'Warn4')),
        CONSTRAINT  BENCH_RUN_PK
		PRIMARY KEY(RUN_DATE,RUN_NO,MACHINE,SYS,INST_NO),
        CONSTRAINT  BENCH_RUN_REF_BENCH_IMPL
		FOREIGN KEY(SYS, BENCH, IMPL) REFERENCES BENCH_IMPL,
	CONSTRAINT  BENCH_RUN_REF_SYS_INSTALL
		FOREIGN KEY(SYS, INST_NO, MACHINE) REFERENCES SYS_INSTALL,
	CONSTRAINT  BENCH_RUN_REF_INPUT_FILE
		FOREIGN KEY(FILE_ID) REFERENCES INPUT_FILE);

-- All times are stored in seconds.
-- MEM is the maximum resident set size.


-- ============================================================================
-- Experimental Table for Runtime Estimation Parameters:
-- ============================================================================


CREATE TABLE ESTIMATE_PAR(
	SYS         VARCHAR(20)  NOT NULL,
	BENCH       VARCHAR(20)  NOT NULL,
	STARTUP     NUMERIC(10)  NOT NULL,
	LOAD_IDX    NUMERIC(10)  NOT NULL,
	RULE_APP    NUMERIC(10)  NOT NULL,
	RESULT      NUMERIC(10)  NOT NULL,
	HITS        NUMERIC(10)  NOT NULL,
        CONSTRAINT  ESTIMATE_PAR_STARTUP_NONNEGATIVE
		CHECK(STARTUP >= 0),
        CONSTRAINT  ESTIMATE_PAR_LOAD_IDX_NONNEGATIVE
		CHECK(LOAD_IDX >= 0),
        CONSTRAINT  ESTIMATE_PAR_RULE_APP_NONNEGATIVE
		CHECK(RULE_APP >= 0),
        CONSTRAINT  ESTIMATE_PAR_RESULT_NONNEGATIVE
		CHECK(RESULT >= 0),
        CONSTRAINT  ESTIMATE_PAR_HITS_NONNEGATIVE
		CHECK(HITS >= 0),
        CONSTRAINT  ESTIMATE_PAR_PK
		PRIMARY KEY(SYS,BENCH,STARTUP,LOAD_IDX,RULE_APP,RESULT),
        CONSTRAINT  ESTIMATE_PAR_REF_SYS
		FOREIGN KEY(SYS) REFERENCES TEST_SYS,
        CONSTRAINT  ESTIMATE_PAR_REF_BENCH
		FOREIGN KEY(BENCH) REFERENCES BENCHMARK);

-- The data for this table are in file "estimate_par.sql".


-- ============================================================================
-- Dummy Table with Exactly One Row Used in Some Views:
-- ============================================================================


CREATE TABLE DUMMY_TAB(
	ATTR NUMERIC(1)
);

INSERT INTO DUMMY_TAB VALUES(0);


-- ############################################################################
-- Views for Statistical Evaluation of Run Data:
-- ############################################################################


-- ============================================================================
-- Full Statistical Evaluation of Run Data:
-- ============================================================================

CREATE VIEW RUN_STATISTICS AS
SELECT BENCH, IMPL, FILE_ID, MACHINE, SYS, INST_NO, WITH_OUTPUT,
	-- Load Time:
	ROUND(AVG(LOAD_TIME),3) AS AVG_LOAD_TIME,
	MIN(LOAD_TIME) AS MIN_LOAD_TIME,
	MAX(LOAD_TIME) AS MAX_LOAD_TIME,
	MAX(LOAD_TIME)-MIN(LOAD_TIME) AS DIFF_LOAD_TIME,
	STDDEV(LOAD_TIME) AS SDEV_LOAD_TIME,
	-- Exec Time:
	ROUND(AVG(EXEC_TIME),3) AS AVG_EXEC_TIME,
	MIN(EXEC_TIME) AS MIN_EXEC_TIME,
	MAX(EXEC_TIME) AS MAX_EXEC_TIME,
	MAX(EXEC_TIME)-MIN(EXEC_TIME) AS DIFF_EXEC_TIME,
	STDDEV(EXEC_TIME) AS SDEV_EXEC_TIME,
	-- Total Time:
	ROUND(AVG(TOTAL_TIME),3) AS AVG_TOTAL_TIME,
	MIN(TOTAL_TIME) AS MIN_TOTAL_TIME,
	MAX(TOTAL_TIME) AS MAX_TOTAL_TIME,
	MAX(TOTAL_TIME)-MIN(TOTAL_TIME) AS DIFF_TOTAL_TIME,
	STDDEV(TOTAL_TIME) AS SDEV_TOTAL_TIME,
	-- User Mode Time:
	ROUND(AVG(USER_TIME),3) AS AVG_USER_TIME,
	MIN(USER_TIME) AS MIN_USER_TIME,
	MAX(USER_TIME) AS MAX_USER_TIME,
	MAX(USER_TIME)-MIN(USER_TIME) AS DIFF_USER_TIME,
	STDDEV(USER_TIME) AS SDEV_USER_TIME,
	-- System Mode Time:
	ROUND(AVG(SYS_TIME),3) AS AVG_SYS_TIME,
	MIN(SYS_TIME) AS MIN_SYS_TIME,
	MAX(SYS_TIME) AS MAX_SYS_TIME,
	MAX(SYS_TIME)-MIN(SYS_TIME) AS DIFF_SYS_TIME,
	STDDEV(SYS_TIME) AS SDEV_SYS_TIME,
	-- CPU Time:
	ROUND(AVG(SYS_TIME+USER_TIME),3) AS AVG_CPU_TIME,
	MIN(SYS_TIME+USER_TIME) AS MIN_CPU_TIME,
	MAX(SYS_TIME+USER_TIME) AS MAX_CPU_TIME,
	MAX(SYS_TIME+USER_TIME)-MIN(SYS_TIME+USER_TIME) AS DIFF_CPU_TIME,
	STDDEV(SYS_TIME+USER_TIME) AS SDEV_CPU_TIME,
	-- Real Time:
	ROUND(AVG(REAL_TIME),3) AS AVG_REAL_TIME,
	MIN(REAL_TIME) AS MIN_REAL_TIME,
	MAX(REAL_TIME) AS MAX_REAL_TIME,
	MAX(REAL_TIME)-MIN(REAL_TIME) AS DIFF_REAL_TIME,
	STDDEV(REAL_TIME) AS SDEV_REAL_TIME,
	-- Memory:
	ROUND(AVG(MEM),3) AS AVG_MEM,
	MIN(MEM) AS MIN_MEM,
	MAX(MEM) AS MAX_MEM,
	MAX(MEM)-MIN(MEM) AS DIFF_MEM,
	STDDEV(MEM) AS SDEV_MEM,
	-- Number of Runs:
	COUNT(*) AS NUM_BENCH_RUN
FROM BENCH_RUN
WHERE STATUS <> 'FAILED' AND STATUS <> 'OUTL'
GROUP BY MACHINE, SYS, INST_NO, BENCH, IMPL, FILE_ID, WITH_OUTPUT
ORDER BY MACHINE, SYS, INST_NO, BENCH, IMPL, FILE_ID, WITH_OUTPUT;

-- ============================================================================
-- Statistical Evaluation of Run Data, without STDDEV - This works for SQLite:
-- ============================================================================

CREATE VIEW RUN_STAT AS
SELECT BENCH, IMPL, FILE_ID, MACHINE, SYS, INST_NO, WITH_OUTPUT,
	-- Load Time:
	ROUND(AVG(LOAD_TIME),3) AS AVG_LOAD_TIME,
	MIN(LOAD_TIME) AS MIN_LOAD_TIME,
	MAX(LOAD_TIME) AS MAX_LOAD_TIME,
	MAX(LOAD_TIME)-MIN(LOAD_TIME) AS DIFF_LOAD_TIME,
	-- Exec Time:
	ROUND(AVG(EXEC_TIME),3) AS AVG_EXEC_TIME,
	MIN(EXEC_TIME) AS MIN_EXEC_TIME,
	MAX(EXEC_TIME) AS MAX_EXEC_TIME,
	MAX(EXEC_TIME)-MIN(EXEC_TIME) AS DIFF_EXEC_TIME,
	-- Total Time:
	ROUND(AVG(TOTAL_TIME),3) AS AVG_TOTAL_TIME,
	MIN(TOTAL_TIME) AS MIN_TOTAL_TIME,
	MAX(TOTAL_TIME) AS MAX_TOTAL_TIME,
	MAX(TOTAL_TIME)-MIN(TOTAL_TIME) AS DIFF_TOTAL_TIME,
	-- User Mode Time:
	ROUND(AVG(USER_TIME),3) AS AVG_USER_TIME,
	MIN(USER_TIME) AS MIN_USER_TIME,
	MAX(USER_TIME) AS MAX_USER_TIME,
	MAX(USER_TIME)-MIN(USER_TIME) AS DIFF_USER_TIME,
	-- System Mode Time:
	ROUND(AVG(SYS_TIME),3) AS AVG_SYS_TIME,
	MIN(SYS_TIME) AS MIN_SYS_TIME,
	MAX(SYS_TIME) AS MAX_SYS_TIME,
	MAX(SYS_TIME)-MIN(SYS_TIME) AS DIFF_SYS_TIME,
	-- CPU Time:
	ROUND(AVG(SYS_TIME+USER_TIME),3) AS AVG_CPU_TIME,
	MIN(SYS_TIME+USER_TIME) AS MIN_CPU_TIME,
	MAX(SYS_TIME+USER_TIME) AS MAX_CPU_TIME,
	MAX(SYS_TIME+USER_TIME)-MIN(SYS_TIME+USER_TIME) AS DIFF_CPU_TIME,
	-- Real Time:
	ROUND(AVG(REAL_TIME),3) AS AVG_REAL_TIME,
	MIN(REAL_TIME) AS MIN_REAL_TIME,
	MAX(REAL_TIME) AS MAX_REAL_TIME,
	MAX(REAL_TIME)-MIN(REAL_TIME) AS DIFF_REAL_TIME,
	-- Memory:
	ROUND(AVG(MEM),3) AS AVG_MEM,
	MIN(MEM) AS MIN_MEM,
	MAX(MEM) AS MAX_MEM,
	MAX(MEM)-MIN(MEM) AS DIFF_MEM,
	-- Number of Runs:
	COUNT(*) AS NUM_BENCH_RUN
FROM BENCH_RUN
WHERE STATUS <> 'FAILED' AND STATUS <> 'OUTL'
GROUP BY MACHINE, SYS, INST_NO, BENCH, IMPL, FILE_ID, WITH_OUTPUT
ORDER BY MACHINE, SYS, INST_NO, BENCH, IMPL, FILE_ID, WITH_OUTPUT;

-- ============================================================================
-- Benchmark Results for Current Version on the Standard Machine without Output:
-- ============================================================================

CREATE VIEW RESULT AS
SELECT R.BENCH AS BENCH,
	R.IMPL AS IMPL,
	R.FILE_ID AS FILE_ID,
	R.SYS AS SYS,
	R.AVG_LOAD_TIME AS LOAD_T,
	R.AVG_EXEC_TIME AS EXEC_T,
	R.AVG_TOTAL_TIME AS TOTAL_T,
	R.AVG_CPU_TIME AS CPU_T,
	R.AVG_REAL_TIME AS REAL_T,
	R.AVG_MEM AS MEM_MB
FROM   RUN_STAT R, SYS_INSTALL V, MACHINE M, INPUT_FILE F
WHERE  R.MACHINE = M.MACHINE AND V.MACHINE = R.MACHINE AND M.STATUS = 'S'
AND    R.SYS = V.SYS AND R.INST_NO = V.INST_NO AND V.STATUS = 'C'
AND    R.WITH_OUTPUT = 'N'
AND    R.FILE_ID = F.FILE_ID
ORDER BY R.BENCH, R.IMPL, F.SEQ_NO, R.SYS;

-- ============================================================================
-- Benchmark Results (Current Ver., Std. Machine, No Output, Best Settings):
-- ============================================================================

CREATE VIEW BEST_RESULT AS
SELECT R.BENCH AS BENCH,
	R.IMPL AS IMPL,
	R.FILE_ID AS FILE_ID,
	R.SYS AS SYS,
	R.AVG_LOAD_TIME AS LOAD_T,
	R.AVG_EXEC_TIME AS EXEC_T,
	R.AVG_TOTAL_TIME AS TOTAL_T,
	R.AVG_CPU_TIME AS CPU_T,
	R.AVG_REAL_TIME AS REAL_T,
	R.AVG_MEM AS MEM_MB
FROM   RUN_STAT R, SYS_INSTALL V, MACHINE M, BENCH_IMPL I, INPUT_FILE F
WHERE  R.MACHINE = M.MACHINE AND V.MACHINE = R.MACHINE AND M.STATUS = 'S'
AND    R.SYS = V.SYS AND R.INST_NO = V.INST_NO AND V.STATUS = 'C'
AND    R.WITH_OUTPUT = 'N'
AND    I.SYS = R.SYS AND I.BENCH = R.BENCH AND I.IMPL = R.IMPL
AND    I.STATUS = 'B'
AND    R.FILE_ID = F.FILE_ID
ORDER BY R.BENCH, R.IMPL, F.SEQ_NO, R.SYS;


-- ############################################################################
-- Outlier Detection, Checking of Results for Plausibility:
-- ############################################################################


-- ============================================================================
-- Detect Outliers with regard to Real Time:
-- ============================================================================

CREATE VIEW OUTLIER_REAL AS
SELECT R.REAL_TIME, S.AVG_REAL_TIME, R.SYS, R.INST_NO, R.STATUS,
       R.BENCH, R.IMPL, R.FILE_ID, R.MACHINE, R.WITH_OUTPUT,
       R.RUN_DATE, R.RUN_NO
FROM   BENCH_RUN R, RUN_STAT S
WHERE  R.STATUS <> 'FAILED'
AND    R.BENCH       = S.BENCH
AND    R.IMPL        = S.IMPL
AND    R.FILE_ID     = S.FILE_ID
AND    R.MACHINE     = S.MACHINE
AND    R.SYS         = S.SYS
AND    R.INST_NO     = S.INST_NO
AND    R.WITH_OUTPUT = S.WITH_OUTPUT
AND    R.REAL_TIME   > 2.0 * S.AVG_REAL_TIME
AND    R.REAL_TIME   > S.AVG_REAL_TIME + 1
ORDER BY MACHINE, SYS, INST_NO, BENCH, IMPL, FILE_ID, WITH_OUTPUT;

-- ============================================================================
-- Detect Outliers with regard to CPU Time:
-- ============================================================================

CREATE VIEW OUTLIER_CPU AS
SELECT R.SYS_TIME+R.USER_TIME AS CPU_TIME, S.AVG_CPU_TIME,
       R.SYS, R.INST_NO, R.STATUS,
       R.BENCH, R.IMPL, R.FILE_ID, R.MACHINE, R.WITH_OUTPUT,
       R.RUN_DATE, R.RUN_NO
FROM   BENCH_RUN R, RUN_STAT S
WHERE  R.STATUS <> 'FAILED'
AND    R.BENCH       = S.BENCH
AND    R.IMPL        = S.IMPL
AND    R.FILE_ID     = S.FILE_ID
AND    R.MACHINE     = S.MACHINE
AND    R.SYS         = S.SYS
AND    R.INST_NO     = S.INST_NO
AND    R.WITH_OUTPUT = S.WITH_OUTPUT
AND    R.SYS_TIME+R.USER_TIME > 2.0 * S.AVG_CPU_TIME
AND    R.SYS_TIME+R.USER_TIME > S.AVG_CPU_TIME + 1
ORDER BY MACHINE, SYS, INST_NO, BENCH, IMPL, FILE_ID, WITH_OUTPUT;

-- ============================================================================
-- Detect Outliers with regard to Memory:
-- ============================================================================

CREATE VIEW OUTLIER_MEM AS
SELECT R.MEM, S.AVG_MEM, R.SYS, R.INST_NO, R.STATUS,
       R.BENCH, R.IMPL, R.FILE_ID, R.MACHINE, R.WITH_OUTPUT,
       R.RUN_DATE, R.RUN_NO
FROM   BENCH_RUN R, RUN_STAT S
WHERE  R.STATUS <> 'FAILED'
AND    R.BENCH       = S.BENCH
AND    R.IMPL        = S.IMPL
AND    R.FILE_ID     = S.FILE_ID
AND    R.MACHINE     = S.MACHINE
AND    R.SYS         = S.SYS
AND    R.INST_NO     = S.INST_NO
AND    R.WITH_OUTPUT = S.WITH_OUTPUT
AND    R.MEM         > 2.0 * S.AVG_MEM
AND    R.MEM         > S.AVG_MEM + 5
ORDER BY MACHINE, SYS, INST_NO, BENCH, IMPL, FILE_ID, WITH_OUTPUT;

-- ============================================================================
-- Check where Realtime is much more than CPU time (at least double):
-- ============================================================================

CREATE VIEW REAL_TOO_LARGE AS
SELECT R.REAL_TIME, S.AVG_REAL_TIME,
       R.SYS_TIME+R.USER_TIME AS CPU_TIME, S.AVG_CPU_TIME,
       R.SYS, R.INST_NO, R.STATUS,
       R.BENCH, R.IMPL, R.FILE_ID, R.MACHINE, R.WITH_OUTPUT,
       R.RUN_DATE, R.RUN_NO
FROM   BENCH_RUN R, RUN_STAT S
WHERE  R.STATUS <> 'FAILED'
AND    R.BENCH       = S.BENCH
AND    R.IMPL        = S.IMPL
AND    R.FILE_ID     = S.FILE_ID
AND    R.MACHINE     = S.MACHINE
AND    R.SYS         = S.SYS
AND    R.INST_NO     = S.INST_NO
AND    R.WITH_OUTPUT = S.WITH_OUTPUT
AND    R.REAL_TIME   > 2.0 * (R.SYS_TIME+R.USER_TIME)
AND    R.REAL_TIME   > (R.SYS_TIME+R.USER_TIME) + 1
ORDER BY MACHINE, SYS, INST_NO, BENCH, IMPL, FILE_ID, WITH_OUTPUT;

-- ============================================================================
-- Compare Real Time with CPU Time:
-- ============================================================================

CREATE VIEW PCT_CPU AS
SELECT SYS, RUN_DATE, MACHINE, BENCH,
	ROUND(AVG((SYS_TIME+USER_TIME) * 100 / REAL_TIME)) AS AVG_CPU,
	ROUND(MIN((SYS_TIME+USER_TIME) * 100 / REAL_TIME)) AS MIN_CPU,
	ROUND(MAX((SYS_TIME+USER_TIME) * 100 / REAL_TIME)) AS MAX_CPU
FROM   BENCH_RUN
WHERE  SYS_TIME IS NOT NULL AND USER_TIME IS NOT NULL
AND    REAL_TIME > 0
AND    STATUS <> 'FAILED' AND STATUS <> 'OUTL'
AND    WITH_OUTPUT = 'N'
GROUP  BY SYS, RUN_DATE, MACHINE, BENCH
ORDER  BY SYS, RUN_DATE, MACHINE, BENCH;


-- ============================================================================
-- Compare Real Time with CPU Time:
-- ============================================================================

CREATE VIEW PCT_CPU_INPUT AS
SELECT SYS, MACHINE, BENCH, FILE_ID,
	ROUND(AVG((SYS_TIME+USER_TIME) * 100 / REAL_TIME)) AS AVG_CPU,
	ROUND(MIN((SYS_TIME+USER_TIME) * 100 / REAL_TIME)) AS MIN_CPU,
	ROUND(MAX((SYS_TIME+USER_TIME) * 100 / REAL_TIME)) AS MAX_CPU
FROM   BENCH_RUN
WHERE  SYS_TIME IS NOT NULL AND USER_TIME IS NOT NULL
AND    REAL_TIME > 0
AND    STATUS <> 'FAILED' AND STATUS <> 'OUTL'
AND    WITH_OUTPUT = 'N'
GROUP  BY SYS, MACHINE, BENCH, FILE_ID
ORDER  BY SYS, MACHINE, BENCH, FILE_ID;


-- ############################################################################
-- Views with Benchmark Results for a Specific System:
-- ############################################################################


-- ============================================================================
-- Benchmark Results for BAM (Current Version, Standard Machine, No Output):
-- ============================================================================

CREATE VIEW BAM AS
SELECT BENCH,
	IMPL,
	FILE_ID AS FILE,
	LOAD_T,
	EXEC_T,
	-- TOTAL_T,
	CPU_T,
	REAL_T,
	MEM_MB
FROM RESULT
WHERE SYS = 'bam';

-- ============================================================================
-- Benchmark Results for XSB (Current Version, Standard Machine, No Output):
-- ============================================================================

CREATE VIEW XSB AS
SELECT BENCH,
	IMPL,
	FILE_ID AS FILE,
	LOAD_T,
	EXEC_T,
	-- TOTAL_T,
	CPU_T,
	REAL_T,
	MEM_MB
FROM RESULT
WHERE SYS = 'xsb';

-- ============================================================================
-- Benchmark Results for YAP (Current Version, Standard Machine, No Output):
-- ============================================================================

CREATE VIEW YAP AS
SELECT BENCH,
	IMPL,
	FILE_ID AS FILE,
	LOAD_T,
	EXEC_T,
	-- TOTAL_T,
	CPU_T,
	REAL_T,
	MEM_MB
FROM RESULT
WHERE SYS = 'yap';

-- ============================================================================
-- Benchmark Results for SQLite (Current Ver., Standard Machine, No Output):
-- ============================================================================

CREATE VIEW SQLITE AS
SELECT BENCH,
	IMPL,
	FILE_ID AS FILE,
	-- LOAD_T,
	-- EXEC_T,
	-- TOTAL_T,
	CPU_T,
	REAL_T,
	MEM_MB
FROM RESULT
WHERE SYS = 'sqlite3';

-- ============================================================================
-- Benchmark Results for PostgreSQL (Current Ver., Standard Machine, No Output):
-- ============================================================================

CREATE VIEW PSQL AS
SELECT BENCH,
	IMPL,
	FILE_ID,
	-- LOAD_T,
	-- EXEC_T,
	-- TOTAL_T,
	CPU_T,
	REAL_T,
	MEM_MB
FROM RESULT
WHERE SYS = 'psql';

-- ============================================================================
-- Benchmark Results for MariaDB (Current Ver., Standard Machine, No Output):
-- ============================================================================

CREATE VIEW MARIA AS
SELECT BENCH,
	IMPL,
	FILE_ID,
	-- LOAD_T,
	-- EXEC_T,
	-- TOTAL_T,
	CPU_T,
	REAL_T,
	MEM_MB
FROM RESULT
WHERE SYS = 'mariadb';

-- ============================================================================
-- Benchmark Results for Souffle (Current Ver., Standard Machine, No Output):
-- ============================================================================

CREATE VIEW SOUFFLE AS
SELECT BENCH,
	IMPL,
	FILE_ID,
	-- LOAD_T,
	-- EXEC_T,
	-- TOTAL_T,
	CPU_T,
	REAL_T,
	MEM_MB
FROM RESULT
WHERE SYS = 'souffle';

-- ============================================================================
-- Benchmark Results for Jena (Current Version, Standard Machine, No Output):
-- ============================================================================

CREATE VIEW JENA AS
SELECT BENCH,
	IMPL,
	FILE_ID,
	-- LOAD_T,
	-- EXEC_T,
	-- TOTAL_T,
	CPU_T,
	REAL_T,
	MEM_MB
FROM RESULT
WHERE SYS = 'jena';

-- ============================================================================
-- Benchmark Results for DataScript (Current Ver., Standard Machine, No Output):
-- ============================================================================

CREATE VIEW DATASCRIPT AS
SELECT BENCH,
	IMPL,
	FILE_ID AS FILE,
	-- LOAD_T,
	-- EXEC_T,
	-- TOTAL_T,
	CPU_T,
	REAL_T,
	MEM_MB
FROM RESULT
WHERE SYS = 'datascript';


-- ############################################################################
-- Views to Find/Check Best Option Settings for a System and a Benchmark:
-- ############################################################################


-- ============================================================================
-- Best Option Settings (in Real Time) for a Program for a Benchmark:
-- ============================================================================

-- This looks only at the runtimes of the standard program version
-- on the standard machine.

CREATE VIEW BEST_IMPL_REAL AS
SELECT	R.BENCH,
	R.FILE_ID,
	R.SYS,
	R.IMPL,
	R.REAL_T,
	I.STATUS
FROM	RESULT R, BENCHMARK B, INPUT_FILE F, BENCH_IMPL I
WHERE   R.BENCH = B.BENCH AND R.FILE_ID = F.FILE_ID
AND	I.BENCH = R.BENCH AND I.IMPL = R.IMPL AND I.SYS = R.SYS
AND NOT EXISTS(SELECT * FROM RESULT R2
		WHERE R2.BENCH = R.BENCH AND R2.SYS = R.SYS
		AND   R2.FILE_ID = R.FILE_ID
		AND   R2.REAL_T < R.REAL_T)
-- Eliminate trivial tests with run time 0: All options are best.
AND EXISTS(SELECT * FROM RESULT R3
		WHERE R3.BENCH = R.BENCH AND R3.SYS = R.SYS
		AND   R3.FILE_ID = R.FILE_ID
		AND   R3.REAL_T > 0)
ORDER BY B.SEQ_NO, F.SEQ_NO;


-- ============================================================================
-- Best Option Settings (in CPU Time) for a Program for a Benchmark:
-- ============================================================================

-- This looks only at the runtimes of the standard program version
-- on the standard machine.

CREATE VIEW BEST_IMPL_CPU AS
SELECT	R.BENCH,
	R.FILE_ID,
	R.SYS,
	R.IMPL,
	R.CPU_T,
	I.STATUS
FROM	RESULT R, BENCHMARK B, INPUT_FILE F, BENCH_IMPL I
WHERE   R.BENCH = B.BENCH AND R.FILE_ID = F.FILE_ID
AND	I.BENCH = R.BENCH AND I.IMPL = R.IMPL AND I.SYS = R.SYS
AND NOT EXISTS(SELECT * FROM RESULT R2
		WHERE R2.BENCH = R.BENCH AND R2.SYS = R.SYS
		AND   R2.FILE_ID = R.FILE_ID
		AND   R2.CPU_T < R.CPU_T)
-- Eliminate trivial tests with run time 0: All options are best.
AND EXISTS(SELECT * FROM RESULT R3
		WHERE R3.BENCH = R.BENCH AND R3.SYS = R.SYS
		AND   R3.FILE_ID = R.FILE_ID
		AND   R3.CPU_T > 0)
ORDER BY B.SEQ_NO, F.SEQ_NO;

-- ============================================================================
-- Compare Storage Managers for MariaDB:
-- ============================================================================

CREATE VIEW MARIA_COMPARE AS
SELECT INNO.BENCH,
	INNO.FILE_ID,
        F.LINES,
	INNO.REAL_T AS INNO_REAL_T,
	MEM.REAL_T AS MEM_REAL_T,
	CASE WHEN INNO.REAL_T > 0 THEN
                ROUND((INNO.REAL_T-MEM.REAL_T)*100/INNO.REAL_T)
                ELSE NULL END AS MEM_BETTER_PCT,
	CASE WHEN INNO.REAL_T < MEM.REAL_T THEN 'Y' ELSE ' ' END AS INNO_BETTER
FROM MARIA INNO, MARIA MEM, INPUT_FILE F
WHERE INNO.BENCH = MEM.BENCH AND INNO.FILE_ID = MEM.FILE_ID AND
        INNO.IMPL = 'i' AND MEM.IMPL = 'm' AND
	F.FILE_ID = INNO.FILE_ID
ORDER BY INNO.BENCH, F.SEQ_NO;


-- ############################################################################
-- Views for Plots of Performance Data vs. Cost Measures (only for TCFF):
-- ############################################################################


-- ============================================================================
-- Plot of Real Time vs. Cost for TCFF Benchmark:
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.

CREATE VIEW PLOT_REAL AS
SELECT	R.SYS,
	R.FILE_ID,
	F.SEQ_NO AS NO,
	G.COST,
	R.REAL_T,
	'"' || R.FILE_ID || '","' || G.COST || '","' || R.REAL_T || '"' AS CSV,
	R.FILE_ID || '	' || G.COST || '	' || R.REAL_T AS TSV
FROM	RESULT R, INPUT_FILE F, INPUT_GRAPH G, BENCH_IMPL I
WHERE   R.BENCH = 'tcff' AND R.FILE_ID = F.FILE_ID AND R.FILE_ID = G.FILE_ID
AND	I.BENCH = R.BENCH AND I.IMPL = R.IMPL AND I.SYS = R.SYS
AND	I.STATUS = 'B'
UNION
SELECT  DISTINCT P.SYS, '', -1, 0 AS COST, 0 AS REAL_T,
	'"df","cost","tavg"' as CSV,
	'df	cost	tavg' as TSV
FROM	TEST_SYS P
WHERE	EXISTS(SELECT * FROM RESULT R
		WHERE R.BENCH = 'tcff' AND R.SYS = P.SYS)
ORDER BY SYS, NO;

-- ============================================================================
-- Plot of CPU Time vs. Cost for TCFF Benchmark:
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.

CREATE VIEW PLOT_CPU AS
SELECT	R.SYS,
	R.FILE_ID,
	F.SEQ_NO AS NO,
	G.COST,
	R.CPU_T,
	'"' || R.FILE_ID || '","' || G.COST || '","' || R.CPU_T || '"' AS CSV,
	R.FILE_ID || '	' || G.COST || '	' || R.CPU_T AS TSV
FROM	RESULT R, INPUT_FILE F, INPUT_GRAPH G, BENCH_IMPL I
WHERE   R.BENCH = 'tcff' AND R.FILE_ID = F.FILE_ID AND R.FILE_ID = G.FILE_ID
AND	I.BENCH = R.BENCH AND I.IMPL = R.IMPL AND I.SYS = R.SYS
AND	I.STATUS = 'B'
UNION
SELECT  DISTINCT P.SYS, '', -1, 0 AS COST, 0 AS CPU_T,
	'"df","cost","tavg"' as CSV,
	'df	cost	tavg' as TSV
FROM	TEST_SYS P
WHERE	EXISTS(SELECT * FROM RESULT R
		WHERE R.BENCH = 'tcff' AND R.SYS = P.SYS)
ORDER BY SYS, NO;

-- ============================================================================
-- Plot of Memory vs. Number of Tuples for TCFF Benchmark:
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.

CREATE VIEW PLOT_MEM AS
SELECT	R.SYS,
	R.FILE_ID,
	F.SEQ_NO AS NO,
	G.NUM_EDGES + G.TC_SIZE AS FACTS,
	R.MEM_MB,
	'"' || R.FILE_ID || '","' || (G.NUM_EDGES+G.TC_SIZE) || '","' ||
		R.MEM_MB || '"' AS CSV,
	R.FILE_ID || '	' || (G.NUM_EDGES+G.TC_SIZE) || '	' ||
		R.MEM_MB AS TSV
FROM	RESULT R, INPUT_FILE F, INPUT_GRAPH G, BENCH_IMPL I
WHERE   R.BENCH = 'tcff' AND R.FILE_ID = F.FILE_ID AND R.FILE_ID = G.FILE_ID
AND	I.BENCH = R.BENCH AND I.IMPL = R.IMPL AND I.SYS = R.SYS
AND	I.STATUS = 'B'
UNION
SELECT  DISTINCT P.SYS, '', -1, 0 AS FACTS, 0 AS CPU_T,
	'"df","facts","mem"' as CSV,
	'df	facts	mem' as TSV
FROM	TEST_SYS P
WHERE	EXISTS(SELECT * FROM RESULT R
		WHERE R.BENCH = 'tcff' AND R.SYS = P.SYS)
ORDER BY SYS, NO;


-- ############################################################################
-- Benchmark Results for TCFF - Comparison of Different Systems:
-- ############################################################################


-- ============================================================================
-- Benchmark Results for TCFF (Comparison of Various Systems on CPU Time):
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.

CREATE VIEW TCFF_CPU AS
SELECT	BAM.FILE_ID,
	BAM.CPU_T AS BAM_CPU,
	XSB.CPU_T AS XSB_CPU,
	(CASE WHEN BAM.CPU_T=0 THEN NULL ELSE ROUND(XSB.CPU_T/BAM.CPU_T,1) END)
		AS XSB_F,
	YAP.CPU_T AS YAP_CPU,
	(CASE WHEN BAM.CPU_T=0 THEN NULL ELSE ROUND(YAP.CPU_T/BAM.CPU_T,1) END)
		AS YAP_F,
	JENA.CPU_T AS JENA_CPU,
	(CASE WHEN BAM.CPU_T=0 THEN NULL ELSE ROUND(JENA.CPU_T/BAM.CPU_T,1) END)
		AS JENA_F,
	SQLITE.CPU_T AS SQLITE_CPU,
	(CASE WHEN BAM.CPU_T=0 THEN NULL ELSE ROUND(SQLITE.CPU_T/BAM.CPU_T,1)
		END) AS SQLITE_F
FROM	(SELECT * FROM BEST_RESULT WHERE SYS='bam'AND BENCH='tcff') BAM
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='xsb' AND BENCH='tcff') XSB
	ON XSB.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='yap' AND BENCH='tcff') YAP
	ON YAP.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='jena' AND BENCH='tcff') JENA
	ON JENA.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='sqlite3' AND BENCH='tcff') SQLITE
	ON SQLITE.FILE_ID = BAM.FILE_ID
	JOIN INPUT_FILE F
        ON BAM.FILE_ID = F.FILE_ID
ORDER BY F.SEQ_NO;

-- ============================================================================
-- Benchmark Results for TCFF (Comparison of Various Systems on REAL Time):
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.
-- This version contains the runtime factors compared with BAM.

CREATE VIEW TCFF_REAL AS
SELECT	BAM.FILE_ID,
        F.SEQ_NO,
	BAM.REAL_T AS BAM_REAL,
	XSB.REAL_T AS XSB_REAL,
	(CASE WHEN BAM.REAL_T=0 THEN NULL
		ELSE ROUND(XSB.REAL_T/BAM.REAL_T,1) END)
		AS XSB_F,
	YAP.REAL_T AS YAP_REAL,
	(CASE WHEN BAM.REAL_T=0 THEN NULL
		ELSE ROUND(YAP.REAL_T/BAM.REAL_T,1) END)
		AS YAP_F,
	SQLITE.REAL_T AS SQLITE_REAL,
	(CASE WHEN BAM.REAL_T=0 THEN NULL
		ELSE ROUND(SQLITE.REAL_T/BAM.REAL_T,1) END)
		AS SQLITE_F,
	PSQL.REAL_T AS PSQL_REAL,
	(CASE WHEN BAM.REAL_T=0 THEN NULL
		ELSE ROUND(PSQL.REAL_T/BAM.REAL_T,1) END)
		AS PSQL_F,
	MARIA.REAL_T AS MARIA_REAL,
	(CASE WHEN BAM.REAL_T=0 THEN NULL
		ELSE ROUND(MARIA.REAL_T/BAM.REAL_T,1) END)
		AS MARIA_F,
	JENA.REAL_T AS JENA_REAL,
	(CASE WHEN BAM.REAL_T=0 THEN NULL
		ELSE ROUND(JENA.REAL_T/BAM.REAL_T,1) END)
		AS JENA_F,
	SOUFFLE.REAL_T AS SOUFFLE_REAL,
	(CASE WHEN BAM.REAL_T=0 THEN NULL
		ELSE ROUND(SOUFFLE.REAL_T/BAM.REAL_T,1) END)
		AS SOUFFLE_F
FROM	(SELECT * FROM BEST_RESULT WHERE SYS='bam'AND BENCH='tcff') BAM
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='xsb' AND BENCH='tcff') XSB
	ON XSB.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='yap' AND BENCH='tcff') YAP
	ON YAP.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='sqlite3' AND BENCH='tcff') SQLITE
	ON SQLITE.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='psql' AND BENCH='tcff') PSQL
	ON PSQL.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='mariadb' AND BENCH='tcff') MARIA
	ON MARIA.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='jena' AND BENCH='tcff') JENA
	ON JENA.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='souffle' AND BENCH='tcff')SOUFFLE
	ON SOUFFLE.FILE_ID = BAM.FILE_ID
	JOIN INPUT_FILE F
        ON BAM.FILE_ID = F.FILE_ID
ORDER BY F.SEQ_NO;

-- ============================================================================
-- Benchmark Results for TCFF (Comparison on REAL Time, Only Factors):
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.

-- It is a projection of TCFF_REAL that shows only the factors compared to BAM.
-- This makes the output rows a little smaller.

CREATE VIEW TCFF_REAL_FACTORS AS
SELECT	FILE_ID, BAM_REAL,
	XSB_F, YAP_F, SQLITE_F, PSQL_F, MARIA_F, JENA_F, SOUFFLE_F
FROM	TCFF_REAL
ORDER BY SEQ_NO;

-- ============================================================================
-- Benchmark Results for TCFF (Comparison of Various Systems on REAL Time):
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.

CREATE VIEW TCFF_REAL_LATEX AS
SELECT	FILE_ID,
	RPAD(REPLACE(FILE_ID, '_', '\_'), 12) || '&' ||
	TO_CHAR(BAM_REAL, '9990.999') || '&' ||
	(CASE WHEN XSB_F IS NULL THEN '      '
		ELSE TO_CHAR(XSB_F, '990.9') END) || '&' ||
	(CASE WHEN YAP_F IS NULL THEN '      '
		ELSE TO_CHAR(YAP_F, '990.9') END) || '&' ||
	(CASE WHEN SQLITE_F IS NULL THEN '      '
		ELSE TO_CHAR(SQLITE_F, '990.9') END) || '&' ||
	(CASE WHEN PSQL_F IS NULL THEN '      '
		ELSE TO_CHAR(PSQL_F, '990.9') END) || '&' ||
	(CASE WHEN MARIA_F IS NULL THEN '      '
		ELSE TO_CHAR(MARIA_F, '990.9') END) || '&' ||
	(CASE WHEN JENA_F IS NULL THEN '      '
		ELSE TO_CHAR(JENA_F, '990.9') END) || '&' ||
	(CASE WHEN SOUFFLE_F IS NULL THEN '      '
		ELSE TO_CHAR(SOUFFLE_F, '990.9') END) || '\\'
	AS LATEX
FROM	TCFF_REAL
ORDER BY SEQ_NO;

-- ============================================================================
-- Benchmark Results for TCFF (Comparison of Various Systems on REAL Time):
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.

-- This version contains the runtime factors compared with XSB
-- and rounds the factors to two decimal places after the decimal point.
-- Otherwise it is identical to TCFF_REAL (which compares with BAM).

CREATE VIEW TCFF_REAL_XSB AS
SELECT	XSB.FILE_ID,
        F.SEQ_NO,
	XSB.REAL_T AS XSB_REAL,
	BAM.REAL_T AS BAM_REAL,
	(CASE WHEN XSB.REAL_T=0 THEN NULL
		ELSE ROUND(BAM.REAL_T/XSB.REAL_T,2) END)
		AS BAM_F,
	YAP.REAL_T AS YAP_REAL,
	(CASE WHEN XSB.REAL_T=0 THEN NULL
		ELSE ROUND(YAP.REAL_T/XSB.REAL_T,2) END)
		AS YAP_F,
	SQLITE.REAL_T AS SQLITE_REAL,
	(CASE WHEN XSB.REAL_T=0 THEN NULL
		ELSE ROUND(SQLITE.REAL_T/XSB.REAL_T,2) END)
		AS SQLITE_F,
	PSQL.REAL_T AS PSQL_REAL,
	(CASE WHEN XSB.REAL_T=0 THEN NULL
		ELSE ROUND(PSQL.REAL_T/XSB.REAL_T,2) END)
		AS PSQL_F,
	MARIA.REAL_T AS MARIA_REAL,
	(CASE WHEN XSB.REAL_T=0 THEN NULL
		ELSE ROUND(MARIA.REAL_T/XSB.REAL_T,2) END)
		AS MARIA_F,
	JENA.REAL_T AS JENA_REAL,
	(CASE WHEN XSB.REAL_T=0 THEN NULL
		ELSE ROUND(JENA.REAL_T/XSB.REAL_T,2) END)
		AS JENA_F,
	SOUFFLE.REAL_T AS SOUFFLE_REAL,
	(CASE WHEN XSB.REAL_T=0 THEN NULL
		ELSE ROUND(SOUFFLE.REAL_T/XSB.REAL_T,2) END)
		AS SOUFFLE_F
FROM	(SELECT * FROM BEST_RESULT WHERE SYS='xsb'AND BENCH='tcff') XSB
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='bam' AND BENCH='tcff') BAM
	ON BAM.FILE_ID = XSB.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='yap' AND BENCH='tcff') YAP
	ON YAP.FILE_ID = XSB.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='sqlite3' AND BENCH='tcff') SQLITE
	ON SQLITE.FILE_ID = XSB.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='psql' AND BENCH='tcff') PSQL
	ON PSQL.FILE_ID = XSB.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='mariadb' AND BENCH='tcff') MARIA
	ON MARIA.FILE_ID = XSB.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='jena' AND BENCH='tcff') JENA
	ON JENA.FILE_ID = XSB.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='souffle' AND BENCH='tcff')SOUFFLE
	ON SOUFFLE.FILE_ID = XSB.FILE_ID
	JOIN INPUT_FILE F
        ON XSB.FILE_ID = F.FILE_ID
ORDER BY F.SEQ_NO;

-- ============================================================================
-- Benchmark Results for TCFF (Comparison on REAL Time, Only Factors):
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.

-- It is a projection of TCFF_REAL that shows only the factors compared to BAM.
-- This makes the output rows a little smaller.

CREATE VIEW TCFF_REAL_XSB_FACTORS AS
SELECT	FILE_ID, XSB_REAL,
	BAM_F, YAP_F, SQLITE_F, PSQL_F, MARIA_F, JENA_F, SOUFFLE_F
FROM	TCFF_REAL_XSB
ORDER BY SEQ_NO;


-- ============================================================================
-- Benchmark Results for TCFF (Comparison of Various Systems on REAL Time):
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.

CREATE VIEW TCFF_REAL_XSB_LATEX AS
SELECT	FILE_ID,
	RPAD(REPLACE(FILE_ID, '_', '\_'), 10) || '&' ||
	TO_CHAR(XSB_REAL, '9990.999') || '&' ||
	(CASE WHEN BAM_F IS NULL     THEN '       '
		ELSE TO_CHAR(BAM_F, '990.99') END) || '&' ||
	(CASE WHEN YAP_F IS NULL     THEN '       '
		ELSE TO_CHAR(YAP_F, '990.99') END) || '&' ||
	(CASE WHEN SQLITE_F IS NULL  THEN '       '
		ELSE TO_CHAR(SQLITE_F, '990.99') END) || '&' ||
	(CASE WHEN PSQL_F IS NULL    THEN '       '
		ELSE TO_CHAR(PSQL_F, '990.99') END) || '&' ||
	(CASE WHEN MARIA_F IS NULL   THEN '       '
		ELSE TO_CHAR(MARIA_F, '990.99') END) || '&' ||
	(CASE WHEN JENA_F IS NULL    THEN '       '
		ELSE TO_CHAR(JENA_F, '990.99') END) || '&' ||
	(CASE WHEN SOUFFLE_F IS NULL THEN '       '
		ELSE TO_CHAR(SOUFFLE_F, '990.99') END) || '\\'
	AS LATEX
FROM	TCFF_REAL_XSB
ORDER BY SEQ_NO;


-- ============================================================================
-- Benchmark Results for TCFF (Comparison of Various Systems on REAL Time):
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.

CREATE VIEW TCFF_REAL_XSB_HTML AS
SELECT -2 AS SEQ_NO, NULL AS FILE_ID,
	'<table>' AS HTML
FROM    DUMMY_TAB
UNION ALL
SELECT -1 AS SEQ_NO, NULL AS FILE_ID,
	'<tr>' ||
	     '<th class="colGraph">'   || 'Graph'          || '</th>' || '
	' || '<th class="colXSBTime">' || 'XSB (s)'        || '</th>' || '
	' || '<th class="colBAM"    >' || 'BAM'            || '</th>' || '
	' || '<th class="colYAP"    >' || 'YAP'            || '</th>' || '
	' || '<th class="colSQLite" >' || 'SQLite'         || '</th>' || '
	' || '<th class="colPSQL"   >' || 'PostgreSQL'     || '</th>' || '
	' || '<th class="colMaria"  >' || 'MariaDB'        || '</th>' || '
	' || '<th class="colJena"   >' || 'Jena'           || '</th>' || '
	' || '<th class="colSouffle">' || 'Souffl&eacute;' || '</th>' || '
	' || '</tr>'
	AS HTML
FROM    DUMMY_TAB
UNION ALL
SELECT	SEQ_NO, FILE_ID,
	'<tr>' ||
	'<td class="colGraph">' || FILE_ID || '</td>' || '
	' || '<td class="colXSBTime">' || TO_CHAR(XSB_REAL, '9990.999') ||
		'&nbsp;s' || '</td>' || '
	' || '<td class="colBAM"    >' ||
		(CASE WHEN BAM_F IS NULL     THEN '&nbsp; '
			ELSE TO_CHAR(BAM_F,     '990.99') END) || '</td>' || '
	' || '<td class="colYAP"    >' ||
		(CASE WHEN YAP_F IS NULL     THEN '&nbsp; '
			ELSE TO_CHAR(YAP_F,     '990.99') END) || '</td>' || '
	' || '<td class="colSQLite" >' ||
		(CASE WHEN SQLITE_F IS NULL  THEN '&nbsp; '
			ELSE TO_CHAR(SQLITE_F,  '990.99') END) || '</td>' || '
	' || '<td class="colPSQL"   >' ||
		(CASE WHEN PSQL_F IS NULL    THEN '&nbsp; '
			ELSE TO_CHAR(PSQL_F,    '990.99') END) || '</td>' || '
	' || '<td class="colMaria"  >' ||
		(CASE WHEN MARIA_F IS NULL   THEN '&nbsp; '
			ELSE TO_CHAR(MARIA_F,   '990.99') END) || '</td>' || '
	' || '<td class="colJena"   >' ||
		(CASE WHEN JENA_F IS NULL    THEN '&nbsp; '
			ELSE TO_CHAR(JENA_F,    '990.99') END) || '</td>' || '
	' || '<td class="colSouffle">' ||
		(CASE WHEN SOUFFLE_F IS NULL THEN '&nbsp; '
			ELSE TO_CHAR(SOUFFLE_F, '990.99') END) || '</td>' || '
	' || '</tr>'
	AS HTML
FROM	TCFF_REAL_XSB
UNION ALL
SELECT 1000000 AS SEQ_NO, NULL AS FILE_ID,
	'</table>' AS HTML
FROM    DUMMY_TAB
ORDER BY SEQ_NO;


-- ============================================================================
-- Benchmark Results for TCFF (Comparison of Various Systems on Memory):
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.

CREATE VIEW TCFF_MEM AS
SELECT	BAM.FILE_ID,
	BAM.MEM_MB AS BAM_MEM,
	XSB.MEM_MB AS XSB_MEM,
	(CASE WHEN BAM.MEM_MB=0 THEN NULL
		ELSE ROUND(XSB.MEM_MB/BAM.MEM_MB,1) END)
		AS XSB_F,
	YAP.MEM_MB AS YAP_MEM,
	(CASE WHEN BAM.MEM_MB=0 THEN NULL
		ELSE ROUND(YAP.MEM_MB/BAM.MEM_MB,1) END)
		AS YAP_F,
	JENA.MEM_MB AS JENA_MEM,
	(CASE WHEN BAM.MEM_MB=0 THEN NULL
		ELSE ROUND(JENA.MEM_MB/BAM.MEM_MB,1) END)
		AS JENA_F,
	SQLITE.MEM_MB AS SQLITE_MEM,
	(CASE WHEN BAM.MEM_MB=0 THEN NULL
		ELSE ROUND(SQLITE.MEM_MB/BAM.MEM_MB,1) END)
		AS SQLITE_F
FROM	(SELECT * FROM BEST_RESULT WHERE SYS='bam'AND BENCH='tcff') BAM
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='xsb' AND BENCH='tcff') XSB
	ON XSB.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='yap' AND BENCH='tcff') YAP
	ON YAP.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='jena' AND BENCH='tcff') JENA
	ON JENA.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='sqlite3' AND BENCH='tcff') SQLITE
	ON SQLITE.FILE_ID = BAM.FILE_ID
	JOIN INPUT_FILE F
        ON BAM.FILE_ID = F.FILE_ID
ORDER BY F.SEQ_NO;

-- ============================================================================
-- Benchmark Results for TCFF (Comparison of Various Systems on CPU Time):
-- ============================================================================

-- This looks only at the runtimes of the standard program version,
-- best implementation, on the standard machine.

CREATE VIEW TCFF_CPU_LATEX AS
SELECT	BAM.FILE_ID,
	RPAD(REPLACE(BAM.FILE_ID, '_', '\_'), 12) || '&' ||
	TO_CHAR(BAM.CPU_T, '9990.999') || '&' ||
	(CASE WHEN XSB.CPU_T IS NULL THEN '         &       ' ELSE
		(TO_CHAR(XSB.CPU_T, '9990.999') || '&' ||
		(CASE WHEN BAM.CPU_T=0 THEN '       '
			ELSE TO_CHAR(XSB.CPU_T/BAM.CPU_T, '9990.9') END))
	END) || '&' ||
	(CASE WHEN YAP.CPU_T IS NULL THEN '         &       ' ELSE
		(TO_CHAR(YAP.CPU_T, '9990.999') || '&' ||
		(CASE WHEN BAM.CPU_T=0 THEN '       '
			ELSE TO_CHAR(YAP.CPU_T/BAM.CPU_T, '9990.9') END))
	END) || '&' ||
	(CASE WHEN JENA.CPU_T IS NULL THEN '         &       ' ELSE
		(TO_CHAR(JENA.CPU_T,'9990.999') || '&' ||
		(CASE WHEN BAM.CPU_T=0 THEN '       '
			ELSE TO_CHAR(JENA.CPU_T/BAM.CPU_T, '9990.9') END))
	END) || '&' ||
	(CASE WHEN SQLITE.CPU_T IS NULL THEN '         &       ' ELSE
		(TO_CHAR(SQLITE.CPU_T,'9990.999') || '&' ||
		(CASE WHEN BAM.CPU_T=0 THEN '       '
			ELSE TO_CHAR(SQLITE.CPU_T/BAM.CPU_T, '9990.9') END))
	END) || '\\'
	AS LATEX
FROM	(SELECT * FROM BEST_RESULT WHERE SYS='bam'AND BENCH='tcff') BAM
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='xsb' AND BENCH='tcff') XSB
	ON XSB.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='yap' AND BENCH='tcff') YAP
	ON YAP.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='jena' AND BENCH='tcff') JENA
	ON JENA.FILE_ID = BAM.FILE_ID
	LEFT JOIN
	(SELECT * FROM BEST_RESULT WHERE SYS='sqlite3' AND BENCH='tcff') SQLITE
	ON SQLITE.FILE_ID = BAM.FILE_ID
	JOIN INPUT_FILE F
	ON F.FILE_ID = BAM.FILE_ID
ORDER BY F.SEQ_NO;


-- ############################################################################
-- Graph Data in Various Formats:
-- ############################################################################


-- ============================================================================
-- Graph Data as LaTeX:
-- ============================================================================

CREATE VIEW INPUT_GRAPH_LATEX AS
SELECT	G.FILE_ID,
	REPLACE(G.FILE_ID, '_', '\_') || '&' ||
	G.NUM_NODES || '&' || G.NUM_EDGES || '&' ||
--	G.MIN_IN_DEGREE || '&' || G.MAX_IN_DEGREE || '&' ||
--	G.MIN_OUT_DEGREE || '&' || G.MAX_OUT_DEGREE || '&' ||
        (CASE WHEN G.MIN_IN_DEGREE = G.MAX_IN_DEGREE
		THEN G.MIN_IN_DEGREE || '&'
		ELSE G.MIN_IN_DEGREE || '--' || G.MAX_IN_DEGREE || '&' END) ||
        (CASE WHEN G.MIN_OUT_DEGREE = G.MAX_OUT_DEGREE
		THEN G.MIN_OUT_DEGREE || '&'
		ELSE G.MIN_OUT_DEGREE || '--' || G.MAX_OUT_DEGREE || '&' END)||
	(CASE WHEN CYCLES='Y' THEN 'yes'
		WHEN CYCLES='N' THEN 'no'
		ELSE '?' END) || '&' ||
	G.TC_SIZE || '&' ||
		G.NUM_ITER || '&' ||
		G.COST || '\\'
		AS LATEX
FROM	INPUT_GRAPH G, INPUT_FILE F
WHERE   G.FILE_ID = F.FILE_ID
ORDER BY F.SEQ_NO;

-- ============================================================================
-- Graph Data as HTML:
-- ============================================================================

CREATE VIEW INPUT_GRAPH_HTML AS
SELECT -2 AS SEQ_NO, NULL AS FILE_ID,
	'<table>' AS HTML
FROM    DUMMY_TAB
UNION ALL
SELECT -1 AS SEQ_NO, NULL AS FILE_ID,
	'<tr>' ||
	'<th class="colGraph">' || 'Graph'          || '</th>' || '
	' || '<th class="colNodes"   >' || 'Nodes'           || '</th>' || '
	' || '<th class="colEdges"   >' || 'Edges'           || '</th>' || '
	' || '<th class="colInDeg"   >' || 'In-Deg.'         || '</th>' || '
	' || '<th class="colOutDeg"  >' || 'Out-Deg.'        || '</th>' || '
	' || '<th class="colCyc"     >' || 'Cyc.'            || '</th>' || '
	' || '<th class="colTCSize"  >' || 'TC&nbsp;Size.'   || '</th>' || '
	' || '<th class="colIter"    >' || 'Iter.'           || '</th>' || '
	' || '<th class="colRuleInst">' || 'Rule&nbsp;Inst.' || '</th>' || '
	' || '</tr>'
	AS HTML
FROM    DUMMY_TAB
UNION ALL
SELECT	F.SEQ_NO, F.FILE_ID,
	'<tr>' ||
	'<td class="colGraph">' ||
		'<a href="data_tsv/' || F.FILENAME || '.tsv">' ||
		F.FILE_ID || '</a>' || '</td>' || '
	' || '<td class="colNodes"   >' || NUM_NODES || '</td>' || '
	' || '<td class="colEdges"   >' || NUM_EDGES || '</td>' || '
	' || '<td class="colInDeg"   >' ||
		(CASE WHEN G.MIN_IN_DEGREE = G.MAX_IN_DEGREE
			THEN CAST(G.MIN_IN_DEGREE AS VARCHAR)
			ELSE G.MIN_IN_DEGREE || '-' || G.MAX_IN_DEGREE END)
	                                              || '</td>' || '
	' || '<td class="colOutDeg"  >' ||
		(CASE WHEN G.MIN_OUT_DEGREE = G.MAX_OUT_DEGREE
			THEN CAST(G.MIN_OUT_DEGREE AS VARCHAR)
			ELSE G.MIN_OUT_DEGREE || '-' || G.MAX_OUT_DEGREE END)
	                                              || '</td>' || '
	' || '<td class="colCyc"     >' ||
		(CASE WHEN CYCLES='Y' THEN 'yes'
			WHEN CYCLES='N' THEN 'no'
			ELSE '?' END)                 || '</td>' || '
	' || '<td class="colTCSize"  >' || TC_SIZE   || '</td>' || '
	' || '<td class="colIter"    >' || NUM_ITER  || '</td>' || '
	' || '<td class="colRuleInst">' || COST      || '</td>' || '
	' || '</tr>'
	AS HTML
FROM	INPUT_GRAPH G, INPUT_FILE F
WHERE   G.FILE_ID = F.FILE_ID
UNION ALL
SELECT 1000000 AS SEQ_NO, NULL AS FILE_ID,
	'</table>' AS HTML
FROM    DUMMY_TAB
ORDER BY SEQ_NO;


-- ############################################################################
-- Support for Reasearch on Runtime Estimation:
-- ############################################################################


-- ============================================================================
-- System Speed in Real Time:
-- ============================================================================

CREATE VIEW SPEED_REAL AS
SELECT SYS, FILE_ID, ROUND((COST/REAL_T)/1000) as SPEED, COST, REAL_T
FROM   PLOT_REAL
WHERE  REAL_T > 0
ORDER  BY SYS, SPEED;

-- ============================================================================
-- System Speed in Real Time With Startup Time:
-- ============================================================================

CREATE VIEW SPEED_REAL_DIFF AS
SELECT P.SYS, P.FILE_ID,
       ROUND(((P.COST-S.COST)/(P.REAL_T-S.REAL_T))/1000) as SPEED,
       P.COST, P.REAL_T,
       (P.COST-S.COST) AS DIFF_COST, (P.REAL_T-S.REAL_T) AS DIFF_REAL_T
FROM   PLOT_REAL P, PLOT_REAL S
WHERE  P.REAL_T-S.REAL_T > 0
AND    P.SYS = S.SYS
AND    S.FILE_ID = 'm4_2ki'
ORDER  BY SYS, SPEED;

-- ============================================================================
-- System Speed in CPU Time:
-- ============================================================================

CREATE VIEW SPEED_CPU AS
SELECT SYS, FILE_ID, ROUND((COST/CPU_T)/1000) as SPEED, COST, CPU_T
FROM   PLOT_CPU
WHERE  CPU_T > 0
ORDER  BY SYS, SPEED;

-- ============================================================================
-- Memory Efficiency:
-- ============================================================================

CREATE VIEW MEM_EFFICIENCY AS
SELECT SYS, FILE_ID, ROUND(MEM_MB*1024*1024/FACTS) as BYTES_PER_FACT,
       FACTS, MEM_MB
FROM   PLOT_MEM
WHERE  FACTS > 0
ORDER  BY SYS, BYTES_PER_FACT;

-- ============================================================================
-- Compute Run Time Estimates (Real Time):
-- ============================================================================

CREATE VIEW ESTIMATE_FORMULA AS
SELECT R.SYS, E.STARTUP, E.LOAD_IDX, E.RULE_APP, E.RESULT,
       G.FILE_ID,
       ROUND(CAST(
         E.STARTUP +
         (G.NUM_EDGES * (LOG(G.NUM_EDGES)/LOG(2)) * E.LOAD_IDX / 1000000000) +
         (G.COST * E.RULE_APP / 1000000000) +
         (G.TC_SIZE * E.RESULT / 1000000000) AS NUMERIC),
       3) AS ESTIMATE,
       R.REAL_T
FROM   RESULT R, INPUT_GRAPH G, ESTIMATE_PAR E
WHERE  R.SYS = E.SYS AND E.BENCH = 'tcff' AND R.BENCH = 'tcff'
AND    R.FILE_ID = G.FILE_ID
ORDER  BY R.SYS, E.STARTUP, E.LOAD_IDX, E.RULE_APP, E.RESULT, ESTIMATE;

-- ============================================================================
-- Check System Speed Estimates (Real Time):
-- ============================================================================

CREATE VIEW ESTIMATE_ERR AS
SELECT E.SYS, E.STARTUP, E.LOAD_IDX, E.RULE_APP, E.RESULT,
       E.FILE_ID, E.ESTIMATE, E.REAL_T,
       E.REAL_T - E.ESTIMATE AS DIFF,
       ROUND((E.REAL_T - E.ESTIMATE) * 100 / E.ESTIMATE) AS DIFF_PCT
FROM   ESTIMATE_FORMULA E
ORDER  BY E.SYS, E.STARTUP, E.LOAD_IDX, E.RULE_APP, E.RESULT, DIFF_PCT;

-- ============================================================================
-- Count the Number of Wrong Estimates (more than 2s wrong and more than 20%):
-- ============================================================================

CREATE VIEW NUM_BAD_ESTIMATES AS
SELECT SYS, STARTUP, LOAD_IDX, RULE_APP, RESULT,
	COUNT(*) AS BAD_ESTIMATES, SUM(ABS(DIFF)) AS ABS_ERR
FROM   ESTIMATE_ERR
WHERE  ABS(DIFF_PCT) > 20
GROUP  BY SYS, STARTUP, LOAD_IDX, RULE_APP, RESULT
ORDER  BY SYS, STARTUP, LOAD_IDX, RULE_APP, RESULT;

-- ============================================================================
-- Generate C++ Objects with Graph Data and Runtime (for Optimization):
-- ============================================================================

CREATE VIEW ESTIMATE_DATA AS
SELECT R.SYS, R.BENCH, R.IMPL,
       'data_c("' || G.FILE_ID || '", ' ||
		G.NUM_NODES || ', ' ||
		G.NUM_EDGES || ', ' ||
		'INT64_C(' || G.TC_SIZE || '), ' ||
		'INT64_C(' || G.COST    || '), ' ||
		G.NUM_ITER  || ', ' ||
		ROUND(R.REAL_T*1000)|| '),' AS OBJ
FROM	INPUT_GRAPH G, INPUT_FILE F, BEST_RESULT R
WHERE   G.FILE_ID = F.FILE_ID AND R.FILE_ID = F.FILE_ID
ORDER BY R.SYS, R.BENCH, F.SEQ_NO;

-- ============================================================================
-- Load time per million edges in s:
-- ============================================================================

CREATE VIEW LOAD_EFFICIENCY AS
SELECT R.SYS, R.BENCH, R.IMPL, R.LOAD_T, G.NUM_EDGES,
		ROUND(R.LOAD_T * 1000 * 1000 / G.NUM_EDGES, 3)
			AS LOAD_PER_M_EDGES
FROM    INPUT_GRAPH G, INPUT_FILE F, BEST_RESULT R
WHERE   G.FILE_ID = F.FILE_ID AND R.FILE_ID = F.FILE_ID
ORDER BY R.SYS, R.BENCH, F.SEQ_NO;

-- ============================================================================
-- Search Range for Parameter Values:
-- ============================================================================

CREATE VIEW PAR_RANGE AS
SELECT R.SYS, G.FILE_ID,
		ROUND(R.REAL_T * 1000 * 1000 / G.NUM_EDGES, 3) AS PAR_LOAD,
		ROUND(R.REAL_T * 1000 * 1000 / G.COST, 3)      AS PAR_RULE,
		ROUND(R.REAL_T * 1000 * 1000 / G.TC_SIZE, 3)   AS PAR_RESULT,
		ROUND(R.REAL_T * 1000 * 1000 / G.NUM_NODES, 3) AS PAR_DATA,
		ROUND(R.REAL_T * 1000 * 1000 / G.NUM_ITER, 3)  AS PAR_ITER
FROM    INPUT_GRAPH G, INPUT_FILE F, BEST_RESULT R
WHERE   G.FILE_ID = F.FILE_ID AND R.FILE_ID = F.FILE_ID
AND     R.BENCH = 'tcff'
ORDER BY R.SYS, R.BENCH, F.SEQ_NO;



-- ############################################################################
-- Output Table Data in Markdown Format (for Documentation)
-- ############################################################################


-- ============================================================================
-- BENCH_IMPL_MD:
-- ============================================================================

CREATE VIEW BENCH_IMPL_MD AS
SELECT 0 AS SECTION, ' ' AS SYS, ' ' AS BENCH, ' ' AS IMPL,
	'|SYS|BENCH|IMPL|STATUS|DESCRIPTION|' AS MD
UNION ALL
SELECT 1 AS SECTION, ' ' AS SYS, ' ' AS BENCH, ' ' AS IMPL,
	'|----|-----|----|------|-----------|' AS MD
UNION ALL
SELECT 2 AS LINE, SYS, BENCH, IMPL,
       '|' || SYS || '|' ||  BENCH || '|' || IMPL || '|' || STATUS ||
	'|' || DESCRIPTION || '|' AS MD
FROM   BENCH_IMPL
ORDER BY SECTION, SYS, BENCH, IMPL;

-- ============================================================================
-- INPUT_FILE_MD:
-- ============================================================================

CREATE VIEW INPUT_FILE_MD AS
SELECT 0 AS SECTION, 0 AS SEQ_NO,
	'|FILE_ID|SEQ_NO|ITYPE|LINES|BYTES|FILENAME|' AS MD
UNION ALL
SELECT 1 AS SECTION, 0 AS SEQ_NO,
	'|-------|-----:|-----|----:|----:|--------|' AS MD
UNION ALL
SELECT 2 AS SECTION, SEQ_NO,
       '|' || FILE_ID || '|' ||  SEQ_NO || '|' || ITYPE || '|' || LINES ||
	'|' || BYTES || '|' || FILENAME || '|' AS MD
FROM   INPUT_FILE
ORDER BY SECTION, SEQ_NO;


-- ============================================================================
-- INPUT_GRAPH_MD:
-- ============================================================================

CREATE VIEW INPUT_GRAPH_MD AS
SELECT 0 AS SECTION, 0 AS SEQ_NO,
	'|FILE_ID|' ||
	'GTYPE|PARAM_1|PARAM_2|' ||
	'NUM_NODES|NUM_EDGES|' ||
	'DUP_EDGES|LOOPS|' ||
	'MIN_IN_DEGREE|MAX_IN_DEGREE|' ||
	'MIN_OUT_DEGREE|MAX_OUT_DEGREE|' ||
	'CYCLES|TC_SIZE|NUM_ITER|COST|' AS MD
UNION ALL
SELECT 1 AS SECTION, 0 AS SEQ_NO,
	'|-------|' ||
	'-----|------:|------:|' ||
	'--------:|--------:|' ||
	'--------:|:---:|' ||
	'------------:|------------:|' ||
	'-------------:|-------------:|' ||
	':----:|------:|-------:|---:|' AS MD
UNION ALL
SELECT 2 AS SECTION, F.SEQ_NO,
       '|' || G.FILE_ID || '|' ||
	GTYPE || '|' ||
	(CASE WHEN PARAM_1 IS NULL THEN ' ' ELSE CAST(PARAM_1 AS VARCHAR) END)
        || '|' ||
	(CASE WHEN PARAM_2 IS NULL THEN ' ' ELSE CAST(PARAM_2 AS VARCHAR) END)
	|| '|' ||
	NUM_NODES || '|' ||  NUM_EDGES || '|' ||
	DUP_EDGES || '|' || LOOPS || '|' ||
	MIN_IN_DEGREE || '|' || MAX_IN_DEGREE || '|' ||
	MIN_OUT_DEGREE || '|' || MAX_OUT_DEGREE || '|' ||
	CYCLES || '|' || TC_SIZE || '|' || NUM_ITER || '|' || COST || '|' AS MD
FROM   INPUT_GRAPH G, INPUT_FILE F
WHERE  G.FILE_ID = F.FILE_ID
ORDER BY SECTION, SEQ_NO;


-- ############################################################################
-- Support for Benchmark Execution Scripts:
-- ############################################################################

-- ============================================================================
-- BENCH_FILE_LIST: For RUN_BENCH Scripts
-- ============================================================================

-- This produces the benchmark file list for use in the RUN_BENCH scripts.
-- The result is to be used in a "case" switch in the bash script.

CREATE VIEW BENCH_FILE_LIST AS
SELECT '' AS FILE_ID, '' AS ITYPE, T.SEQ_NO AS TNO, -2 AS FNO,
	'' AS DATA_P,
	'' AS DATA_TSV
FROM INPUT_TYPE T
UNION ALL
SELECT '' AS FILE_ID, '' AS ITYPE, T.SEQ_NO AS TNO, -1 AS FNO,
	'    ' || '# ' || T.HEADLINE || ':' AS DATA_P,
	'    ' || '# ' || T.HEADLINE || ':' AS DATA_TSV
FROM INPUT_TYPE T
UNION ALL
SELECT F.FILE_ID, F.ITYPE, T.SEQ_NO AS TNO, F.SEQ_NO AS FNO,
	'    ' || RPAD((F.FILE_ID || ')'), 11) ||
	' FTYPE=' || RPAD((F.ITYPE || ';'), 11) ||
	'FILE=../data_p/' || F.FILENAME || '.P' || ' ;;' AS DATA_P,
	'    ' || RPAD((F.FILE_ID || ')'), 9) ||
	' FTYPE=' || RPAD((F.ITYPE || ';'), 9) ||
	'FILE=../data_tsv/' || F.FILENAME || '.tsv' || ' ;;' AS DATA_TSV
FROM INPUT_FILE F, INPUT_TYPE T
WHERE F.ITYPE = T.ITYPE
ORDER BY TNO, FNO;

-- Note that RPAD cuts off strings that are too long (in PostgreSQL).
-- This is only safe because FILE_ID is declared as VARCHAR(10)
-- and ITYPE as VARCHAR(8).

-- ============================================================================
-- TC_FILE_IDs: For RUN_BENCH_LIST Scripts
-- ============================================================================

CREATE VIEW TC_FILE_IDS AS
SELECT -1 AS SEQ_NO, ' ' AS GTYPE,
	'TC_FILE_IDS="\' AS FILE_IDS
UNION ALL
SELECT T.SEQ_NO, T.GTYPE,
	CHR(9) || STRING_AGG(G.FILE_ID, ' ') || ' ' || '\' AS FILE_IDS
FROM INPUT_GRAPH G, GRAPH_TYPE T
WHERE G.GTYPE = T.GTYPE
GROUP BY T.SEQ_NO, T.GTYPE
UNION ALL
SELECT 1000000 AS SEQ_NO, ' ' AS GTYPE,
	CHR(9) || '"' AS FILE_IDS
ORDER BY SEQ_NO;

-- CHR(9) is a TAB character.
-- PostgreSQL ignores literal TAB characters on input.

-- For MySQL, replace
--	STRING_AGG(G.FILE_ID, ' ')
-- by
--	GROUP_CONCAT((G.FILE_ID SEPARATOR ' '))

