Documentation for the rbench Benchmark Database
===============================================

This is a database to store benchmark results
for programs that answer queries from large input data files,
e.g. logic programming systems, deductive databases, relational databases,
and graph databases/RDF stores.

The benchmarks were inspired by the
[OpenRuleBench](http://rulebench.semwebcentral.org/),
see also
[OpenRuleBench Download](http://www3.cs.stonybrook.edu/~pfodor/openrulebench/download.html).

However, the OpenRuleBench scripts do not support executing the same
benchmark for the same program multiple times.
This is important to get reliable data,
especially when one is also interested in slight differences.
Since we are developing our own deductive database based on the
[Push Method](http://users.informatik.uni-halle.de/~brass/push/),
we are interested even in minor runtime improvements.
Furthermore, it is very important for us to compare different versions
of our software,
or the same version with different compiler settings.
Moreover, we run benchmarks on different machines
with different operating systems.
The OpenRuleBench scripts have no support for this
(they were developed with a different application in mind,
namely to get an overview of the performance of rule-based systems
for semantic web applications).

One soon loses the overview if one keeps the measurements in simple text files.
Therefore we developed a relational database
to store and evaluate runtime measurements (and memory consumption, too).

A problem for fair benchmarks is to find the best settings
for each systems for a given benchmark,
e.g. which indexes should be used for a relational database.
As we understood the OpenRuleBench,
the developers of each system were ask to define the optimal configuration
of their system for each benchmark.
However,
we did find some non-optimal configurations.

In order to give some confidence in our benchmark results,
we tried several different settings
for the same program and benchmark,
and store the results in our benchmark database.
In the main comparison,
we use only the *best* settings,
but in the database it can be seen what we tried and what were the results.
Sometimes,
it is not clear,
what are the best settings,
because most benchmarks are run on different input data files,
and sometimes the best settings are different for different data files.
However,
for the main comparison of the systems,
we require that for each benchmark
there is only one setting manually selected as best.
But our database contains all measurements,
and one can check the alternatives.

Of course,
we would be grateful for settings recommended by the system developers
or experts for one of the systems we tested in our benchmarks.

Another design goal of our database was
to be able to generate results in different formats
(e.g. HTML, LaTeX, csv).
In particular,
we want to be able to generate a web site
with the benchmark results more or less automatically
(so that there are no errors in copying the data,
and updates are simple).


Tables of the Relational Database
=================================


`MACHINE`: List of Machines on which Benchmarks are Executed
------------------------------------------------------------


This table contains some data about the machines on which benchmarks were run.
One machine is marked as the stardard machine
by the value 'S' in column STATUS.
However,
the database permits to store and evaluate
benchmark measurements for different machines
with different architectures and operating systems.

* MACHINE
    + Short name (ID) of the machine, e.g. 'dbs1'.
    + VARCHAR(20)  NOT NULL, PRIMARY KEY
* SEQ_NO
    + Sequential number for sorting a list of machines,
      starting at 1.
    + NUMERIC(2)   NOT NULL,
      UNIQUE,
      CHECK(SEQ_NO > 0),
* STATUS
    + The standard machine is marked by 'S' in this column,
      all other machines contain a space in this column.
      Many views for comparing benchmark results look only
      at measurements for the standard machine.
      Only one row of the table should have STATUS='S'.
      In future, there many be more STATUS values,
      e.g. for comparing the standard machine to an alternative machine.
    + CHAR(1)      NOT NULL,
      CHECK(STATUS IN ('S', ' '))
* MODEL
    + The model name of the machine.
      E.g. the main machine that we use for the benchmarks is a
      'HP Blade Server, HPE BL460c Gen9'.
    + VARCHAR(60)  NULL
* YEAR
    + The year when the machine was bought.
      E.g. out main test machine was bought in October 2016,
      so this column contains 2016.
    + NUMERIC(4)   NULL,
      CHECK(YEAR >= 2000)
* MONTH
    + The month when the machine was bought.
      For our machine, bought in October 2016, this contains 10.
    + NUMERIC(2)   NULL
      CHECK(MONTH >= 1 AND MONTH <= 12)
* PRICE
    + The price of the machine in euros.
      E.g. the price of our main test machine was 3319 euros.
    + NUMERIC(7)   NULL
      CHECK(PRICE >= 0)
* NUM_CPUS
    + The number of CPUs in the machine.
      E.g. the main machine we use for the benchmarks contains two CPUs.
    + NUMERIC(4)   NOT NULL,
      CHECK(NUM_CPUS > 0)
* CPU
    + The name of the CPU,
      e.g. 'Intel(R) Xeon(R) CPU E5-2630 v4 @ 2.20GHz'.
    + VARCHAR(60)  NOT NULL
* CPU_CORES
    + The number of cores per CPU in the machine.
      E.g. the Intel Xeon E5-2630 has 10 cores.
    + NUMERIC(4)   NULL,
      CHECK(CPU_CORES > 0)
* CPU_THREADS
    + Because of hyperthreading,
      the number of hardware-supported threads might be larger than the
      number of cores.
      This column contains the total number of hardware threads per CPU.
      E.g. the Intel Xeon E5-2630 supports 20 concurrent threads.
    + NUMERIC(4)   NULL,
      CHECK(CPU_THREADS > 0)
* GHZ
    + The clock frequency of the CPU in GHz,
      e.g. 2.2 in case of the Intel Xeon E5-2630.
    + NUMERIC(5,2) NULL
* ARCH
    + The architecture of the CPU,
      e.g. 'x86_64'.
    + VARCHAR(20)  NULL
* RAM
    + The main memory size (RAM) in Gigabytes.
      Our main test machine has 64 GB.
    + NUMERIC(8,3) NOT NULL,
      CHECK(RAM > 0)
* NUM_DISKS
    + The number of disks of the machine.
    + NUMERIC(4)   NULL,
      CHECK(NUM_DISKS > 0)
* DISK
    + The type of disk of the machine.
    + VARCHAR(40)  NULL
* SSD
    + If the disk that used for reading data of the benchmarks
      is solid state disk,
      then this column is 'Y'.
      If it is a standard magnetic disk,
      the value of this column is 'N'.
    + CHAR(1)      NULL,
      CHECK(SSD IN ('Y','N'))
* OS
    + The operating system running on the machine,
      e.g. 'Debian GNU/Linux 8.10 (jessie)'.
    + VARCHAR(40)  NULL
* OS_CORE
    + In case of a Linux operting system,
      the version of the core,
      e.g.  'Linux 3.16.0-5-amd64 x86_64'.
    + VARCHAR(40)  NULL
    + Note that due to updates,
      this might not be current.
      The above value was current when we started the benchmarks in May 2018.
      It gives some impression of the machine and its operating system.
      We are planning to create a table MACHINE_HISTORY
      that contains data of a machine that might change.


`INPUT_TYPE`: List of Input Data (Problem Instance) Types
---------------------------------------------------------


We have many input data files for the transitive closure benchmark,
but since they contain general graph data,
they can be used for different benchmarks.
The entries in this table are used to classify data files.
For instance,
there are also data files for the DBLP benchmark
and for the Join1 benchmark,
and so on.

* ITYPE
    + A short ID for a class of data files,
      e.g. 'tc'.
    + VARCHAR(8) NOT NULL, PRIMARY KEY
* SEQ_NO
    + A number used for sorting input types.
    + NUMERIC(2)  NOT NULL
      UNIQUE
      CHECK(SEQ_NO >= 0)
* HEADLINE
    + A headline for the list of files of this type,
      e.g. 'Input Files for the TC Benchmarks (Directed Graphs)'.
    + VARCHAR(80) NOT NULL
* DESCRIPTION
    + A description of this class of data files,
      e.g. 'Directed graph with edges stored in predicate par'.
    + VARCHAR(80) NOT NULL

The currently defined input types are:

ITYPE|SEQ_NO|DESCRIPTION                                                     |
-----|-----:|----------------------------------------------------------------|
tc   |     1|Directed graph with edges stored in predicate par               |
j1   |     2|Directed graph with edges in five relations: c2, c3, c4, d1, d2 |
dblp |     3|DBLP data in an EAV table att                                   |
wine |     4|Wine ontology facts                                             |


`BENCHMARK`: List of Benchmarks
-------------------------------


This is the list of all benchmarks,
e.g. 'tcff' is a benchmark identifier in column BENCH.
It corresponds to the computation of the transitive closure
with query 'tc(X,Y)',
i.e. with both arguments *free*.
Another benchmark, 'tcbf',
is the computation of transitive closure with query 'tc(1,X)',
i.e. the first argument *bound* (to a given value),
and the second argument *free*.

* BENCH
    + This is a short ID of the benchmark,
      e.g. 'tcff'.
    + VARCHAR(20)  NOT NULL,
      PRIMARY KEY
* SEQ_NO
    + The number in this column is used for ordering benchmarks,
      when a web page with all benchmark results is generated.
    + NUMERIC(2)   NOT NULL,
      CHECK(SEQ_NO >= 0),
      UNIQUE
* ITYPE
    + This is the class of data files
      that can be used as input for the benchmark.
    + VARCHAR(8)  NULL,
      FOREIGN KEY (ITYPE) REFERENCES INPUT_TYPE
* HEADLINE
    + This is a headline to be printed in a web page
      with results for this benchmark.
    + VARCHAR(80)  NOT NULL

The currently defined benchmarks are:

BENCH |SEQ_NO|ITYPE|HEADLINE                                                  |
------|-----:|-----|----------------------------------------------------------|
dblp  |   1|dblp|DBLP Benchmark                                               |
j1axy |   2|j1  |The Join1 Benchmark, Query a(X,Y)                            |
j1axyd|   3|j1  |The Join1 Benchmark (without duplicate elimination)          |
tcff  |   4|tc  |Transitive Closure Benchmark without Query Bindings          |
tcbf  |   5|tc  |Transitive Closure Benchmark with the First Argument Bound   |
tcbfs |   6|tc  |Transitive Closure Benchmark, First Arg. Bound, with SLDMagic|
tcfb  |   7|tc  |Transitive Closure Benchmark with the Second Argument Bound  |
wine  |   8|wine|Wine Ontology Benchmark                                      |
loadd |   9|dblp|Test of Loader for DBLP Data                                 |
loadt |  10|tc  |Test of Loader for Transitive Closure                        |
set1  |  11|NULL|Test of Set Insertion and Lookup (One Column)                |
set2  |  12|NULL|Test of Set Insertion and Lookup (Two Columns)               |


`SYS_TYPES`: System Types
-------------------------

This table defines broad classes of systems (programs to be tested),
e.g. Prolog systems and Relational Databases.
We use it for generating subheadlines in the list of all tested programs.

* STYPE
    + A short ID of the class of programs,
      e.g. 'Prolog'.
    + VARCHAR(10)  NOT NULL,
      PRIMARY KEY
* SEQ_NO
    + A number used to define the output sequence.
    + NUMERIC(2)   NOT NULL,
      UNIQUE,
      CHECK(SEQ_NO >= 0)
* HEADLINE
    + The subheadline for this program class in the list of programs.
    + VARCHAR(80)  NOT NULL,

The current entries are:

STYPE   |SEQ_NO| HEADLINE                       |
--------|-----:|--------------------------------|
Prolog  |    1 | Prolog Systems                 |
Datalog |    2 | Datalog Systems                |
RDB     |    3 | Relational Database Systems    |
Graph   |    4 | Graph Databases and RDF Stores |


`TEST_SYS`: List of Systems to be Tested/Compared
-------------------------------------------------

* SYS
    + A short name (ID) of the program, e.g. 'psql'.
    + VARCHAR(20)  NOT NULL
      PRIMARY KEY
* SEQ_NO
    + A number used for sorting the list of programs.
    + NUMERIC(2)   NOT NULL
      CHECK(SEQ_NO >= 0)
      UNIQUE
* FULL_NAME
    + The full name of the program,
      to be used in output.
      E.g. 'PostgreSQL'.
    + VARCHAR(40)  NOT NULL
* STYPE
    + Type of the program for structured lists of all tested programs.
      E.g. 'RDB'.
      This is a foreign key referencing SYS_TYPE.
    + VARCHAR(10)  NOT NULL
      REFERENCES SYS_TYPE
* URL
    + A web address where one can download the program
      or get more information about it.
    + VARCHAR(80)  NULL
* REMARK
    + An optional remark (not yet used).
    + VARCHAR(800) NULL

The current list of programs is:

|SYS|SEQ_NO|FULL_NAME|STYPE|URL|
|----|-----:|---------|-----|---|
|xsb| 10| XSB| Prolog| http://xsb.sourceforge.net/|
|yap| 11| YAProlog| Prolog| https://www.dcc.fc.up.pt/~vsc/Yap/|
|swipl| 12| SWI-Prolog| Prolog| http://www.swi-prolog.org/|
|mercury| 13| Melbourne Mercury| Prolog| http://www.mercurylang.org/|
|bam| 21| Push (bam)| Datalog| http://www.informatik.uni-halle.de/~brass/push/|
|ydb| 22| Push (ydb)| Datalog| http://www.informatik.uni-halle.de/~brass/push/|
|dlv| 23| DLV| Datalog| http://www.dlvsystem.com/|
|souffle| 24| Souffle| Datalog| http://souffle-lang.org/|
|datascript| 25| Datascript| Datalog| https://github.com/tonsky/datascript|
|hsql| 31| HSQLDB| RDB| http://hsqldb.org/|
|sqlite3| 32| SQLite 3| RDB| https://www.sqlite.org/index.html|
|mariadb| 33| MariaDB| RDB| https://www.mariadb.com|
|psql| 34| PostgreSQL| RDB| https://www.postgresql.org|
|jena| 41| Apache Jena| Graph| https://jena.apache.org/|


`SYS_VER`: Program Versions
----------------------------

This table contains information about program versions.
Note that this table describes a version of the source code.
SYS_INSTALL (below) describes a version of the program binary,
which also depends on the machine, the compiler, compiler settings,
and configuration settings.

* SYS
    + The ID of the program (see table TEST_SYS),
      e.g. 'xsb'.
    + VARCHAR(20)  NOT NULL
      REFERENCES TEST_SYS
* VER
    + A version number (any string),
      e.g. '3.7.'.
      Together with the program ID,
      this forms the primary key of this table.
    + VARCHAR(20)  NOT NULL,
    + PRIMARY KEY (SYS, VER)
* VER_ADDTEXT
    + Some program versions have code names or other additions.
      On output, this should be concatenated
      with the value of the attribute VER
      to get the full name of the program version.
      However, it is not necessary for identifying the program version.
      E.g., for XSB 3.7, this attribute has the value ' (Clan MacGregor)'.
      Note the leading space
      (VER and VER_ADDTEXT are simply concatenated).
    + VARCHAR(80)  NOT NULL,
* VER_DATE
    + The release date of the program version.
      It must be a legal date value of the used database system.
      For many systems, the ISO format 'YYYY-MM-DD' works.
      E.g., XSB 3.7 was released on July 6, 2016: '2016-07-06'.
    + DATE         NULL
* REMARK
    + This optional column can be used to store additional information
      about the program version (e.g., major changes or new features).
    + VARCHAR(800) NULL


|SYS|VER|VER_ADDTEXT|VER_DATE|
|----|---|-----------|--------|
|xsb| 3.7.|  (Clan MacGregor)| 2016-07-06|
|xsb| 3.8.0|  (Three-Buck Chuck)| 2017-10-28|
|yap| 6.2.2|  (x86_64-linux)| 2011-10-21|
|yap| 6.3.3|  (x86_64-linux)| 2013-01-20|
|swipl| 7.7.15|  (x86_64-linux)| 2018-06-01|
|swipl| 7.5.15|  (x86_64-linux)| 2017-09-12|
|swipl| 7.6.4|  (x86_64-linux)| 2018-01-12|
|swipl| 7.7.7|  (x86_64-linux)| 2017-12-21|
|ydb| 0.1| | 2016-11-21|
|bam| 0.1a| | 2018-04-24|
|souffle| 1.0| | NULL|
|datascript| 0.16.6| | 2018-06-06|
|sqlite3| 3.8.7.1| | 2014-10-29|
|sqlite3| 3.22.0| | 2018-01-22|
|sqlite3| 3.24.0| | 2018-06-04|
|sqlite3| 3.25.1| | 2018-09-18|
|sqlite3| 3.25.2| | 2018-09-25|
|psql| 9.4.15| | 2017-11-09|
|mariadb| 10.0.35| | 2018-05-03|
|mariadb| 10.3.9| | 2018-08-15|
|jena| 3.7.0| | 2018-04-05|
|jena| 3.8.0| | 2018-06-25|


`SYS_INSTALL`: Table for Installed Program Versions on Machines
----------------------------------------------------------------

This table lists program versions that are or were installed
on a test machine.
They are identified here by a number INST_NO that is unique together with
the program name and the machine.
These three elements (program, installation ID, and machine)
are also encoded in the file names with the test results.

Whereas SYS_VER describes a version of the source code,
SYS_INSTALL describes a binary program.
For instance, if the same source code is compiled with g++ and with clang,
these are two different entries in SYS_INSTALL, both referring to the
same entry in SYS_VER.
Also different configuration settings may give different binaries,
and therefore different entries in SYS_INSTALL.

I try to make sure that the same version of a program
gets the same INST_NO number even on different machines,
but the database schema does not require this
(and it is impossible when there are really different installations
of the same program version).

Only program installations that are listed here can appear in the test runs.

This table also define a current version (more precisely, installation)
of a program on a machine (by STATUS='C').
Some views show only results for the current version.
It is also possible to define a previous version (STATUS='P') as reference,
if the result of a program update should be evaluated.
Finally, one can mark an alternative version (STATUS='A'),
e.g. compiled with a different compiler or compiler options,
which can also be used for comparison.

* SYS
    + The ID of the program, see table TEST_SYS.
      E.g. 'psql'.
    + VARCHAR(20)  NOT NULL
* VER
    + VARCHAR(20)  NOT NULL
    + FOREIGN KEY (SYS, VER) REFERENCES SYS_VER,
* MACHINE
    + The ID of the machine on which the program is installed
      (see table MACHINE).
      E.g. 'dbs1'.
    + VARCHAR(20)  NOT NULL
      REFERENCES MACHINE
* INST_NO
    + A number that uniquely identifies the program installation
      together with the program ID (column SYS) and the machine (MACHINE).
      Since the version VER can be an arbitrary string,
      it is useful to have a single small number instead.
      In addition, this is necessary,
      because there might be different installations of the same program
      version (with different configuration settings or compilers).
    + NUMERIC(4)   NOT NULL
    + PRIMARY KEY (SYS, INST_NO, MACHINE),
* COMPILER
    + Sometimes we want to check whether different compilers (e.g. g++ or clang)
      or different compiler options (e.g. -O2 vs. -O3) produce faster code.
      This column can be used to specify the compiler.
      However,
      one can leave it blank if one is not interested in this information,
      or a standard compiler defined in the installation script was used.
    + VARCHAR(80)  NOT NULL
* STATUS
    + This column is used to mark a *current version* of the program
      with the value 'C'.
      Many views for selecting benchmark results look only at the
      current version on the standard machine.
      There are other options defined below.
      They could be used in views,
      e.g. for comparing the performance of the current version with the
      previous version.
      Only one installation of a program on a machine should be marked
      as current version.
    + CHAR(1)      NOT NULL
      CHECK(STATUS IN ('C','P','A',' '))
    + The meaning of the status values are:
         - 'C': Current version
         - 'P': Previous version
         - 'A': Alternate version
         - ' ': Any other version
* REMARK
    + VARCHAR(800) NULL

|SYS|VER|MACHINE|INST_NO|COMPILER|STATUS|
|----|---|-------|-------|--------|------|
|xsb| 3.7.| dbs1| 1| | C|
|xsb| 3.8.0| dbs1| 2| | C|
|xsb| 3.8.0| bach| 2| | C|
|yap| 6.2.2| dbs1| 1| | C|
|yap| 6.3.3| bach| 2| | C|
|swipl| 7.7.15| dbs1| 1| | C|
|swipl| 7.5.15| dbs1| 2| | P|
|swipl| 7.6.4| dbs1| 3| C|
|swipl| 7.7.7| dbs1| 4| | C|
|ydb| 0.1| bach| 1| gcc version 4.8.5 (-O2)| C|
|ydb| 0.1| dbs1| 1| gcc version 4.9.2 (-O2)| C|
|bam| 0.1a| bach| 1| gcc version 4.8.5 (-O2)| C|
|bam| 0.1a| dbs1| 1| gcc version 4.9.2 (-O2)| C|
|souffle| 1.0| dbs1| 1| | C|
|datascript| 0.16.6| dbs1| 1| node 10.5.0| C|
|sqlite3| 3.8.7.1| dbs1| 1| | P|
|sqlite3| 3.22.0| dbs1| 2| | P|
|sqlite3| 3.24.0| dbs1| 3| | P|
|sqlite3| 3.25.1| dbs1| 4| | C|
|sqlite3| 3.25.2| dbs1| 5| | A|
|psql| 9.4.15| dbs1| 1| | C|
|mariadb| 10.0.35| dbs1| 1| | P|
|mariadb| 10.3.9| dbs1| 2| | C|
|jena| 3.7.0| dbs1| 1| openJDK 8| P|
|jena| 3.8.0| dbs1| 2| openJDK 8| C|


`BENCH_IMPL`: Implementation Variants of Benchmarks
---------------------------------------------------

In general,
there are different option settings (e.g. index selections)
that should be tested for the same program
to execute a given benchmark as fast as possible.
We call such a combination of benchmark, program and option settings
a *benchmark implementation*.
STATUS='B' marks the best implementation
that is used in some views to compare different programs.
This is selected manually,
but there are also views that help to find the best implementation.
For instance,
we tested the InnoDB und the Memory storage engine for MariaDB.
It depended on the input data,
which was faster.
We selected InnoDB in this table for the main comparison between the systems,
but the benchmark database contains results for both.

* SYS
    + ID of the program that executes the benchmark,
      see table TEST_SYS.
      E.g. 'psql'.
    + VARCHAR(20)  NOT NULL
      REFERENCES TEST_SYS
* BENCH
    + ID of the benchmark,
      see table BENCHMARK.
      E.g. 'tcff'.
    + VARCHAR(20)  NOT NULL
      REFERENCES BENCHMARK
* IMPL
    + This value is used to distinguish between several implementations
      of the same benchmark with the same program.
      For instance,
      the implemenations on a relational database might differ in
      the indexes that have been created.
      If there is only one implementation,
      it is simplest to choose a single space.
      The program for loading benchmark runtimes from tsv-files
      will map a missing implementation to a space.
    + VARCHAR(20)  NOT NULL
    + PRIMARY KEY (SYS, BENCH, IMPL),
* STATUS
    + CHAR(1)      NOT NULL
      CHECK(STATUS IN ('B','A',' '))
    + Meaning of the STATUS values:
         - STATUS='B' marks the implementation variant (option settings)
           that is considered best and is used in comparisons between systems.
           One can use the views BEST_IMPL_REAL and BEST_IMPL_CPU to check
           whether these implementation variants are really best.
           Sometimes the best implementation depends on the specific data file,
           but it would somehow be unfair to adapt the options to the data file.
         - STATUS='A' marks an interesting alternative variants
           that appear in some result tables.
           There should be only one *best* version,
           but there can be several alterative versions.
         - STATUS=' ' means that the variant was tested
           in order to prove that the best version is really good,
           but it will not be listed in summary tables.
* DESCRIPTION
    + In this attribute,
      one would typically explain which indexes have been chosen,
      or other information on the specific benchmark implementation.
    + VARCHAR(80)  NULL

The current list of benchmark implementations is:

|SYS|BENCH|IMPL|STATUS|DESCRIPTION|
|----|-----|----|------|-----------|
|bam|j1axy|1||Join1, a(X,Y) [V1: RelAttr, -goto, fullDupCh]|
|bam|j1axy|2||Join1, a(X,Y) [V2: no goto, bitmap-dupcheck]|
|bam|j1axy|3|B|Join1, a(X,Y) [V3: goto, bitmap-dupC, cursor]|
|bam|j1axy|4|A|Join1, a(X,Y) [V4: functions, bitmap-dupC]|
|bam|j1axy|5||Join1, a(X,Y) [V5: fun without args, bitmap]|
|bam|j1axy|6|A|Join1, a(X,Y) [V6: Abstract Machine, bitmap]|
|bam|tcff|1||tc(X,Y) [V1: RelLocalVar]|
|bam|tcff|2|A|tc(X,Y) [V2: Cursor Interface, switch]|
|bam|tcff|3|A|tc(X,Y) [V3: Seminaive Evaluation]|
|bam|tcff|4|A|tc(X,Y) [V4: Functions]|
|bam|tcff|5|B|tc(X,Y) [V5: Abstract Machine]|
|bam|loadt|1|A|transitive closure data file (with handcrafted loader)|
|bam|loadt|2|B|transitive closure data file (with general loader)|
|bam|set1|1|B|Set with 1 column, extendible Hash table|
|bam|set2|1|B|Set with 2 columns, extendible Hash table|
|xsb|dblp|dyn|A|load_dyn, no index|
|xsb|dblp|dync|A|load_dync, no index|
|xsb|dblp|dynin|A|load_dyn, index on col 1+2|
|xsb|dblp|dyncin|B|load_dync, index on col 1+2|
|xsb|dblp|dyntr|A|load_dyn, trie index|
|xsb|dblp|dynctr|A|load_dync, trie index|
|xsb|j1axyd|none|B|Join1 with duplicates, a(X,Y) [V0: no tabling]|
|xsb|j1axy|a||Join1, a(X,Y) [V1: only a tabled]|
|xsb|j1axy|all||Join1, a(X,Y) [V2: all derived preds tabled (subsumptive)]|
|xsb|j1axy|allv|B|Join1, a(X,Y) [V4: all derived preds tabled, variant tabled]|
|xsb|j1axy|alli||Join1, a(X,Y) [V4: all derived preds tabled, with indexes]|
|xsb|j1axy|allt|A|Join1, a(X,Y) [V5: all derived preds tabled, with trie index]|
|xsb|tcff|dyn||tc(X,Y) [V1: load_dyn, trie-index for par]|
|xsb|tcff||B|tc(X,Y) [V2: load_dync, trie-index for par]|
|xsb|tcbf|dyn||tc(1,X) [V1: load_dyn, trie-index for par]|
|xsb|tcbf||B|tc(1,X) [V2: load_dync, trie-index for par]|
|xsb|tcbfs|slddc|B|tc(1,X) [V3: SLDMagic, load_dync, trie-index]|
|xsb|tcfb||B|tc(X,1) [unknown implementation]|
|xsb|wine|is|B|wine ontology [V1: load_dync, subsumptive tabling, std. indexes]|
|yap|dblp|1|B|DBLP Benchmark|
|yap|j1axyd|none|B|Join1 with duplicates, a(X,Y) [V0: no tabling]|
|yap|j1axy|a||Join1, a(X,Y) [V1: only a tabled]|
|yap|j1axy|all|B|Join1, a(X,Y) [V2: all derived preds tabled]|
|yap|tcff||B|tc(X,Y) [V1: tc tabled]|
|yap|tcbf||B|tc(1,X) [V1: tc tabled]|
|yap|tcbfs|sldm|B|tc(1,X) [V2: with SLDMagic, all preds tabled]|
|yap|tcfb||B|tc(X,1) [V1: tc tabled]|
|yap|wine|tab|B|wine ontology [V1: table 225 derived preds, reorder body lits|
|swipl|dblp||B|DBLP Benchmark|
|swipl|j1axyd||B|Join1 with duplicates, a(X,Y)|
|swipl|j1axy||B|Join1, a(X,Y)|
|swipl|tcff||B|tc(X,Y) [tc tabled]|
|swipl|tcbf||B|tc(1,X) [tc tabled]|
|swipl|tcbfs|sldm|B|tc(1,X) [with SLDMagic, all preds tabled]|
|swipl|tcfb||B|tc(X,1) [tc tabled]|
|swipl|wine||B|wine ontology|
|sqlite3|tcff||B|transitive closure [ff] for SQLite3|
|sqlite3|tcbf||B|transitive closure [bf] for SQLite3|
|sqlite3|tcfb||B|transitive closure [fb] for SQLite3|
|psql|tcff||B|transitive closure [ff] for PostgreSQL|
|psql|tcbf||B|transitive closure [bf] for PostgreSQL|
|psql|tcfb||B|transitive closure [fb] for PostgreSQL|
|mariadb|tcff|m|A|transitive closure for MariaDB, storage manager MEMORY|
|mariadb|tcff|i|B|transitive closure for MariaDB, storage manager InnoDB|
|mariadb|tcbf||B|transitive closure for MariaDB, unknown storage manager|
|mariadb|tcfb||B|transitive closure for MariaDB, unknown storage manager|
|jena|tcff||B|transitive closure [ff] with Property Paths for Apache Jena|
|jena|tcbf||B|transitive closure [bf] with Property Paths for Apache Jena|
|jena|tcfb||B|transitive closure [fb] with Property Paths for Apache Jena|
|datascript|tcff||B|transitive closure for DataScript|
|souffle|tcff||B|tc(X,Y) [V1: compiled, -j1]|


`INPUT_FILE`: Information about Input Data Files
------------------------------------------------

This a list of data files
that can be used as input for a benchmark.
Each file has a type ITYPE,
and in the table BENCHMARK (see above)
it is listed what kind of input is required for a benchmark.
For instance,
we have many data files that define directed graphs
with a predicate *par*
(parent, OpenRuleBench used this name for the transitive closure benchmarks).

Different programs need different file formats,
therefore each file is available in several formats such as
+ Prolog facts (with extension '.P')
+ Tab-Separated-Values (extension '.tsv')
+ Comma-separated Values (extension '.csv')
+ SQL INSERT Statements (extension '.sql')
+ RDF in Turtle Syntax (extension '.ttl')
+ JSON (extension '.json')
+ graphviz (extension '.dot')

The columns of the table are:

* FILE_ID
    + A short name (ID) that uniquely identifies the file.
      E.g. 'k1000' for the complete graph with 1000 nodes.
    + VARCHAR(10)  NOT NULL
      PRIMARY KEY
* SEQ_NO
    + A number that defines a sort order for the files
      (or benchmark results for files).
    + NUMERIC(3)   NOT NULL
      UNIQUE
      CHECK(SEQ_NO >= 0)
* ITYPE
    + The type of data in the file,
      see table INPUT_TYPE above.
      E.g. 'tc'.
    + VARCHAR(8)  NOT NULL
      REFERENCES INPUT_TYPE
* LINES
    + The number of lines in the file.
      Formally, the Prolog facts version is chosen to count the lines.
      But most other file formats will have the same number of lines.
      This is also the number of facts in the file
      (rows in the corresponding tables).
    + NUMERIC(12)  NOT NULL
      CHECK(LINES >= 0)
* BYTES
    + This is the file size in bytes of the Prolog facts version.
      It should give an impression of the approximate file size
      in the other versions, too.
    + NUMERIC(12)  NOT NULL
      CHECK(BYTES >= 0)
* FILENAME
    + This is the file name of the data file without the extension.
    + VARCHAR(40)  NOT NULL

Note that the data of this table are contained in the file 'input_files.sql'.
The current contents of this table is:

|FILE_ID|SEQ_NO|ITYPE|LINES|BYTES|FILENAME|
|-------|-----:|-----|----:|----:|--------|
|k465|101|tc|216225|3141075|tc_complete_465|
|k500|102|tc|250000|3640000|tc_complete_500|
|k585|103|tc|342225|5004675|tc_complete_585|
|k670|104|tc|448900|6586100|tc_complete_670|
|k735|105|tc|540225|7941675|tc_complete_735|
|k795|106|tc|632025|9305475|tc_complete_795|
|k845|107|tc|714025|10524475|tc_complete_845|
|k890|108|tc|792100|11685700|tc_complete_890|
|k930|109|tc|864900|12768900|tc_complete_930|
|k965|110|tc|931225|13756075|tc_complete_965|
|k1k|111|tc|1000000|14780000|tc_complete_1000|
|k2k|112|tc|4000000|63560000|tc_complete_2000|
|t500|121|tc|124750|1816360|tc_total_500|
|t585|122|tc|170820|2498060|tc_total_585|
|t735|123|tc|269745|3965435|tc_total_735|
|t845|124|tc|356590|5256010|tc_total_845|
|t930|125|tc|431985|6377585|tc_total_930|
|t1k|126|tc|499500|7382610|tc_total_1000|
|t1065|127|tc|566580|8450820|tc_total_1065|
|t1120|128|tc|626640|9410790|tc_total_1120|
|t1170|129|tc|683865|10328115|tc_total_1170|
|t1215|130|tc|737505|11190045|tc_total_1215|
|t1260|131|tc|793170|12086400|tc_total_1260|
|t2k|132|tc|1999000|31764110|tc_total_2000|
|c1k|141|tc|1000|14780|tc_cycle_1000|
|c2k|142|tc|2000|31780|tc_cycle_2000|
|c3k|143|tc|3000|48780|tc_cycle_3000|
|c4k|144|tc|4000|65780|tc_cycle_4000|
|c10k|145|tc|10000|167780|tc_cycle_10000|
|c14k1|146|tc|14100|245680|tc_cycle_14100|
|c17k3|147|tc|17300|306480|tc_cycle_17300|
|c20k|148|tc|20000|357780|tc_cycle_20000|
|c22k4|149|tc|22400|403380|tc_cycle_22400|
|c24k5|150|tc|24500|443280|tc_cycle_24500|
|c26k5|151|tc|26500|481280|tc_cycle_26500|
|c28k3|152|tc|28300|515480|tc_cycle_28300|
|c30k|153|tc|30000|547780|tc_cycle_30000|
|c31k6|154|tc|31600|578180|tc_cycle_31600|
|s1k_1|161|tc|2000|29560|tc_shorted_cycle_1000_1|
|s1k_2|162|tc|3000|44340|tc_shorted_cycle_1000_2|
|s1k_3|163|tc|4000|59120|tc_shorted_cycle_1000_3|
|s1k_4|164|tc|5000|73900|tc_shorted_cycle_1000_4|
|s1k_5|165|tc|6000|88680|tc_shorted_cycle_1000_5|
|s1k2_1|166|tc|2400|36360|tc_shorted_cycle_1200_1|
|s1k2_2|167|tc|3600|54540|tc_shorted_cycle_1200_2|
|s1k2_4|168|tc|6000|90900|tc_shorted_cycle_1200_4|
|s1k2_5|169|tc|3357|50265|tc_shorted_cycle_1200_5|
|s2k_1|170|tc|4000|63560|tc_shorted_cycle_2000_1|
|s2k_2|171|tc|6000|95340|tc_shorted_cycle_2000_2|
|s2k_3|172|tc|8000|127120|tc_shorted_cycle_2000_3|
|s2k_4|173|tc|10000|158900|tc_shorted_cycle_2000_4|
|s2k_5|174|tc|12000|190680|tc_shorted_cycle_2000_5|
|s2k4_1|175|tc|4800|77160|tc_shorted_cycle_2400_1|
|s2k4_2|176|tc|7200|115740|tc_shorted_cycle_2400_2|
|s2k4_4|177|tc|12000|192900|tc_shorted_cycle_2400_4|
|s2k4_5|178|tc|14400|231480|tc_shorted_cycle_2400_5|
|s4k8_1|179|tc|9600|158760|tc_shorted_cycle_4800_1|
|s4k8_2|180|tc|14400|238140|tc_shorted_cycle_4800_2|
|s4k8_4|181|tc|24000|396900|tc_shorted_cycle_4800_4|
|s4k8_5|182|tc|28800|476280|tc_shorted_cycle_4800_5|
|s6k3_1|183|tc|12600|209760|tc_shorted_cycle_6300_1|
|s6k3_2|184|tc|18900|314640|tc_shorted_cycle_6300_2|
|s6k3_4|185|tc|31500|524400|tc_shorted_cycle_6300_4|
|s6k3_5|186|tc|37800|629280|tc_shorted_cycle_6300_5|
|s7k2_1|187|tc|14400|240360|tc_shorted_cycle_7200_1|
|s7k2_2|188|tc|21600|360540|tc_shorted_cycle_7200_2|
|s7k2_4|189|tc|36000|600900|tc_shorted_cycle_7200_4|
|s7k2_5|190|tc|43200|721080|tc_shorted_cycle_7200_5|
|s9k_1|191|tc|18000|301560|tc_shorted_cycle_9000_1|
|s9k_2|192|tc|27000|452340|tc_shorted_cycle_9000_2|
|s9k_4|193|tc|45000|753900|tc_shorted_cycle_9000_4|
|s9k_5|194|tc|54000|904680|tc_shorted_cycle_9000_5|
|s12k_1|195|tc|24000|411560|tc_shorted_cycle_12000_1|
|s12k_2|196|tc|36000|617340|tc_shorted_cycle_12000_2|
|s12k_4|197|tc|60000|1028900|tc_shorted_cycle_12000_4|
|s12k_5|198|tc|72000|1234680|tc_shorted_cycle_12000_5|
|s13k5_1|199|tc|27000|468560|tc_shorted_cycle_13500_1|
|s13k5_2|200|tc|40500|702840|tc_shorted_cycle_13500_2|
|s13k5_4|201|tc|67500|1171400|tc_shorted_cycle_13500_4|
|s13k5_5|202|tc|81000|1405680|tc_shorted_cycle_13500_5|
|s15k9_1|203|tc|31800|559760|tc_shorted_cycle_15900_1|
|s15k9_2|204|tc|47700|839640|tc_shorted_cycle_15900_2|
|s15k9_4|205|tc|79500|1399400|tc_shorted_cycle_15900_4|
|s15k9_5|206|tc|95400|1679280|tc_shorted_cycle_15900_5|
|s18k6_1|207|tc|37200|662360|tc_shorted_cycle_18600_1|
|s18k6_2|208|tc|55800|993540|tc_shorted_cycle_18600_2|
|s18k6_4|209|tc|93000|1655900|tc_shorted_cycle_18600_4|
|s18k6_5|210|tc|111600|1987080|tc_shorted_cycle_18600_5|
|p1k|221|tc|999|14767|tc_path_1000|
|p2k|222|tc|1999|31766|tc_path_2000|
|p3k|223|tc|2999|48766|tc_path_3000|
|p4k|224|tc|3999|65766|tc_path_4000|
|p14k1|225|tc|14099|245665|tc_path_14100|
|p20k|226|tc|19999|357765|tc_path_20000|
|p24k5|227|tc|24499|443265|tc_path_24500|
|p28k3|228|tc|28299|515465|tc_path_28300|
|p31k6|229|tc|31599|578165|tc_path_31600|
|p34k6|230|tc|34599|635165|tc_path_34600|
|p37k4|231|tc|37399|688365|tc_path_37400|
|p40k|232|tc|39999|737765|tc_path_40000|
|p42k4|233|tc|42399|783365|tc_path_42400|
|p44k7|234|tc|44699|827065|tc_path_44700|
|m4_2ki|241|tc|6144|103341|tc_multipath_4_2048|
|m16_512|242|tc|7680|128966|tc_multipath_16_512|
|m64_128|243|tc|8064|135110|tc_multipath_64_128|
|m256_32|244|tc|8160|136579|tc_multipath_256_32|
|m1ki_8|245|tc|8184|136938|tc_multipath_1024_8|
|m4ki_2|246|tc|8190|137022|tc_multipath_4096_2|
|b17|251|tc|131070|2488088|tc_bintree_17|
|b18|252|tc|262142|5171672|tc_bintree_18|
|b19|253|tc|524286|10676696|tc_bintree_19|
|b20|254|tc|1048574|21735320|tc_bintree_20|
|b21|255|tc|2097150|44901144|tc_bintree_21|
|b22|256|tc|4194302|93135640|tc_bintree_22|
|b23|257|tc|8388606|189604632|tc_bintree_23|
|b24|258|tc|16777214|389319832|tc_bintree_24|
|b25|259|tc|33554430|805527448|tc_bintree_25|
|b26|260|tc|67108862|1644388248|tc_bintree_26|
|y500_4k|271|tc|4499|74378|tc_y_500_4000|
|y500_8k|272|tc|8499|142378|tc_y_500_8000|
|y1k_4k|273|tc|4999|83876|tc_y_1000_4000|
|y1k_8k|274|tc|8999|151876|tc_y_1000_8000|
|y1k_15k|275|tc|15999|282877|tc_y_1000_15000|
|y1k_25k|276|tc|25999|472877|tc_y_1000_25000|
|y1k_35k|277|tc|35999|662877|tc_y_1000_35000|
|y5k_15k|278|tc|19999|358877|tc_y_5000_15000|
|y5k_25k|279|tc|29999|548877|tc_y_5000_25000|
|y5k_35k|280|tc|39999|738877|tc_y_5000_35000|
|y10k_15k|281|tc|24999|463875|tc_y_10000_15000|
|y10k_25k|282|tc|34999|653875|tc_y_10000_25000|
|y10k_35k|283|tc|44999|843875|tc_y_10000_35000|
|u1k_50k|301|tc|50000|739011|tc_uniform_1000_50000|
|u1k_125k|302|tc|125000|1847511|tc_uniform_1000_125000|
|u1k_250k|303|tc|250000|3695016|tc_uniform_1000_250000|
|u2k_200k|304|tc|200000|3178014|tc_uniform_2000_200000|
|u2k_250k|305|tc|250000|3972534|tc_uniform_2000_250000|
|u2k_500k|306|tc|500000|7944982|tc_uniform_2000_500000|
|u2k_1m|307|tc|1000000|15890006|tc_uniform_2000_1000000|
|a1k_50k|311|tc|50000|738966|tc_acyc_1000_50000|
|a1k_125k|312|tc|125000|1847531|tc_acyc_1000_125000|
|a1k_250k|313|tc|250000|3695001|tc_acyc_1000_250000|
|a2k_200k|314|tc|200000|3177896|tc_acyc_2000_200000|
|a2k_500k|315|tc|500000|7944738|tc_acyc_2000_500000|
|a2k_1m|316|tc|1000000|15890152|tc_acyc_2000_1000000|
|f1|401|tc|50000|689318|tc_d1000_parsize50000_xsb_nocyc|
|f2|402|tc|50000|689318|tc_d1000_parsize50000_xsb_cyc|
|f3|403|tc|250000|3446506|tc_d1000_parsize250000_xsb_nocyc|
|f4|404|tc|250000|3446506|tc_d1000_parsize250000_xsb_cyc|
|f5|405|tc|500000|6893031|tc_d1000_parsize500000_xsb_nocyc|
|f6|406|tc|500000|6893031|tc_d1000_parsize500000_xsb_cyc|
|f7|407|tc|500000|7446477|tc_d2000_parsize500000_xsb_nocyc|
|f8|408|tc|500000|7446477|tc_d2000_parsize500000_xsb_cyc|
|f9|409|tc|1000000|14892982|tc_d2000_parsize1000000_xsb_nocyc|
|f10|410|tc|1000000|14892982|tc_d2000_parsize1000000_xsb_cyc|
|10k|501|j1|50000|639300|d1000_relsize10000_xsb_cyc|
|50k|502|j1|250000|3196504|d1000_relsize50000_xsb_cyc|
|250k|503|j1|1250000|15982480|d1000_relsize250000_xsb_cyc|
|std|601|dblp|2437867|128444409|dblp|
|big|602|dblp|50221773|3082908204|dblp_big|
|wdat|701|wine|654|26386|wine_data|


`INPUT_SET`: Collections of Input Data Files Used for Publications
------------------------------------------------------------------

Sometimes,
we have new ideas for interesting input data files.
In order to make the results of previous publications easily reproducible,
we want to store which files were used for which publications.
Actually,
this mechanism can be used to attach arbitrary tags to input data files,
so that one can use any collection of data files for evaluations.

* SET_ID
    + An identifier for the set of input files
      or the tag that one wishes to attach to a file.
    + VARCHAR(10)  NOT NULL
* FILE_ID
    + This identifies the input file, see table INPUT_FILE.
    + VARCHAR(10)  NOT NULL
      REFERENCES INPUT_FILE
    + PRIMARY KEY (SET_ID, FILE_ID)

Note that the data for this table are contained in 'input_sets.sql'.

Currently, only one input set 'oct18' is defined
that contains the 43 graphs used in a paper finished in October 2018.


`GRAPH_TYPES`: For Classifying Graphs (A Subclass of Input Files)
-----------------------------------------------------------------

A subclass of the input data files are graphs.
This table defines types of graphs
that can be used e.g. for subheadlines in the list of data files.

* GTYPE
    + A one-letter ID of the graph type.
      This is also the first letter of the FILE_ID of the graph
      (maybe with different case).
    + CHAR(1)      NOT NULL
      PRIMARY KEY
* SEQ_NO
    + This defines an output order for the graph types.
    + NUMERIC(2)   NOT NULL
      CHECK(SEQ_NO >= 0)
      UNIQUE
* NAME
    + Name of the graph type
      (possibly with short explanation).
    + VARCHAR(80)  NOT NULL

The current list of graph types is:

|GTYPE|SEQ_NO|NAME                                                |
|-----|-----:|----------------------------------------------------|
| K   |    1 | Complete Graph (with loops)                        |
| T   |    2 | Total Order Graph (i.e. maximal acyclic graph)     |
| C   |    3 | Cycle Graph                                        |
| S   |    4 | Cycle Graph with Shortcuts                         |
| P   |    5 | Path                                               |
| M   |    6 | Multi-Path                                         |
| B   |    7 | Binary Tree                                        |
| Y   |    8 | Y-Graph (many nodes pointing to first node of path)|
| U   |    9 | Random Graph with Uniform Node Degrees             |
| A   |   10 | Random Acyclic Graph with Uniform Node Degrees     |


`INPUT_GRAPH`: More Data on Input Data Files That are Directed Graphs
---------------------------------------------------------------------

A subset of the data files describe directed graphs
(the data files with ITYPE='tc').
This table contains additional attributes for those data file
(properties of the graph).
We need some of the attributes for out work on runtime estimation.
We computed these numbers from the data files
also as an additional safety measure to find possible errors
in the graph-generating program.
For instance,
the program used in the OpenRuleBench
did produce duplicate edges
for acyclic graphs.

* FILE_ID
    + This identifies the data file.
      It is at the same time foreign key to INPUT_FILE
      and primary key
      (this is a usual translation of subclasses).
    + VARCHAR(10) NOT NULL
      PRIMARY KEY
      REFERENCES INPUT_FILE
* GTYPE
    + This is the type of the graph.
      E.g. all complete graphs have GTYPE='K'.
      See table GRAPH_TYPE.
    + CHAR(1)     NOT NULL
      REFERENCES GRAPH_TYPE
* PARAM_1
    + For graphs that were generated with our graph generation program,
      this is the value of the first (and often only) parameter.
      E.g. for a complete graph with 1000 nodes,
      it will be 1000.
    + NUMERIC(12) NULL
      CHECK(PARAM_1 >= 0)
* PARAM_2
    + This is the value of the second parameter
      for the graph generation program (if applicable).
      E.g. for multi-paths,
      the first parameter is the length of the single paths,
      and the second parameter is the number of paths.
    + NUMERIC(12) NULL
      CHECK(PARAM_2 >= 0)
* NUM_NODES
    + The number of nodes in the graph.
    + NUMERIC(12) NOT NULL
      CHECK(NUM_NODES >= 0)
* NUM_EDGES
    + The number of edges in the graph.
    + NUMERIC(12) NOT NULL
      CHECK(NUM_EDGES >= 0)
* DUP_EDGES
    + If this is not 0,
      the graph contains duplicate edges,
      i.e. more than one edge between the same two nodes.
      Formally, the value of this attribute
      is the number of pairs (x,y) of nodes,
      such that there are at least two edges between x and y.
    + NUMERIC(12) NOT NULL
      CHECK(DUP_EDGES >= 0)
* LOOPS
    + This is 'Y' if the graph contains loops,
      i.e. edges of the form (x,x).
      It is 'N' otherwise.
    + CHAR(1)     NOT NULL
      CHECK(LOOPS  in ('Y','N'))
* MIN_IN_DEGREE
    + This is the minimal in-degree of the nodes,
      i.e. the minimum of the count of incoming edges
      over all nodes.
    + NUMERIC(12) NOT NULL
      CHECK(MIN_IN_DEGREE >= 0)
* MAX_IN_DEGREE
    + This is the maximal in-degree of the nodes,
      i.e. the minimum of the count of incoming edges
      over all nodes.
    + NUMERIC(12) NOT NULL
      CHECK(MAX_IN_DEGREE >= 0)
* MIN_OUT_DEGREE
    + This is the minimal out-degree of the nodes,
      i.e. the minimum of the count of outgoing edges
      over all nodes.
    + NUMERIC(12) NOT NULL
      CHECK(MIN_OUT_DEGREE >= 0)
* MAX_OUT_DEGREE
    + This is the maximal out-degree of the nodes,
      i.e. the maximum of the count of outgoing edges
      over all nodes.
    + NUMERIC(12) NOT NULL
      CHECK(MAX_OUT_DEGREE >= 0)
* CYCLES
    + This is 'Y' if the graph is cyclic,
      i.e. there is a node x such that there is a non-empty path
      starting at x and ending at x.
      It is 'N' for acyclic graphs.
    + CHAR(1)     NOT NULL
      CHECK(CYCLES IN ('Y','N'))
* TC_SIZE
    + This is the size of the transitive closure of the graph,
      i.e. the number of node pairs (x,y)
      such that y is reachable from x via one or more edges.
    + NUMERIC(12) NOT NULL
      CHECK(TC_SIZE >= 0)
* NUM_ITER
    + This is the number of iterations
      that naive/seminaive evaluation would need to compute
      the transitive closure.
      I.e. it is the length of the longest path
      for which there is no shorter path between the same two nodes.
      In other words,
      it is the maximum of the distance (length of the shortest path)
      of all nodes x and y such that y is reachable from x.
      For connected graphs,
      this is known as the diameter of the graph.
    + NUMERIC(12) NOT NULL
      CHECK(NUM_ITER >= 0)
* COST
    + This is the number of applicable rule instances
      in the standard tail-recursive Prolog/Datalog program
      for computing the transitive closure
      (where 'applicable' means
      that the rule body is true in the minimal Herbrand model).
      One can compute this as the number of edges
      (for the rule tc(X,Y) <- par(X,Y))
      plus the size of the join of the edge relation
      with the transitive closure relation
      (for the rule tc(X,Z) <- par(X,Y), tc(Y,Z)).
    + NUMERIC(12) NOT NULL
      CHECK(COST >= 0)

Note that the data for this table are contained in the file 'input_graphs.sql'.
The current contents of this table is:

|FILE_ID|GTYPE|PARAM_1|PARAM_2|NUM_NODES|NUM_EDGES|DUP_EDGES|LOOPS|MIN_IN_DEGREE|MAX_IN_DEGREE|MIN_OUT_DEGREE|MAX_OUT_DEGREE|CYCLES|TC_SIZE|NUM_ITER|COST|
|-------|-----|------:|------:|--------:|--------:|--------:|:---:|------------:|------------:|-------------:|-------------:|:----:|------:|-------:|---:|
|k500|K|500| |500|250000|0|Y|500|500|500|500|Y|250000|1|125250000|
|k1k|K|1000| |1000|1000000|0|Y|1000|1000|1000|1000|Y|1000000|1|1001000000|
|k2k|K|2000| |2000|4000000|0|Y|2000|2000|2000|2000|Y|4000000|1|8004000000|
|t500|T|500| |500|124750|0|N|0|499|0|499|N|124750|1|62126000|
|t1k|T|1000| |1000|499500|0|N|0|999|0|999|N|499500|1|166666500|
|t2k|T|2000| |2000|1999000|0|N|0|1999|0|1999|N|1999000|1|1333333000|
|c1k|C|1000| |1000|1000|0|N|1|1|1|1|Y|1000000|1000|1001000|
|c2k|C|2000| |2000|2000|0|N|1|1|1|1|Y|4000000|2000|4002000|
|c3k|C|3000| |3000|3000|0|N|1|1|1|1|Y|9000000|3000|9003000|
|c4k|C|4000| |4000|4000|0|N|1|1|1|1|Y|16000000|4000|16004000|
|s1k_1|S|1000|1|1000|2000|0|N|2|2|2|2|Y|1000000|500|2002000|
|s1k_2|S|1000|2|1000|3000|0|N|3|3|3|3|Y|1000000|201|3003000|
|s1k_3|S|1000|3|1000|4000|0|N|4|4|4|4|Y|1000000|250|4004000|
|s1k_4|S|1000|4|1000|5000|0|N|5|5|5|5|Y|1000000|200|5005000|
|s1k_5|S|1000|5|1000|6000|0|N|6|6|6|6|Y|1000000|42|6006000|
|s2k_1|S|2000|1|2000|4000|0|N|2|2|2|2|Y|4000000|1000|8004000|
|s2k_2|S|2000|2|2000|6000|0|N|3|3|3|3|Y|4000000|288|12006000|
|s2k_3|S|2000|3|2000|8000|0|N|4|4|4|4|Y|4000000|500|16008000|
|s2k_4|S|2000|4|2000|10000|0|N|5|5|5|5|Y|4000000|400|20010000|
|s2k_5|S|2000|5|2000|12000|0|N|6|6|6|6|Y|4000000|127|24012000|
|p1k|P|1000| |1000|999|0|N|0|1|0|1|N|499500|999|499500|
|p2k|P|2000| |2000|1999|0|N|0|1|0|1|N|1999000|1999|1999000|
|p3k|P|3000| |3000|2999|0|N|0|1|0|1|N|4498500|2999|4498500|
|p4k|P|4000| |4000|3999|0|N|0|1|0|1|N|7998000|3999|7998000|
|m4_2ki|M|4|2048|8192|6144|0|N|0|1|0|1|N|12288|3|12288|
|m16_512|M|16|512|8192|7680|0|N|0|1|0|1|N|61440|15|61440|
|m64_128|M|64|128|8192|8064|0|N|0|1|0|1|N|258048|63|258048|
|m256_32|M|256|32|8192|8160|0|N|0|1|0|1|N|1044480|255|1044480|
|m1ki_8|M|1024|8|8192|8184|0|N|0|1|0|1|N|4190208|1023|4190208|
|m4ki_2|M|4096|2|8192|8190|0|N|0|1|0|1|N|16773120|4095|16773120|
|b17|B|17| |131071|131070|0|N|0|1|0|2|N|1966082|16|1966082|
|b18|B|18| |262143|262142|0|N|0|1|0|2|N|4194306|17|4194306|
|b19|B|19| |524287|524286|0|N|0|1|0|2|N|8912898|18|8912898|
|y500_4k|Y|500|4000|4500|4499|0|N|0|500|0|1|N|9998000|4000|9998000|
|y500_8k|Y|500|8000|8500|8499|0|N|0|500|0|1|N|35996000|8000|35996000|
|y1k_4k|Y|1000|4000|5000|4999|0|N|0|1000|0|1|N|11998000|4000|11998000|
|y1k_8k|Y|1000|8000|9000|8999|0|N|0|1000|0|1|N|39996000|8000|39996000|
|u1k_50k|U|1000|50000|1000|50000|0|N|46|51|46|51|Y|1000000|3|50050000|
|u1k_125k|U|1000|125000|1000|125000|0|N|122|127|122|127|Y|1000000|2|125125000|
|u1k_250k|U|1000|250000|1000|250000|0|N|248|251|248|251|Y|1000000|2|250250000|
|u2k_200k|U|2000|200000|2000|200000|0|N|98|101|98|101|Y|4000000|3|400200000|
|u2k_500k|U|2000|500000|2000|500000|0|N|248|251|248|251|Y|4000000|2|1000500000|
|u2k_1m|U|2000|1000000|2000|1000000|0|N|499|502|499|502|Y|4000000|2|2001000000|
|a1k_50k|A|1000|50000|1000|50000|0|N|0|102|0|102|N|471621|8|15269270|
|a1k_125k|A|1000|125000|1000|125000|0|N|0|250|0|250|N|492680|5|40853746|
|a1k_250k|A|1000|250000|1000|250000|0|N|0|500|0|501|N|497763|4|82941061|
|a2k_200k|A|2000|200000|2000|200000|0|N|0|201|0|201|N|1944729|8|128420988|
|a2k_500k|A|2000|500000|2000|500000|0|N|0|500|0|500|N|1985321|6|330666855|
|a2k_1m|A|2000|1000000|2000|1000000|0|N|0|1000|0|1000|N|1995698|4|666033613|


`BENCH_RUN`: Results of Benchmark Runs
--------------------------------------

This is finally the main table,
in which the runtime of benchmark runs is stored.

There are three sources for timing information:
Measurements done by the program itself,
external measurements done with `/usr/bin/time`,
and measurements through the `/proc/[procid]/stats` virtual file system.

In order to distinguish between the time needed for loading the data
and for executing the query,
we printed these times from the program that runs the benchmark
(where possible).
For instance,
Prolog systems usually have a timing function.
In addition,
we printed the total runtime as computed by the program itself.
This should normally be simply the sum of load time and execution time.
It differs from the externally measured time
by not including the time needed for compiling the program.
If the program can compute CPU time and real time (wallclock time),
we selected CPU time.
If that is not available,
we take wallclock time.
These measurements are anyway not used for comparisons
between the programs.
They are used to analyze where the runtime is used up.
For instance,
XSB was earlier quite slow in loading the data
(this was improved with load_dync),
but fast in query execution.

The CPU time in user mode (USER_TIME),
the CPU time in system/kernel mode (SYS_TIME)
and the REAL_TIME are measured externally with /usr/bin/time.
They should be comparable for all systems.

We also measured 'maximum resident set size' with /usr/bin/time.
This is the 'high water mark'
for the amount of memory a process uses
and that is present in real RAM (e.g., not swapped out).
It includes shared libraries (as far as actually used).

For server-based systems such as MariaDB and PostgreSQL
the measurements are more difficult,
because we cannot use /usr/bin/time
for the server process.
The startup time for the server process is not included.

All times are stored in seconds.

* RUN_DATE
    + The date on which the benchmark was run.
    + DATE         NOT NULL
* RUN_NO
    + To distinguish runs of the same program on the same machine,
      a sequential number RUN_NO was added to the key.
    + NUMERIC(4)   NOT NULL
      CHECK(RUN_NO > 0)
* SYS
    + The ID of the program that was used to execute the benchmark,
      e.g. 'xsb'
      (see table TEST_SYS above).
    + VARCHAR(20)  NOT NULL
* INST_NO
    + The installation number of the program on the machine
      (see table SYS_INSTALL above).
    + NUMERIC(4)   NOT NULL
* MACHINE
    + The ID of the machine on which the benchmark was executed
      (see table MACHINE above).
    + VARCHAR(20)  NOT NULL
    + PRIMARY KEY(RUN_DATE,RUN_NO,SYS,INST_NO,MACHINE)
    + FOREIGN KEY(SYS, INST_NO, MACHINE) REFERENCES SYS_INSTALL
* BENCH
    + The ID of the benchmark,
      e.g. 'tcff'
      (see table BENCHMARK above).
    + VARCHAR(20)  NOT NULL
* IMPL
    + This identifies a concrete implementation
      of a benchmark on a program/system,
      e.g. one can distinguish different option settings or indexes in this way
      (see table BENCH_IMPL above).
    + VARCHAR(20)  NOT NULL
    + FOREIGN KEY(SYS, BENCH, IMPL) REFERENCES BENCH_IMPL
* FILE_ID
    + The ID of the input file for which the benchmark was executed
      (see table INPUT_FILE above).
    + VARCHAR(10)  NOT NULL
      REFERENCES INPUT_FILE
* WITH_OUTPUT
    + Since the result of some benchmarks is very large,
      we normally do not printed it
      in the benchmark runs.
      If possible,
      we compute the size of the query result,
      e.g. with 'SELECT COUNT(*)'.
      If not even that is possible
      (e.g. for Prolog systems),
      we simply backtrack over all solutions.
      However, sometimes we check that the system indeed computes
      the correct output.
      In this way,
      we might also detect if the optimizer does too clever optimizations
      because the output is not printed.
      The column WITH_OUTPUT is 'Y' if output is printed,
      and 'N' if not.
      Most views for comparing benchmark results
      only select rows where WITH_OUTPUT='N'.
    + CHAR(1)      NOT NULL
      CHECK(WITH_OUTPUT IN ('Y','N'))
* LOAD_TIME
    + The time (in seconds) that the program itself reported
      as time for loading the data
      (CPU time if possible, otherwise real time).
      This column is NULL if we were not able to get this information easily.
    + NUMERIC(10,3) NULL
      CHECK(LOAD_TIME >= 0)
* EXEC_TIME
    + The time (in seconds) that the program itself reported
      as time for executing the query.
      As for the LOAD_TIME this can be CPU time, real time, or null.
    + NUMERIC(10,3) NULL
      CHECK(EXEC_TIME >= 0)
* TOTAL_TIME
    + The total runtime that the program reported for executing the benchmark.
      Normally, this should be the sum of LOAD_TIME and EXEC_TIME.
      Again, it can be null if the program has no built-in timeing function.
    + NUMERIC(10,3) NULL
      CHECK(TOTAL_TIME >= 0)
* USER_TIME
    + The CPU time that the program needed in user mode
      for executing the benchmark.
      This time is measured externally,
      normally with /usr/bin/time.
      For server-based systems,
      it might not be available (NULL).
    + NUMERIC(10,3) NULL
      CHECK(USER_TIME >= 0),
* SYS_TIME
    + The CPU time that the program needed in system/kernel mode.
      This will mainly be the time for the operating system calls
      for reading the input data.
      For server-based systems,
      it might not be available (NULL).
    + NUMERIC(10,3) NULL
      CHECK(SYS_TIME >= 0)
* REAL_TIME
    + This is the only performance measurement
      that is always available (never NULL).
      It is the real time that the benchmark program has run.
      For server based systems,
      it is only the time that the client has run,
      i.e. the startup and shutdown time for the server is not included.
    + NUMERIC(10,3) NOT NULL
      CHECK(REAL_TIME >= 0)
* MEM
    + This is the amount of memory (RAM) that the program used.
      More precisely,
      it is the 'maximum resident set size' reported by /usr/bin/time
      (in kilobytes).
      Shared libraries do count for the maximum resident set size,
      as far as they are actually used.
      Memory that was longer not accessed by the process does not count
      if the operating system reclaims the page.
      This in particular includes memory that is actually swapped out to disk.
      See e.g.
         - [https://utcc.utoronto.ca/~cks/space/blog/unix/UnderstandingRSS]
         - [https://unix.stackexchange.com/questions/35129/need-explanation-on-resident-set-size-virtual-size]
         - [https://unix.stackexchange.com/questions/30940/getrusage-system-call-what-is-maximum-resident-set-size/30941]
    + NUMERIC(10,3) NULL
      CHECK(MEM >= 0)
* STATUS
    + This should normally be just 'OK'.
      If the benchmark program exited with non-zero exit code,
      it is 'FAILED'.
      This might include cases where a limit was reached,
      e.g. the program used too much memory.
      The valued 'OUTL' is used for marking outliers.
      Measurements with values 'FAILED' and 'OUTL' are ignored
      in the benchmark comparisons.
      There are also various warnings generated by programs
      when real time and CPU time differ unusually much.
    + VARCHAR(80)  NULL
      CHECK(STATUS IN ('OK', 'FAILED', 'OUTL',
      'Warn', 'Warn1', 'Warn2', 'Warn3', 'Warn4')),
* REMARK
    + One could add a remark or tag to a measurement in this column.
    + VARCHAR(800) NULL


`ESTIMATE_PAR`: Experimental Table for Runtime Estimation Parameters
--------------------------------------------------------------------

These are parameters for a formula
that tries to generate an approximation of the measured runtime
from characteristics of the input data.
Currently,
it is only used for the tcff benchmark.
Note that more than one set of parameters can be stored
for the same program.
It is possible to compare the proposed parameters in this way.

* SYS
    + Program ID, see table TEST_SYS above.
    + VARCHAR(20)  NOT NULL
      REFERENCES TEST_SYS
* BENCH
    + Benchmark ID, see table BENCHMARK above.
      Currently, this is actually only used for the tcff benchmark.
    + VARCHAR(20)  NOT NULL
      REFERENCES BENCHMARK
* STARTUP
    + Startup/Initialization time in milliseconds.
    + NUMERIC(10)  NOT NULL
      CHECK(STARTUP >= 0)
* LOAD_IDX
    + Time in milliseconds for loading data
      per one million 'edges * log(edges)'.
      The logarithmic factor simulate the creation of an index.
    + NUMERIC(10)  NOT NULL
      CHECK(LOAD_IDX >= 0)
* RULE_APP
    + Time in milliseconds per one million rule applications (deductions).
      For the tcff benchmark,
      the number of necessary rule applications are stored in the column COST
      of the table INPUT_GRAPH above.
    + NUMERIC(10)  NOT NULL
      CHECK(RULE_APP >= 0)
* RESULT
    + Time in milliseconds be one million result tuples (answers).
      This can be used to give a different weight
      to successful rule applications
      in contrast to computing a duplicate of an already known result.
    + NUMERIC(10)  NOT NULL
      CHECK(RESULT >= 0)
    + PRIMARY KEY(SYS,BENCH,STARTUP,LOAD_IDX,RULE_APP,RESULT)
* HITS
    + This is the number of input graphs (from 43)
      for which the formula generates an acceptable approximation
      of the real runtime.
      Of course,
      we can compute this number with a view (NUM_BAD_ESTIMATES) below.
      The number stored in this table is what our optimization program
      reported.
    + NUMERIC(10)  NOT NULL
      CHECK(HITS >= 0)

`DUMMY_TAB`: Dummy Table with Exactly One Row
---------------------------------------------

This is used in some views
(to write a FROM clause).

The table has only one column:

* ATTR NUMERIC(1)

The table contents is:

|ATTR|
|---:|
|   0|


Views for Analyzing the Collected Data
======================================

Sorry,
currently we can only list the views.
More documentation will be added later.

Statistical Evaluation of Run Data:
-----------------------------------

* RUN_STATISTICS:
  Full statistical evaluation of run data.
  We run each benchmark several times (usually ten times),
  and this view computes averages, minimum, maximum, interval size
  and standard deviation
  for several performance parameters.
* RUN_STAT:
  This is the same as RUN_STATISTICS,
  but without STDDEV, which is not supported in SQLite.
  Since most other views are based on the averages,
  we actually use RUN_STAT in the other views.
* RESULT:
  This is a selection of the benchmark results in RUN_STAT
  for current program version (installation)
  on the standard machine without output.
* BEST_RESULT:
  This is again a subset of RESULT
  that only considers the benchmark implementation for a program
  that was manually marked 'best' in SYS_IMPL.


Outlier Detection, Checking of Results for Plausibility
-------------------------------------------------------

* OUTLIER_REAL:
  Detect Outliers with regard to Real Time.
* OUTLIER_CPU:
  Detect Outliers with regard to CPU Time.
* OUTLIER_MEM:
  Detect Outliers with regard to Memory.
* REAL_TOO_LARGE:
  Check where Realtime is much more than CPU time (at least double).
* PCT_CPU:
  Compare Real Time with CPU Time.
* PCT_CPU_INPUT:
  Compare Real Time with CPU Time for each input file.


Views with Benchmark Results for a Specific System
--------------------------------------------------

There is one view for each tested program
that lists the benchmark results for that program.

* BAM
* XSB
* YAP
* SQLITE
* PSQL
* MARIA
* SOUFFLE
* JENA
* DATASCRIPT


Views to Find/Check Best Option Settings for a System and a Benchmark
---------------------------------------------------------------------

These views are intended to help finding/checking the best implementation
of a benchmark for a program,
e.g. which indexes should be chosen.
The views consider only the runtimes
of the current program version (installation)
on the standard machine.

* BEST_IMPL_REAL:
  Best option settings in real time.
* BEST_IMPL_CPU:
  Best Option Settings (in CPU Time) for a Program for a Benchmark.
* MARIA_COMPARE:
  Compare Storage Managers for MariaDB.


Views for Plots of Performance Data vs. Cost Measures (only for TCFF)
---------------------------------------------------------------------

We use these views for getting the data for plots
of performance data vs. cost measures,
in particular the run time vs. the number of applicable rule instances.
The results are available in csv and tsv formats.
These views look only at the current program version (installation)
on the standard machine,
with the best implementation of the benchmark for the system
(best indexes, options).

* PLOT_REAL:
  Plot of Real Time vs. Cost for TCFF Benchmark:
* PLOT_CPU:
  Plot of CPU Time vs. Cost for TCFF Benchmark
* PLOT_MEM:
  Plot of Memory vs. Number of Tuples for TCFF Benchmark.

Benchmark Results for TCFF - Comparison of Different Systems
------------------------------------------------------------

The following views are for comparing system runtimes for the TCFF benchmark.
These views look only at the current program version (installation)
on the standard machine,
with the best implementation of the benchmark for the system
(best indexes, options).

* TCFF_CPU:
  Benchmark Results for TCFF (Comparison of Various Systems on CPU Time).
* TCFF_REAL:
  Benchmark Results for TCFF (Comparison of Various Systems on REAL Time).
  This version contains the runtime factors compared with BAM.
* TCFF_REAL_FACTORS:
  This is a projection of TCFF_REAL that shows only the factors compared to BAM.
  This makes the output rows a little smaller.
* TCFF_REAL_LATEX:
  LaTeX output for TCFF_REAL.
* TCFF_REAL_XSB:
  Benchmark Results for TCFF (Comparison of Various Systems on REAL Time).
  This version contains the runtime factors compared with XSB
  and rounds the factors to two decimal places after the decimal point.
  Otherwise it is identical to TCFF_REAL (which compares with BAM).
* TCFF_REAL_XSB_FACTORS:
  Projection of TCFF_REAL_XSB on factors only.
* TCFF_REAL_XSB_LATEX:
  LaTeX output from TCFF_REAL_XSB.
* TCFF_REAL_XSB_HTML:
  HTML output form TCFF_REAL_XSB.
* TCFF_MEM:
   Benchmark Results for TCFF (Comparison of Various Systems on Memory).
* TCFF_CPU_LATEX:
  LatEX output for comparison of old selection of systems on CPU time.


Graph Data in Various Formats
-----------------------------

* INPUT_GRAPH_LATEX:
  The table INPUT_GRAPH in LaTeX format.
* INPUT_GRAPH_HTML:
  The table INPUT_GRAPH in HTML format.

Support for Reasearch on Runtime Estimation
-------------------------------------------

* SPEED_REAL:
  System Speed in Real Time
* SPEED_REAL_DIFF:
  System Speed in Real Time With Startup Time
* SPEED_CPU:
  System Speed in CPU Time
* MEM_EFFICIENCY:
  Memory efficiency (memory usage divided by number of facts).
* ESTIMATE_FORMULA:
  Compute Run Time Estimates (Real Time).
* ESTIMATE_ERR:
  Check System Speed Estimates (Real Time).
* NUM_BAD_ESTIMATES:
  Count the Number of Wrong Estimates (more than 20%)
* ESTIMATE_DATA:
  Generate C++ Objects with Graph Data and Runtime (for Optimization).
* LOAD_EFFICIENCY:
  Load time per million edges in s.
* PAR_RANGE:
  Search Range for Parameter Values.

Output Table Data in Markdown Format (for Documentation)
--------------------------------------------------------

* BENCH_IMPL_MD:
  The table BENCH_IMPL in MD format.
* INPUT_FILE_MD:
  The table INPUT_FILE in MD format.
* INPUT_GRAPH_MD:
  The table INPUT_GRAPH in MD format.

Support for Benchmark Execution Scripts
---------------------------------------

* BENCH_FILE_LIST:
  This produces the benchmark file list for use in the RUN_BENCH scripts.
  The result is to be used in a 'case' switch in the bash script
  that checks the FILE_ID.

