// ============================================================================
// Project:	rbench - Logic Programming and Database Benchmarks
// Filename:	opt/opt.cpp
// Purpose:	Optimize parameters for runtime estimation formula
// Last Change:	04.10.2018
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2018 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================

// This must be compiled with the C++11 Standard, because it uses int64_t.
// E.g.:
//	g++ -std=c++11 -Wall -O2 -o opt opt.cpp
// If one uses the parallel version, one needs to add -pthread, e.g.
//	g++ -std=c++11 -pthread -c -Wall -Wextra -Wshadow -Wpointer-arith \
//		-Wcast-qual -Wcast-align -Wunreachable-code -pedantic -O2 \
//		-o opt.o opt.cpp
//	g++ -pthread -o opt opt.o 

//=============================================================================
// Do parallel optimization? How many parallel jobs?
//=============================================================================

// If NUM_JOBS is 1, the serial version will be chosen.

//#define NUM_JOBS 1
//#define NUM_JOBS 4
#define NUM_JOBS 30

//=============================================================================
// Do correctness checks in inner loops (hurts performance):
//=============================================================================

// #define DO_CHECKS 0
#define DO_CHECKS 1

//=============================================================================
// How often should output be printed (step size for percent of work):
//=============================================================================

#define OUT_PCT 5

//=============================================================================
// Include Files:
//=============================================================================

#include <iostream>
#include <cstdint>

#if NUM_THREADS > 1
#include <thread>
#endif

//=============================================================================
// Type for Large Integers (64 Bit):
//=============================================================================

typedef int64_t bigint_t;

//=============================================================================
// String Data Type:
//=============================================================================

typedef const char *str_t;

//-----------------------------------------------------------------------------
// str_eq: Compare two strings for equality.
//-----------------------------------------------------------------------------

bool	str_eq(str_t str1, str_t str2)
{
	// Avoid crashes:
	if(str1 == nullptr)
		return(str2 == nullptr);
	else if(str2 == nullptr)
		return(false);

	// Compare strings:
	while(*str1 && *str1==*str2) {
		str1++;
		str2++;
	}
	return(*str1==*str2);
}

//=============================================================================
// Minimum Function:
//=============================================================================

template<class T>
inline T MIN(T x, T y) {
        if(x <= y)
                return x;
        else
                return y;
}

//=============================================================================
// Integer Division:
//=============================================================================

//-----------------------------------------------------------------------------
// IDIV_ROUND:
//-----------------------------------------------------------------------------

/**
 * Integer division, round result
 *
 * Note that this works only for non-negative numbers.
 */

#define IDIV_ROUND(X,Y)	(((X)+(Y)/2)/(Y))

//-----------------------------------------------------------------------------
// IDIV_CEILING:
//-----------------------------------------------------------------------------

/**
 * Integer division, round up.
 *
 * Note that this works only for non-negative numbers.
 */

#define IDIV_CEILING(X,Y)	(((X)+(Y)-1)/(Y))

//=============================================================================
// Class for Graph Data:
//=============================================================================

class data_c {
	public:

	data_c(str_t rt_graph,
			long rt_nodes, long rt_edges,
			bigint_t rt_tc_size, bigint_t rt_cost,
			long rt_iter, long rt_realtime)
	{
		Graph    = rt_graph;
		Nodes    = rt_nodes;
		Edges    = rt_edges;
		TCSize   = rt_tc_size;
		Cost     = rt_cost;
		Iter     = rt_iter;
		RealTime = rt_realtime;

		// Compute Product "Iterations * Edges" only once:
		IterEdges= static_cast<bigint_t>(rt_iter) * rt_edges;

		// Compute "Edges * log(Edges)":
		int log = 0;
		for(long n = Edges >> 1; n > 0; n >>= 1)
			log++;
		EdgesLogEdges = Edges * log;

		// Compute "Cost * log(TCSize)":
		log = 0;
		for(long n = TCSize >> 1; n > 0; n >>= 1)
			log++;
		CostLogTC = Cost * log;

	}

	inline str_t graph() const { return Graph; }
	inline long nodes() const { return Nodes; }
	inline long edges() const { return Edges; }
	inline bigint_t tc_size() const { return TCSize; }
	inline bigint_t cost() const { return Cost; }
	inline long iter() const { return Iter; }
	inline long real_time() const { return RealTime; }
	inline bigint_t iter_edges() const { return IterEdges; }
	inline bigint_t edges_log_edges() const { return EdgesLogEdges; }
	inline bigint_t cost_log_tc() const { return CostLogTC; }
	
	private:
	str_t Graph;
	long Nodes;
	long Edges;
	bigint_t TCSize;
	bigint_t Cost;
	long Iter;
	long RealTime;
	bigint_t IterEdges;
	bigint_t EdgesLogEdges;
	bigint_t CostLogTC;
};

//-----------------------------------------------------------------------------
// Pointer Type for Graph Data:
//-----------------------------------------------------------------------------

typedef data_c *data_t;

//=============================================================================
// Data Values (exported form the Database):
//=============================================================================

//-----------------------------------------------------------------------------
// Data for XSB:
//-----------------------------------------------------------------------------

data_c data_xsb[] = {
 data_c("k500", 500, 250000, INT64_C(250000), INT64_C(125250000), 1, 15101),
 data_c("k1k", 1000, 1000000, INT64_C(1000000), INT64_C(1001000000), 1, 122002),
 data_c("t500", 500, 124750, INT64_C(124750), INT64_C(62126000), 1, 2777),
 data_c("t1k", 1000, 499500, INT64_C(499500), INT64_C(166666500), 1, 18929),
 data_c("c1k", 1000, 1000, INT64_C(1000000), INT64_C(1001000), 1000, 643),
 data_c("c2k", 2000, 2000, INT64_C(4000000), INT64_C(4002000), 2000, 1840),
 data_c("c3k", 3000, 3000, INT64_C(9000000), INT64_C(9003000), 3000, 3694),
 data_c("c4k", 4000, 4000, INT64_C(16000000), INT64_C(16004000), 4000, 6638),
 data_c("s1k_1", 1000, 2000, INT64_C(1000000), INT64_C(2002000), 500, 712),
 data_c("s1k_3", 1000, 4000, INT64_C(1000000), INT64_C(4004000), 250, 937),
 data_c("s1k_4", 1000, 5000, INT64_C(1000000), INT64_C(5005000), 200, 1038),
 data_c("s2k_1", 2000, 4000, INT64_C(4000000), INT64_C(8004000), 1000, 2167),
 data_c("s2k_3", 2000, 8000, INT64_C(4000000), INT64_C(16008000), 500, 3508),
 data_c("s2k_4", 2000, 10000, INT64_C(4000000), INT64_C(20010000), 400, 4110),
 data_c("p1k", 1000, 999, INT64_C(499500), INT64_C(499500), 999, 396),
 data_c("p2k", 2000, 1999, INT64_C(1999000), INT64_C(1999000), 1999, 1178),
 data_c("p3k", 3000, 2999, INT64_C(4498500), INT64_C(4498500), 2999, 2104),
 data_c("p4k", 4000, 3999, INT64_C(7998000), INT64_C(7998000), 3999, 3632),
 data_c("m4_2ki", 8192, 6144, INT64_C(12288), INT64_C(12288), 3, 117),
 data_c("m16_512", 8192, 7680, INT64_C(61440), INT64_C(61440), 15, 180),
 data_c("m64_128", 8192, 8064, INT64_C(258048), INT64_C(258048), 63, 316),
 data_c("m256_32", 8192, 8160, INT64_C(1044480), INT64_C(1044480), 255, 765),
 data_c("m1ki_8", 8192, 8184, INT64_C(4190208), INT64_C(4190208), 1023, 2128),
 data_c("m4ki_2", 8192, 8190, INT64_C(16773120), INT64_C(16773120), 4095, 7156),
 data_c("b17", 131071, 131070, INT64_C(1966082), INT64_C(1966082), 16, 1417),
 data_c("b18", 262143, 262142, INT64_C(4194306), INT64_C(4194306), 17, 2385),
 data_c("b19", 524287, 524286, INT64_C(8912898), INT64_C(8912898), 18, 4725),
 data_c("y500_4k", 4500, 4499, INT64_C(9998000), INT64_C(9998000), 4000, 4268),
 data_c("y500_8k", 8500, 8499, INT64_C(35996000), INT64_C(35996000), 8000, 14954),
 data_c("y1k_4k", 5000, 4999, INT64_C(11998000), INT64_C(11998000), 4000, 4632),
 data_c("y1k_8k", 9000, 8999, INT64_C(39996000), INT64_C(39996000), 8000, 16239),
 data_c("u1k_50k", 1000, 50000, INT64_C(1000000), INT64_C(50050000), 3, 6649),
 data_c("u1k_125k", 1000, 125000, INT64_C(1000000), INT64_C(125125000), 2, 14620),
 data_c("u1k_250k", 1000, 250000, INT64_C(1000000), INT64_C(250250000), 2, 28911),
 data_c("u2k_200k", 2000, 200000, INT64_C(4000000), INT64_C(400200000), 3, 48136),
 data_c("u2k_500k", 2000, 500000, INT64_C(4000000), INT64_C(1000500000), 2, 117980),
 data_c("u2k_1m", 2000, 1000000, INT64_C(4000000), INT64_C(2001000000), 2, 223119),
 data_c("a1k_50k", 1000, 50000, INT64_C(471621), INT64_C(15269270), 8, 2150),
 data_c("a1k_125k", 1000, 125000, INT64_C(492680), INT64_C(40853746), 5, 5017),
 data_c("a1k_250k", 1000, 250000, INT64_C(497763), INT64_C(82941061), 4, 10169),
 data_c("a2k_200k", 2000, 200000, INT64_C(1944729), INT64_C(128420988), 8, 15172),
 data_c("a2k_500k", 2000, 500000, INT64_C(1985321), INT64_C(330666855), 6, 37856),
 data_c("a2k_1m", 2000, 1000000, INT64_C(1995698), INT64_C(666033613), 4, 78214),
};

//-----------------------------------------------------------------------------
// Data for BAM:
//-----------------------------------------------------------------------------

data_c data_bam[] = {
 data_c("k500", 500, 250000, INT64_C(250000), INT64_C(125250000), 1, 3914),
 data_c("k1k", 1000, 1000000, INT64_C(1000000), INT64_C(1001000000), 1, 31810),
 data_c("t500", 500, 124750, INT64_C(124750), INT64_C(62126000), 1, 1381),
 data_c("t1k", 1000, 499500, INT64_C(499500), INT64_C(166666500), 1, 10467),
 data_c("c1k", 1000, 1000, INT64_C(1000000), INT64_C(1001000), 1000, 60),
 data_c("c2k", 2000, 2000, INT64_C(4000000), INT64_C(4002000), 2000, 290),
 data_c("c3k", 3000, 3000, INT64_C(9000000), INT64_C(9003000), 3000, 868),
 data_c("c4k", 4000, 4000, INT64_C(16000000), INT64_C(16004000), 4000, 2893),
 data_c("s1k_1", 1000, 2000, INT64_C(1000000), INT64_C(2002000), 500, 95),
 data_c("s1k_3", 1000, 4000, INT64_C(1000000), INT64_C(4004000), 250, 166),
 data_c("s1k_4", 1000, 5000, INT64_C(1000000), INT64_C(5005000), 200, 194),
 data_c("s2k_1", 2000, 4000, INT64_C(4000000), INT64_C(8004000), 1000, 457),
 data_c("s2k_3", 2000, 8000, INT64_C(4000000), INT64_C(16008000), 500, 771),
 data_c("s2k_4", 2000, 10000, INT64_C(4000000), INT64_C(20010000), 400, 921),
 data_c("p1k", 1000, 999, INT64_C(499500), INT64_C(499500), 999, 42),
 data_c("p2k", 2000, 1999, INT64_C(1999000), INT64_C(1999000), 1999, 142),
 data_c("p3k", 3000, 2999, INT64_C(4498500), INT64_C(4498500), 2999, 330),
 data_c("p4k", 4000, 3999, INT64_C(7998000), INT64_C(7998000), 3999, 707),
 data_c("m4_2ki", 8192, 6144, INT64_C(12288), INT64_C(12288), 3, 0),
 data_c("m16_512", 8192, 7680, INT64_C(61440), INT64_C(61440), 15, 0),
 data_c("m64_128", 8192, 8064, INT64_C(258048), INT64_C(258048), 63, 48),
 data_c("m256_32", 8192, 8160, INT64_C(1044480), INT64_C(1044480), 255, 109),
 data_c("m1ki_8", 8192, 8184, INT64_C(4190208), INT64_C(4190208), 1023, 315),
 data_c("m4ki_2", 8192, 8190, INT64_C(16773120), INT64_C(16773120), 4095, 3397),
 data_c("b17", 131071, 131070, INT64_C(1966082), INT64_C(1966082), 16, 371),
 data_c("b18", 262143, 262142, INT64_C(4194306), INT64_C(4194306), 17, 1644),
 data_c("b19", 524287, 524286, INT64_C(8912898), INT64_C(8912898), 18, 8045),
 data_c("y500_4k", 4500, 4499, INT64_C(9998000), INT64_C(9998000), 4000, 1041),
 data_c("y500_8k", 8500, 8499, INT64_C(35996000), INT64_C(35996000), 8000, 15916),
 data_c("y1k_4k", 5000, 4999, INT64_C(11998000), INT64_C(11998000), 4000, 1424),
 data_c("y1k_8k", 9000, 8999, INT64_C(39996000), INT64_C(39996000), 8000, 19931),
 data_c("u1k_50k", 1000, 50000, INT64_C(1000000), INT64_C(50050000), 3, 1754),
 data_c("u1k_125k", 1000, 125000, INT64_C(1000000), INT64_C(125125000), 2, 4451),
 data_c("u1k_250k", 1000, 250000, INT64_C(1000000), INT64_C(250250000), 2, 9047),
 data_c("u2k_200k", 2000, 200000, INT64_C(4000000), INT64_C(400200000), 3, 15233),
 data_c("u2k_500k", 2000, 500000, INT64_C(4000000), INT64_C(1000500000), 2, 39528),
 data_c("u2k_1m", 2000, 1000000, INT64_C(4000000), INT64_C(2001000000), 2, 82529),
 data_c("a1k_50k", 1000, 50000, INT64_C(471621), INT64_C(15269270), 8, 614),
 data_c("a1k_125k", 1000, 125000, INT64_C(492680), INT64_C(40853746), 5, 1628),
 data_c("a1k_250k", 1000, 250000, INT64_C(497763), INT64_C(82941061), 4, 3342),
 data_c("a2k_200k", 2000, 200000, INT64_C(1944729), INT64_C(128420988), 8, 6186),
 data_c("a2k_500k", 2000, 500000, INT64_C(1985321), INT64_C(330666855), 6, 16079),
 data_c("a2k_1m", 2000, 1000000, INT64_C(1995698), INT64_C(666033613), 4, 33653),
};

//-----------------------------------------------------------------------------
// Data for YAP:
//-----------------------------------------------------------------------------

data_c data_yap[] = {
 data_c("k500", 500, 250000, INT64_C(250000), INT64_C(125250000), 1, 10388),
 data_c("k1k", 1000, 1000000, INT64_C(1000000), INT64_C(1001000000), 1, 80514),
 data_c("t500", 500, 124750, INT64_C(124750), INT64_C(62126000), 1, 2203),
 data_c("t1k", 1000, 499500, INT64_C(499500), INT64_C(166666500), 1, 14636),
 data_c("c1k", 1000, 1000, INT64_C(1000000), INT64_C(1001000), 1000, 517),
 data_c("c2k", 2000, 2000, INT64_C(4000000), INT64_C(4002000), 2000, 2292),
 data_c("c3k", 3000, 3000, INT64_C(9000000), INT64_C(9003000), 3000, 5176),
 data_c("c4k", 4000, 4000, INT64_C(16000000), INT64_C(16004000), 4000, 9954),
 data_c("s1k_1", 1000, 2000, INT64_C(1000000), INT64_C(2002000), 500, 745),
 data_c("s1k_3", 1000, 4000, INT64_C(1000000), INT64_C(4004000), 250, 1249),
 data_c("s1k_4", 1000, 5000, INT64_C(1000000), INT64_C(5005000), 200, 1512),
 data_c("s2k_1", 2000, 4000, INT64_C(4000000), INT64_C(8004000), 1000, 3244),
 data_c("s2k_3", 2000, 8000, INT64_C(4000000), INT64_C(16008000), 500, 5128),
 data_c("s2k_4", 2000, 10000, INT64_C(4000000), INT64_C(20010000), 400, 6567),
 data_c("p1k", 1000, 999, INT64_C(499500), INT64_C(499500), 999, 271),
 data_c("p2k", 2000, 1999, INT64_C(1999000), INT64_C(1999000), 1999, 1083),
 data_c("p3k", 3000, 2999, INT64_C(4498500), INT64_C(4498500), 2999, 2456),
 data_c("p4k", 4000, 3999, INT64_C(7998000), INT64_C(7998000), 3999, 4631),
 data_c("m4_2ki", 8192, 6144, INT64_C(12288), INT64_C(12288), 3, 50),
 data_c("m16_512", 8192, 7680, INT64_C(61440), INT64_C(61440), 15, 90),
 data_c("m64_128", 8192, 8064, INT64_C(258048), INT64_C(258048), 63, 250),
 data_c("m256_32", 8192, 8160, INT64_C(1044480), INT64_C(1044480), 255, 862),
 data_c("m1ki_8", 8192, 8184, INT64_C(4190208), INT64_C(4190208), 1023, 2872),
 data_c("m4ki_2", 8192, 8190, INT64_C(16773120), INT64_C(16773120), 4095, 10108),
 data_c("b17", 131071, 131070, INT64_C(1966082), INT64_C(1966082), 16, 1307),
 data_c("b18", 262143, 262142, INT64_C(4194306), INT64_C(4194306), 17, 2533),
 data_c("b19", 524287, 524286, INT64_C(8912898), INT64_C(8912898), 18, 5153),
 data_c("y500_4k", 4500, 4499, INT64_C(9998000), INT64_C(9998000), 4000, 4945),
 data_c("y500_8k", 8500, 8499, INT64_C(35996000), INT64_C(35996000), 8000, 20071),
 data_c("y1k_4k", 5000, 4999, INT64_C(11998000), INT64_C(11998000), 4000, 5065),
 data_c("y1k_8k", 9000, 8999, INT64_C(39996000), INT64_C(39996000), 8000, 21155),
 data_c("u1k_50k", 1000, 50000, INT64_C(1000000), INT64_C(50050000), 3, 13523),
 data_c("u1k_125k", 1000, 125000, INT64_C(1000000), INT64_C(125125000), 2, 33144),
 data_c("u1k_250k", 1000, 250000, INT64_C(1000000), INT64_C(250250000), 2, 62955),
 data_c("u2k_200k", 2000, 200000, INT64_C(4000000), INT64_C(400200000), 3, 127953),
 data_c("u2k_500k", 2000, 500000, INT64_C(4000000), INT64_C(1000500000), 2, 301431),
 data_c("u2k_1m", 2000, 1000000, INT64_C(4000000), INT64_C(2001000000), 2, 553727),
 data_c("a1k_50k", 1000, 50000, INT64_C(471621), INT64_C(15269270), 8, 2526),
 data_c("a1k_125k", 1000, 125000, INT64_C(492680), INT64_C(40853746), 5, 7592),
 data_c("a1k_250k", 1000, 250000, INT64_C(497763), INT64_C(82941061), 4, 17970),
 data_c("a2k_200k", 2000, 200000, INT64_C(1944729), INT64_C(128420988), 8, 18528),
 data_c("a2k_500k", 2000, 500000, INT64_C(1985321), INT64_C(330666855), 6, 62601),
 data_c("a2k_1m", 2000, 1000000, INT64_C(1995698), INT64_C(666033613), 4, 164460),
};

//-----------------------------------------------------------------------------
// Data for SQLite:
//-----------------------------------------------------------------------------


data_c data_sqlite[] = {
 data_c("k500", 500, 250000, INT64_C(250000), INT64_C(125250000), 1, 73760),
 data_c("k1k", 1000, 1000000, INT64_C(1000000), INT64_C(1001000000), 1, 610108),
 data_c("t500", 500, 124750, INT64_C(124750), INT64_C(62126000), 1, 11794),
 data_c("t1k", 1000, 499500, INT64_C(499500), INT64_C(166666500), 1, 97659),
 data_c("c1k", 1000, 1000, INT64_C(1000000), INT64_C(1001000), 1000, 2209),
 data_c("c2k", 2000, 2000, INT64_C(4000000), INT64_C(4002000), 2000, 8497),
 data_c("c3k", 3000, 3000, INT64_C(9000000), INT64_C(9003000), 3000, 19711),
 data_c("c4k", 4000, 4000, INT64_C(16000000), INT64_C(16004000), 4000, 35095),
 data_c("s1k_1", 1000, 2000, INT64_C(1000000), INT64_C(2002000), 500, 2732),
 data_c("s1k_3", 1000, 4000, INT64_C(1000000), INT64_C(4004000), 250, 3703),
 data_c("s1k_4", 1000, 5000, INT64_C(1000000), INT64_C(5005000), 200, 4175),
 data_c("s2k_1", 2000, 4000, INT64_C(4000000), INT64_C(8004000), 1000, 10883),
 data_c("s2k_3", 2000, 8000, INT64_C(4000000), INT64_C(16008000), 500, 15366),
 data_c("s2k_4", 2000, 10000, INT64_C(4000000), INT64_C(20010000), 400, 17292),
 data_c("p1k", 1000, 999, INT64_C(499500), INT64_C(499500), 999, 1181),
 data_c("p2k", 2000, 1999, INT64_C(1999000), INT64_C(1999000), 1999, 4152),
 data_c("p3k", 3000, 2999, INT64_C(4498500), INT64_C(4498500), 2999, 9239),
 data_c("p4k", 4000, 3999, INT64_C(7998000), INT64_C(7998000), 3999, 16358),
 data_c("m4_2ki", 8192, 6144, INT64_C(12288), INT64_C(12288), 3, 120),
 data_c("m16_512", 8192, 7680, INT64_C(61440), INT64_C(61440), 15, 262),
 data_c("m64_128", 8192, 8064, INT64_C(258048), INT64_C(258048), 63, 699),
 data_c("m256_32", 8192, 8160, INT64_C(1044480), INT64_C(1044480), 255, 2189),
 data_c("m1ki_8", 8192, 8184, INT64_C(4190208), INT64_C(4190208), 1023, 8644),
 data_c("m4ki_2", 8192, 8190, INT64_C(16773120), INT64_C(16773120), 4095, 37943),
 data_c("b17", 131071, 131070, INT64_C(1966082), INT64_C(1966082), 16, 4283),
 data_c("b18", 262143, 262142, INT64_C(4194306), INT64_C(4194306), 17, 8744),
 data_c("b19", 524287, 524286, INT64_C(8912898), INT64_C(8912898), 18, 18373),
 data_c("y500_4k", 4500, 4499, INT64_C(9998000), INT64_C(9998000), 4000, 21699),
 data_c("y500_8k", 8500, 8499, INT64_C(35996000), INT64_C(35996000), 8000, 90900),
 data_c("y1k_4k", 5000, 4999, INT64_C(11998000), INT64_C(11998000), 4000, 26871),
 data_c("y1k_8k", 9000, 8999, INT64_C(39996000), INT64_C(39996000), 8000, 103049),
 data_c("u1k_50k", 1000, 50000, INT64_C(1000000), INT64_C(50050000), 3, 35178),
 data_c("u1k_125k", 1000, 125000, INT64_C(1000000), INT64_C(125125000), 2, 82903),
 data_c("u1k_250k", 1000, 250000, INT64_C(1000000), INT64_C(250250000), 2, 163032),
 data_c("u2k_200k", 2000, 200000, INT64_C(4000000), INT64_C(400200000), 3, 310434),
 data_c("u2k_500k", 2000, 500000, INT64_C(4000000), INT64_C(1000500000), 2, 699210),
 data_c("u2k_1m", 2000, 1000000, INT64_C(4000000), INT64_C(2001000000), 2, 1342575),
 data_c("a1k_50k", 1000, 50000, INT64_C(471621), INT64_C(15269270), 8, 10554),
 data_c("a1k_125k", 1000, 125000, INT64_C(492680), INT64_C(40853746), 5, 26000),
 data_c("a1k_250k", 1000, 250000, INT64_C(497763), INT64_C(82941061), 4, 52263),
 data_c("a2k_200k", 2000, 200000, INT64_C(1944729), INT64_C(128420988), 8, 95773),
 data_c("a2k_500k", 2000, 500000, INT64_C(1985321), INT64_C(330666855), 6, 226843),
 data_c("a2k_1m", 2000, 1000000, INT64_C(1995698), INT64_C(666033613), 4, 441177),
};

//-----------------------------------------------------------------------------
// Data for PostgreSQL:
//-----------------------------------------------------------------------------

data_c data_psql[] = {
data_c("k500", 500, 250000, INT64_C(250000), INT64_C(125250000), 1, 43082),
 data_c("k1k", 1000, 1000000, INT64_C(1000000), INT64_C(1001000000), 1, 397291),
 data_c("t500", 500, 124750, INT64_C(124750), INT64_C(62126000), 1, 6654),
 data_c("t1k", 1000, 499500, INT64_C(499500), INT64_C(166666500), 1, 52657),
 data_c("c1k", 1000, 1000, INT64_C(1000000), INT64_C(1001000), 1000, 1522),
 data_c("c2k", 2000, 2000, INT64_C(4000000), INT64_C(4002000), 2000, 5438),
 data_c("c3k", 3000, 3000, INT64_C(9000000), INT64_C(9003000), 3000, 12141),
 data_c("c4k", 4000, 4000, INT64_C(16000000), INT64_C(16004000), 4000, 21916),
 data_c("s1k_1", 1000, 2000, INT64_C(1000000), INT64_C(2002000), 500, 1917),
 data_c("s1k_3", 1000, 4000, INT64_C(1000000), INT64_C(4004000), 250, 2486),
 data_c("s1k_4", 1000, 5000, INT64_C(1000000), INT64_C(5005000), 200, 2776),
 data_c("s2k_1", 2000, 4000, INT64_C(4000000), INT64_C(8004000), 1000, 7057),
 data_c("s2k_3", 2000, 8000, INT64_C(4000000), INT64_C(16008000), 500, 9808),
 data_c("s2k_4", 2000, 10000, INT64_C(4000000), INT64_C(20010000), 400, 11224),
 data_c("p1k", 1000, 999, INT64_C(499500), INT64_C(499500), 999, 848),
 data_c("p2k", 2000, 1999, INT64_C(1999000), INT64_C(1999000), 1999, 2698),
 data_c("p3k", 3000, 2999, INT64_C(4498500), INT64_C(4498500), 2999, 5753),
 data_c("p4k", 4000, 3999, INT64_C(7998000), INT64_C(7998000), 3999, 10111),
 data_c("m4_2ki", 8192, 6144, INT64_C(12288), INT64_C(12288), 3, 154),
 data_c("m16_512", 8192, 7680, INT64_C(61440), INT64_C(61440), 15, 261),
 data_c("m64_128", 8192, 8064, INT64_C(258048), INT64_C(258048), 63, 534),
 data_c("m256_32", 8192, 8160, INT64_C(1044480), INT64_C(1044480), 255, 1543),
 data_c("m1ki_8", 8192, 8184, INT64_C(4190208), INT64_C(4190208), 1023, 5440),
 data_c("m4ki_2", 8192, 8190, INT64_C(16773120), INT64_C(16773120), 4095, 21941),
 data_c("b17", 131071, 131070, INT64_C(1966082), INT64_C(1966082), 16, 3452),
 data_c("b18", 262143, 262142, INT64_C(4194306), INT64_C(4194306), 17, 7044),
 data_c("b19", 524287, 524286, INT64_C(8912898), INT64_C(8912898), 18, 14803),
 data_c("y500_4k", 4500, 4499, INT64_C(9998000), INT64_C(9998000), 4000, 13387),
 data_c("y500_8k", 8500, 8499, INT64_C(35996000), INT64_C(35996000), 8000, 50795),
 data_c("y1k_4k", 5000, 4999, INT64_C(11998000), INT64_C(11998000), 4000, 15581),
 data_c("y1k_8k", 9000, 8999, INT64_C(39996000), INT64_C(39996000), 8000, 55595),
 data_c("u1k_50k", 1000, 50000, INT64_C(1000000), INT64_C(50050000), 3, 22644),
 data_c("u1k_125k", 1000, 125000, INT64_C(1000000), INT64_C(125125000), 2, 54067),
 data_c("u1k_250k", 1000, 250000, INT64_C(1000000), INT64_C(250250000), 2, 108861),
 data_c("u2k_200k", 2000, 200000, INT64_C(4000000), INT64_C(400200000), 3, 208007),
 data_c("u2k_500k", 2000, 500000, INT64_C(4000000), INT64_C(1000500000), 2, 539281),
 data_c("u2k_1m", 2000, 1000000, INT64_C(4000000), INT64_C(2001000000), 2, 1151658),
 data_c("a1k_50k", 1000, 50000, INT64_C(471621), INT64_C(15269270), 8, 6262),
 data_c("a1k_125k", 1000, 125000, INT64_C(492680), INT64_C(40853746), 5, 15848),
 data_c("a1k_250k", 1000, 250000, INT64_C(497763), INT64_C(82941061), 4, 33724),
 data_c("a2k_200k", 2000, 200000, INT64_C(1944729), INT64_C(128420988), 8, 72064),
 data_c("a2k_500k", 2000, 500000, INT64_C(1985321), INT64_C(330666855), 6, 174759),
 data_c("a2k_1m", 2000, 1000000, INT64_C(1995698), INT64_C(666033613), 4, 386183),
};

//-----------------------------------------------------------------------------
// Data for MariaDB (with Memory Storage Manager):
//-----------------------------------------------------------------------------

data_c data_maria_m[] = {
 data_c("k500", 500, 250000, INT64_C(250000), INT64_C(125250000), 1, 52375),
 data_c("k1k", 1000, 1000000, INT64_C(1000000), INT64_C(1001000000), 1, 451933),
 data_c("t500", 500, 124750, INT64_C(124750), INT64_C(62126000), 1, 10348),
 data_c("t1k", 1000, 499500, INT64_C(499500), INT64_C(166666500), 1, 74740),
 data_c("c1k", 1000, 1000, INT64_C(1000000), INT64_C(1001000), 1000, 1167),
 data_c("c2k", 2000, 2000, INT64_C(4000000), INT64_C(4002000), 2000, 46490),
 data_c("c3k", 3000, 3000, INT64_C(9000000), INT64_C(9003000), 3000, 157534),
 data_c("c4k", 4000, 4000, INT64_C(16000000), INT64_C(16004000), 4000, 327722),
 data_c("s1k_1", 1000, 2000, INT64_C(1000000), INT64_C(2002000), 500, 1531),
 data_c("s1k_3", 1000, 4000, INT64_C(1000000), INT64_C(4004000), 250, 2314),
 data_c("s1k_4", 1000, 5000, INT64_C(1000000), INT64_C(5005000), 200, 2746),
 data_c("s2k_1", 2000, 4000, INT64_C(4000000), INT64_C(8004000), 1000, 80884),
 data_c("s2k_3", 2000, 8000, INT64_C(4000000), INT64_C(16008000), 500, 144601),
 data_c("s2k_4", 2000, 10000, INT64_C(4000000), INT64_C(20010000), 400, 169460),
 data_c("p1k", 1000, 999, INT64_C(499500), INT64_C(499500), 999, 685),
 data_c("p2k", 2000, 1999, INT64_C(1999000), INT64_C(1999000), 1999, 6567),
 data_c("p3k", 3000, 2999, INT64_C(4498500), INT64_C(4498500), 2999, 56763),
 data_c("p4k", 4000, 3999, INT64_C(7998000), INT64_C(7998000), 3999, 133877),
 data_c("m4_2ki", 8192, 6144, INT64_C(12288), INT64_C(12288), 3, 54),
 data_c("m16_512", 8192, 7680, INT64_C(61440), INT64_C(61440), 15, 150),
 data_c("m64_128", 8192, 8064, INT64_C(258048), INT64_C(258048), 63, 375),
 data_c("m256_32", 8192, 8160, INT64_C(1044480), INT64_C(1044480), 255, 1161),
 data_c("m1ki_8", 8192, 8184, INT64_C(4190208), INT64_C(4190208), 1023, 52849),
 data_c("m4ki_2", 8192, 8190, INT64_C(16773120), INT64_C(16773120), 4095, 334514),
 data_c("b17", 131071, 131070, INT64_C(1966082), INT64_C(1966082), 16, 6320),
 data_c("b18", 262143, 262142, INT64_C(4194306), INT64_C(4194306), 17, 52041),
 data_c("b19", 524287, 524286, INT64_C(8912898), INT64_C(8912898), 18, 163716),
 data_c("y500_4k", 4500, 4499, INT64_C(9998000), INT64_C(9998000), 4000, 177141),
 data_c("y500_8k", 8500, 8499, INT64_C(35996000), INT64_C(35996000), 8000, 841026),
 data_c("y1k_4k", 5000, 4999, INT64_C(11998000), INT64_C(11998000), 4000, 224334),
 data_c("y1k_8k", 9000, 8999, INT64_C(39996000), INT64_C(39996000), 8000, 939970),
 data_c("u1k_50k", 1000, 50000, INT64_C(1000000), INT64_C(50050000), 3, 30010),
 data_c("u1k_125k", 1000, 125000, INT64_C(1000000), INT64_C(125125000), 2, 74156),
 data_c("u1k_250k", 1000, 250000, INT64_C(1000000), INT64_C(250250000), 2, 155192),
 data_c("u2k_200k", 2000, 200000, INT64_C(4000000), INT64_C(400200000), 3, 5214560),
 data_c("u2k_500k", 2000, 500000, INT64_C(4000000), INT64_C(1000500000), 2, 12960970),
 data_c("u2k_1m", 2000, 1000000, INT64_C(4000000), INT64_C(2001000000), 2, 25949750),
 data_c("a1k_50k", 1000, 50000, INT64_C(471621), INT64_C(15269270), 8, 8022),
 data_c("a1k_125k", 1000, 125000, INT64_C(492680), INT64_C(40853746), 5, 20220),
 data_c("a1k_250k", 1000, 250000, INT64_C(497763), INT64_C(82941061), 4, 44247),
 data_c("a2k_200k", 2000, 200000, INT64_C(1944729), INT64_C(128420988), 8, 1406857),
 data_c("a2k_500k", 2000, 500000, INT64_C(1985321), INT64_C(330666855), 6, 4064160),
 data_c("a2k_1m", 2000, 1000000, INT64_C(1995698), INT64_C(666033613), 4, 8406610),
};

//-----------------------------------------------------------------------------
// Data for MariaDB (with InnoDB Storage Manager):
//-----------------------------------------------------------------------------

data_c data_maria_i[] = {
 data_c("k500", 500, 250000, INT64_C(250000), INT64_C(125250000), 1, 63180),
 data_c("k1k", 1000, 1000000, INT64_C(1000000), INT64_C(1001000000), 1, 533213),
 data_c("t500", 500, 124750, INT64_C(124750), INT64_C(62126000), 1, 13234),
 data_c("t1k", 1000, 499500, INT64_C(499500), INT64_C(166666500), 1, 82985),
 data_c("c1k", 1000, 1000, INT64_C(1000000), INT64_C(1001000), 1000, 1436),
 data_c("c2k", 2000, 2000, INT64_C(4000000), INT64_C(4002000), 2000, 47730),
 data_c("c3k", 3000, 3000, INT64_C(9000000), INT64_C(9003000), 3000, 160576),
 data_c("c4k", 4000, 4000, INT64_C(16000000), INT64_C(16004000), 4000, 336150),
 data_c("s1k_1", 1000, 2000, INT64_C(1000000), INT64_C(2002000), 500, 1984),
 data_c("s1k_3", 1000, 4000, INT64_C(1000000), INT64_C(4004000), 250, 3039),
 data_c("s1k_4", 1000, 5000, INT64_C(1000000), INT64_C(5005000), 200, 3438),
 data_c("s2k_1", 2000, 4000, INT64_C(4000000), INT64_C(8004000), 1000, 87717),
 data_c("s2k_3", 2000, 8000, INT64_C(4000000), INT64_C(16008000), 500, 149004),
 data_c("s2k_4", 2000, 10000, INT64_C(4000000), INT64_C(20010000), 400, 180097),
 data_c("p1k", 1000, 999, INT64_C(499500), INT64_C(499500), 999, 1074),
 data_c("p2k", 2000, 1999, INT64_C(1999000), INT64_C(1999000), 1999, 8269),
 data_c("p3k", 3000, 2999, INT64_C(4498500), INT64_C(4498500), 2999, 64731),
 data_c("p4k", 4000, 3999, INT64_C(7998000), INT64_C(7998000), 3999, 149327),
 data_c("m4_2ki", 8192, 6144, INT64_C(12288), INT64_C(12288), 3, 111),
 data_c("m16_512", 8192, 7680, INT64_C(61440), INT64_C(61440), 15, 256),
 data_c("m64_128", 8192, 8064, INT64_C(258048), INT64_C(258048), 63, 631),
 data_c("m256_32", 8192, 8160, INT64_C(1044480), INT64_C(1044480), 255, 1936),
 data_c("m1ki_8", 8192, 8184, INT64_C(4190208), INT64_C(4190208), 1023, 57826),
 data_c("m4ki_2", 8192, 8190, INT64_C(16773120), INT64_C(16773120), 4095, 367509),
 data_c("b17", 131071, 131070, INT64_C(1966082), INT64_C(1966082), 16, 7255),
 data_c("b18", 262143, 262142, INT64_C(4194306), INT64_C(4194306), 17, 49346),
 data_c("b19", 524287, 524286, INT64_C(8912898), INT64_C(8912898), 18, 148038),
 data_c("y500_4k", 4500, 4499, INT64_C(9998000), INT64_C(9998000), 4000, 219946),
 data_c("y500_8k", 8500, 8499, INT64_C(35996000), INT64_C(35996000), 8000, 1022508),
 data_c("y1k_4k", 5000, 4999, INT64_C(11998000), INT64_C(11998000), 4000, 264605),
 data_c("y1k_8k", 9000, 8999, INT64_C(39996000), INT64_C(39996000), 8000, 1110265),
 data_c("u1k_50k", 1000, 50000, INT64_C(1000000), INT64_C(50050000), 3, 26625),
 data_c("u1k_125k", 1000, 125000, INT64_C(1000000), INT64_C(125125000), 2, 62649),
 data_c("u1k_250k", 1000, 250000, INT64_C(1000000), INT64_C(250250000), 2, 120097),
 data_c("u2k_200k", 2000, 200000, INT64_C(4000000), INT64_C(400200000), 3, 5603130),
 data_c("u2k_500k", 2000, 500000, INT64_C(4000000), INT64_C(1000500000), 2, 13645480),
 data_c("u2k_1m", 2000, 1000000, INT64_C(4000000), INT64_C(2001000000), 2, 25857380),
 data_c("a1k_50k", 1000, 50000, INT64_C(471621), INT64_C(15269270), 8, 8240),
 data_c("a1k_125k", 1000, 125000, INT64_C(492680), INT64_C(40853746), 5, 20341),
 data_c("a1k_250k", 1000, 250000, INT64_C(497763), INT64_C(82941061), 4, 39912),
 data_c("a2k_200k", 2000, 200000, INT64_C(1944729), INT64_C(128420988), 8, 664055),
 data_c("a2k_500k", 2000, 500000, INT64_C(1985321), INT64_C(330666855), 6, 3780760),
 data_c("a2k_1m", 2000, 1000000, INT64_C(1995698), INT64_C(666033613), 4, 5468940),
};

//-----------------------------------------------------------------------------
// Data for Souffle:
//-----------------------------------------------------------------------------

data_c data_souffle[] = {
 data_c("k500", 500, 250000, INT64_C(250000), INT64_C(125250000), 1, 1371),
 data_c("k1k", 1000, 1000000, INT64_C(1000000), INT64_C(1001000000), 1, 8032),
 data_c("t500", 500, 124750, INT64_C(124750), INT64_C(62126000), 1, 369),
 data_c("t1k", 1000, 499500, INT64_C(499500), INT64_C(166666500), 1, 1748),
 data_c("c1k", 1000, 1000, INT64_C(1000000), INT64_C(1001000), 1000, 771),
 data_c("c2k", 2000, 2000, INT64_C(4000000), INT64_C(4002000), 2000, 3013),
 data_c("c3k", 3000, 3000, INT64_C(9000000), INT64_C(9003000), 3000, 6802),
 data_c("c4k", 4000, 4000, INT64_C(16000000), INT64_C(16004000), 4000, 11482),
 data_c("s1k_1", 1000, 2000, INT64_C(1000000), INT64_C(2002000), 500, 547),
 data_c("s1k_3", 1000, 4000, INT64_C(1000000), INT64_C(4004000), 250, 434),
 data_c("s1k_4", 1000, 5000, INT64_C(1000000), INT64_C(5005000), 200, 432),
 data_c("s2k_1", 2000, 4000, INT64_C(4000000), INT64_C(8004000), 1000, 2065),
 data_c("s2k_3", 2000, 8000, INT64_C(4000000), INT64_C(16008000), 500, 1610),
 data_c("s2k_4", 2000, 10000, INT64_C(4000000), INT64_C(20010000), 400, 1534),
 data_c("p1k", 1000, 999, INT64_C(499500), INT64_C(499500), 999, 639),
 data_c("p2k", 2000, 1999, INT64_C(1999000), INT64_C(1999000), 1999, 2222),
 data_c("p3k", 3000, 2999, INT64_C(4498500), INT64_C(4498500), 2999, 4958),
 data_c("p4k", 4000, 3999, INT64_C(7998000), INT64_C(7998000), 3999, 8235),
 data_c("m4_2ki", 8192, 6144, INT64_C(12288), INT64_C(12288), 3, 50),
 data_c("m16_512", 8192, 7680, INT64_C(61440), INT64_C(61440), 15, 154),
 data_c("m64_128", 8192, 8064, INT64_C(258048), INT64_C(258048), 63, 408),
 data_c("m256_32", 8192, 8160, INT64_C(1044480), INT64_C(1044480), 255, 1258),
 data_c("m1ki_8", 8192, 8184, INT64_C(4190208), INT64_C(4190208), 1023, 4067),
 data_c("m4ki_2", 8192, 8190, INT64_C(16773120), INT64_C(16773120), 4095, 16569),
 data_c("b17", 131071, 131070, INT64_C(1966082), INT64_C(1966082), 16, 531),
 data_c("b18", 262143, 262142, INT64_C(4194306), INT64_C(4194306), 17, 971),
 data_c("b19", 524287, 524286, INT64_C(8912898), INT64_C(8912898), 18, 1713),
 data_c("y500_4k", 4500, 4499, INT64_C(9998000), INT64_C(9998000), 4000, 10595),
 data_c("y500_8k", 8500, 8499, INT64_C(35996000), INT64_C(35996000), 8000, 40997),
 data_c("y1k_4k", 5000, 4999, INT64_C(11998000), INT64_C(11998000), 4000, 11688),
 data_c("y1k_8k", 9000, 8999, INT64_C(39996000), INT64_C(39996000), 8000, 43208),
 data_c("u1k_50k", 1000, 50000, INT64_C(1000000), INT64_C(50050000), 3, 789),
 data_c("u1k_125k", 1000, 125000, INT64_C(1000000), INT64_C(125125000), 2, 1452),
 data_c("u1k_250k", 1000, 250000, INT64_C(1000000), INT64_C(250250000), 2, 2859),
 data_c("u2k_200k", 2000, 200000, INT64_C(4000000), INT64_C(400200000), 3, 3771),
 data_c("u2k_500k", 2000, 500000, INT64_C(4000000), INT64_C(1000500000), 2, 9221),
 data_c("u2k_1m", 2000, 1000000, INT64_C(4000000), INT64_C(2001000000), 2, 20321),
 data_c("a1k_50k", 1000, 50000, INT64_C(471621), INT64_C(15269270), 8, 349),
 data_c("a1k_125k", 1000, 125000, INT64_C(492680), INT64_C(40853746), 5, 796),
 data_c("a1k_250k", 1000, 250000, INT64_C(497763), INT64_C(82941061), 4, 1440),
 data_c("a2k_200k", 2000, 200000, INT64_C(1944729), INT64_C(128420988), 8, 1621),
 data_c("a2k_500k", 2000, 500000, INT64_C(1985321), INT64_C(330666855), 6, 3925),
 data_c("a2k_1m", 2000, 1000000, INT64_C(1995698), INT64_C(666033613), 4, 8479),
};

//-----------------------------------------------------------------------------
// Data for Jena:
//-----------------------------------------------------------------------------

data_c data_jena[] = {
 data_c("k500", 500, 250000, INT64_C(250000), INT64_C(125250000), 1, 27022),
 data_c("k1k", 1000, 1000000, INT64_C(1000000), INT64_C(1001000000), 1, 202835),
 data_c("t500", 500, 124750, INT64_C(124750), INT64_C(62126000), 1, 7101),
 data_c("t1k", 1000, 499500, INT64_C(499500), INT64_C(166666500), 1, 38450),
 data_c("c1k", 1000, 1000, INT64_C(1000000), INT64_C(1001000), 1000, 3571),
 data_c("c2k", 2000, 2000, INT64_C(4000000), INT64_C(4002000), 2000, 8228),
 data_c("c3k", 3000, 3000, INT64_C(9000000), INT64_C(9003000), 3000, 14944),
 data_c("c4k", 4000, 4000, INT64_C(16000000), INT64_C(16004000), 4000, 23833),
 data_c("s1k_1", 1000, 2000, INT64_C(1000000), INT64_C(2002000), 500, 3927),
 data_c("s1k_3", 1000, 4000, INT64_C(1000000), INT64_C(4004000), 250, 4491),
 data_c("s1k_4", 1000, 5000, INT64_C(1000000), INT64_C(5005000), 200, 4616),
 data_c("s2k_1", 2000, 4000, INT64_C(4000000), INT64_C(8004000), 1000, 9480),
 data_c("s2k_3", 2000, 8000, INT64_C(4000000), INT64_C(16008000), 500, 10898),
 data_c("s2k_4", 2000, 10000, INT64_C(4000000), INT64_C(20010000), 400, 11484),
 data_c("p1k", 1000, 999, INT64_C(499500), INT64_C(499500), 999, 2653),
 data_c("p2k", 2000, 1999, INT64_C(1999000), INT64_C(1999000), 1999, 5268),
 data_c("p3k", 3000, 2999, INT64_C(4498500), INT64_C(4498500), 2999, 8948),
 data_c("p4k", 4000, 3999, INT64_C(7998000), INT64_C(7998000), 3999, 14285),
 data_c("m4_2ki", 8192, 6144, INT64_C(12288), INT64_C(12288), 3, 1938),
 data_c("m16_512", 8192, 7680, INT64_C(61440), INT64_C(61440), 15, 2166),
 data_c("m64_128", 8192, 8064, INT64_C(258048), INT64_C(258048), 63, 2472),
 data_c("m256_32", 8192, 8160, INT64_C(1044480), INT64_C(1044480), 255, 3891),
 data_c("m1ki_8", 8192, 8184, INT64_C(4190208), INT64_C(4190208), 1023, 8964),
 data_c("m4ki_2", 8192, 8190, INT64_C(16773120), INT64_C(16773120), 4095, 26142),
 data_c("b17", 131071, 131070, INT64_C(1966082), INT64_C(1966082), 16, 8231),
 data_c("b18", 262143, 262142, INT64_C(4194306), INT64_C(4194306), 17, 15447),
 data_c("b19", 524287, 524286, INT64_C(8912898), INT64_C(8912898), 18, 33542),
 data_c("y500_4k", 4500, 4499, INT64_C(9998000), INT64_C(9998000), 4000, 16454),
 data_c("y500_8k", 8500, 8499, INT64_C(35996000), INT64_C(35996000), 8000, 58143),
 data_c("y1k_4k", 5000, 4999, INT64_C(11998000), INT64_C(11998000), 4000, 19440),
 data_c("y1k_8k", 9000, 8999, INT64_C(39996000), INT64_C(39996000), 8000, 63759),
 data_c("u1k_50k", 1000, 50000, INT64_C(1000000), INT64_C(50050000), 3, 15188),
 data_c("u1k_125k", 1000, 125000, INT64_C(1000000), INT64_C(125125000), 2, 31381),
 data_c("u1k_250k", 1000, 250000, INT64_C(1000000), INT64_C(250250000), 2, 56304),
 data_c("u2k_200k", 2000, 200000, INT64_C(4000000), INT64_C(400200000), 3, 96155),
 data_c("u2k_500k", 2000, 500000, INT64_C(4000000), INT64_C(1000500000), 2, 223011),
 data_c("u2k_1m", 2000, 1000000, INT64_C(4000000), INT64_C(2001000000), 2, 405297),
 data_c("a1k_50k", 1000, 50000, INT64_C(471621), INT64_C(15269270), 8, 6710),
 data_c("a1k_125k", 1000, 125000, INT64_C(492680), INT64_C(40853746), 5, 12874),
 data_c("a1k_250k", 1000, 250000, INT64_C(497763), INT64_C(82941061), 4, 22482),
 data_c("a2k_200k", 2000, 200000, INT64_C(1944729), INT64_C(128420988), 8, 35862),
 data_c("a2k_500k", 2000, 500000, INT64_C(1985321), INT64_C(330666855), 6, 81462),
 data_c("a2k_1m", 2000, 1000000, INT64_C(1995698), INT64_C(666033613), 4, 155457),
};

//-----------------------------------------------------------------------------
// Normal Array Size:
//-----------------------------------------------------------------------------

const int NUM_GRAPHS = sizeof(data_xsb)/sizeof(data_c);


//=============================================================================
// Class for Systems (with Array of Data Values):
//=============================================================================

class sys_c {
	public:

	// Constructor:
	sys_c(str_t sys_name, int sys_known_opt,
			data_c *sys_data, size_t sys_data_size) {
		Name = sys_name;
		KnownOpt = sys_known_opt;
		Data = sys_data;
		if(sys_data_size != NUM_GRAPHS) {
			std::cerr << "Wrong number of entries " <<
				"in data array for " << sys_name <<
				" (" << sys_data_size << ", correct: " <<
				NUM_GRAPHS << ")!\n";
			exit(4);
		}
	}

	// Accessor methods:
	str_t name() const { return Name; }
	int known_opt() const { return KnownOpt; }
	data_c *data() const { return Data; }


	private:

	// Attributes:
	str_t Name;
	int KnownOpt;
	data_c *Data;
};

//-----------------------------------------------------------------------------
// Pointer Type for System Data:
//-----------------------------------------------------------------------------

typedef sys_c *sys_t;

//-----------------------------------------------------------------------------
// List of Systems with Data Arrays:
//-----------------------------------------------------------------------------

sys_c systems[] = {
	sys_c("xsb", 41, data_xsb, sizeof(data_xsb)/sizeof(data_c)),
	sys_c("bam", 29, data_bam, sizeof(data_bam)/sizeof(data_c)),
	sys_c("yap", 32, data_yap, sizeof(data_yap)/sizeof(data_c)),
	sys_c("sqlite", 42, data_sqlite,
					sizeof(data_sqlite)/sizeof(data_c)),
	sys_c("psql", 38, data_psql, sizeof(data_psql)/sizeof(data_c)),
	sys_c("maria_m", 18, data_maria_m, sizeof(data_maria_m)/sizeof(data_c)),
	sys_c("maria_i", 19, data_maria_i, sizeof(data_maria_i)/sizeof(data_c)),
	sys_c("souffle", 24, data_souffle, sizeof(data_souffle)/sizeof(data_c)),
	sys_c("jena", 42, data_jena, sizeof(data_jena)/sizeof(data_c)),
};

const int NUM_SYSTEMS = sizeof(systems)/sizeof(sys_c);

//-----------------------------------------------------------------------------
// Get object for a given system:
//-----------------------------------------------------------------------------

sys_t get_sys(str_t name) {
	for(int i = 0; i < NUM_SYSTEMS; i++) {
		if(str_eq(name, systems[i].name()))
			return &(systems[i]);
	}
	return nullptr;
}

//=============================================================================
// Class for Parameters of the Formula:
//=============================================================================

class par_c {
	public:

	//---------------------------------------------------------------------
	// Constant for number of single parameter values:
	//---------------------------------------------------------------------

	static const int DIMENSIONS = 8;

	static const int DIMEN_STARTUP  = 1;
	static const int DIMEN_LOADX    = 2;
	static const int DIMEN_RULE_APP = 3;
	static const int DIMEN_RESULT   = 4;
	static const int DIMEN_DED_LOGR = 5;
	static const int DIMEN_ITER     = 6;
	static const int DIMEN_LOAD     = 7;
	static const int DIMEN_DATA_VAL = 8;

	static const int DIMEN_FIRST = 1;
	static const int DIMEN_LAST  = 8;

	static const int DIMEN_NULL  = 0;

	//---------------------------------------------------------------------
	// Constructor:
	//---------------------------------------------------------------------

	par_c(long par_startup, long par_loadx, long par_rule_app,
		long par_result, long par_ded_logr, long par_iter,
		long par_load, long par_data_val)
	{
		Par[DIMEN_NULL]     = 0;
		Par[DIMEN_STARTUP]  = par_startup;
		Par[DIMEN_LOADX]    = par_loadx;
		Par[DIMEN_RULE_APP] = par_rule_app;
		Par[DIMEN_RESULT]   = par_result;
		Par[DIMEN_DED_LOGR] = par_ded_logr;
		Par[DIMEN_ITER]     = par_iter;
		Par[DIMEN_LOAD]     = par_load;
		Par[DIMEN_DATA_VAL] = par_data_val;
	}

	//---------------------------------------------------------------------
	// Default Constructor:
	//---------------------------------------------------------------------

	par_c()
	{
		Par[DIMEN_NULL]     = 0;
		Par[DIMEN_STARTUP]  = 0;
		Par[DIMEN_LOADX]    = 0;
		Par[DIMEN_RULE_APP] = 0;
		Par[DIMEN_RESULT]   = 0;
		Par[DIMEN_DED_LOGR] = 0;
		Par[DIMEN_ITER]     = 0;
		Par[DIMEN_LOAD]     = 0;
		Par[DIMEN_DATA_VAL] = 0;
	}

	//---------------------------------------------------------------------
	// Set parameter value:
	//---------------------------------------------------------------------

	inline void set(int dimen, long value) {

		// Check dimension:
#if DO_CHECKS
		if(dimen < DIMEN_FIRST || dimen > DIMEN_LAST) {
			std::cerr << "Invalid dimension " << dimen <<
					" in set.\n";
			exit(5);
		}
#endif

		// Set parameter value:
		Par[dimen] = value;
	}

	//---------------------------------------------------------------------
	// Get parameter value:
	//---------------------------------------------------------------------

	inline long get(int dimen) const {

		// Check dimension:
#if DO_CHECKS
		if(dimen < DIMEN_FIRST || dimen > DIMEN_LAST) {
			std::cerr << "Invalid dimension " << dimen <<
					" in get.\n";
			exit(5);
		}
#endif

		// Return parameter value:
		return Par[dimen];
	}

	//---------------------------------------------------------------------
	// Add step size to parameter value:
	//---------------------------------------------------------------------

	inline void add(int dimen, long step) {

		// Check dimension:
#if DO_CHECKS
		if(dimen < DIMEN_FIRST || dimen > DIMEN_LAST) {
			std::cerr << "Invalid dimension " << dimen <<
					" in add.\n";
			exit(5);
		}
#endif

		// Do the incrementation:
		Par[dimen] += step;
	}

	//---------------------------------------------------------------------
	// Change parameter in one dimention +1/-1:
	//---------------------------------------------------------------------

	void move(int dimen)
	{
		int inc = 1;
		if(dimen < 0) {
			dimen = -dimen;
			inc   = -1;
		}

		// Check that index is legal:
		if(dimen > DIMEN_LAST) {
				std::cerr << "Invalid dimension: " <<
						dimen << " in move\n";
				exit(5);
		}

		// Change parameter value:
		Par[dimen] += inc;
	}

	//---------------------------------------------------------------------
	// Milliseconds for startup (independent of input):
	//---------------------------------------------------------------------

	inline long startup_cost() const { return Par[DIMEN_STARTUP]; }

	//---------------------------------------------------------------------
	// Milliseconds for inserting edges in an index (per n*log(n) = 1mio):
	//---------------------------------------------------------------------

	inline long loadx_cost() const { return Par[DIMEN_LOADX]; }

	//---------------------------------------------------------------------
	// Milliseconds to apply one million rule instances:
	//---------------------------------------------------------------------

	inline long rule_app_cost() const { return Par[DIMEN_RULE_APP]; }

	//---------------------------------------------------------------------
	// Number of applied rule instances per millisecond:
	//---------------------------------------------------------------------

	inline long speed() const
	{
		long rule_app = Par[DIMEN_RULE_APP];
		return (rule_app == 0) ? 0 :
					(1000000 + rule_app/2) / rule_app;
	}

	//---------------------------------------------------------------------
	// Milliseconds to derive one million result tuples:
	//---------------------------------------------------------------------

	inline long result_cost() const { return Par[DIMEN_RESULT]; }

	//---------------------------------------------------------------------
	// Milliseconds per one million rule instances * log of result size:
	//---------------------------------------------------------------------

	inline long ded_logr_cost() const { return Par[DIMEN_DED_LOGR]; }

	//---------------------------------------------------------------------
	// Milliseconds per one million scanned tuples (iterations*edges):
	//---------------------------------------------------------------------

	inline long iter_cost() const { return Par[DIMEN_ITER]; }

	//---------------------------------------------------------------------
	// Milliseconds to load one million input tuples:
	//---------------------------------------------------------------------

	inline long load_cost() const { return Par[DIMEN_LOAD]; }

	//---------------------------------------------------------------------
	// Milliseconds to work with one million data values (nodes):
	//---------------------------------------------------------------------

	inline long data_val_cost() const { return Par[DIMEN_DATA_VAL]; }

	//---------------------------------------------------------------------
	// Product of all parameter values (to compute total number of steps):
	//---------------------------------------------------------------------

	inline bigint_t prod() const
	{
		bigint_t product = 1;
		for(int dimen = DIMEN_FIRST; dimen <= DIMEN_LAST; dimen++)
			product *= Par[dimen];
		return product;
	}

	//---------------------------------------------------------------------
	// Print parameters (long version):
	//---------------------------------------------------------------------

	void print(str_t indent)
	{
		std::cout << indent <<	"Startup:  " << Par[DIMEN_STARTUP]
							<< "\n";
		std::cout << indent <<	"LoadX:    " << Par[DIMEN_LOADX]
							<< "\n";
		std::cout << indent <<	"Rule App: " << Par[DIMEN_RULE_APP]
							<< "\n";
		std::cout << indent <<	"Result:   " << Par[DIMEN_RESULT]
							<< "\n";

		// The other parameters are optional:
		if(Par[DIMEN_DED_LOGR] > 0)
			std::cout << indent <<
					"Ded*LogR: " << Par[DIMEN_DED_LOGR]
							<< "\n";
		if(Par[DIMEN_ITER] > 0)
			std::cout << indent <<
					"Iter*Tup: " << Par[DIMEN_ITER]
							<< "\n";

		if(Par[DIMEN_LOAD] > 0)
			std::cout << indent <<
					"LinLoad:  " << Par[DIMEN_LOAD]
							<< "\n";

		if(Par[DIMEN_DATA_VAL] > 0)
			std::cout << indent <<
					"DataVal:  " << Par[DIMEN_DATA_VAL]
							<< "\n";
	}

	//---------------------------------------------------------------------
	// Print parameters (short version):
	//---------------------------------------------------------------------

	void print_short()
	{
		// Standard parameters:
		std::cout << "(" <<
			"S: " << Par[DIMEN_STARTUP]  << ", " <<
			"X: " << Par[DIMEN_LOADX]    << ", " <<
			"A: " << Par[DIMEN_RULE_APP] << ", " <<
			"R: " << Par[DIMEN_RESULT];

		// Optional Parameters:
		if(Par[DIMEN_DED_LOGR] > 0)
			std::cout << ", " << "DL: " << Par[DIMEN_DED_LOGR];
		if(Par[DIMEN_ITER] > 0)
			std::cout << ", " << "IT: " << Par[DIMEN_ITER];
		if(Par[DIMEN_LOAD] > 0)
			std::cout << ", " << "LL: " << Par[DIMEN_LOAD];
		if(Par[DIMEN_DATA_VAL] > 0)
			std::cout << ", " << "DV: " << Par[DIMEN_DATA_VAL];

		// Done:
		std::cout << ")" << std::endl;
	}

	//---------------------------------------------------------------------
	// The Formula to Estimate Runtime from Graph Data:
	//---------------------------------------------------------------------

	inline long formula(data_t data_p)
	{
		bigint_t edges = data_p->edges();
		long load_part = IDIV_ROUND(edges * Par[DIMEN_LOAD], 1000000);

		long rule_part = 0;
		bigint_t cost = data_p->cost();
		rule_part = IDIV_ROUND(cost * Par[DIMEN_RULE_APP], 1000000);
		//bigint_t cost = data_p->cost_log_tc();
		//rule_part = IDIV_ROUND(cost * Par[DIMEN_RULE_APP],
		//		static_cast<bigint_t>(1000000000));

		bigint_t tc_size = data_p->tc_size();
		long result_part = IDIV_ROUND(tc_size * Par[DIMEN_RESULT],
					 1000000);

		bigint_t iter_edges = data_p->iter_edges();
		long iter_part = IDIV_ROUND(iter_edges * Par[DIMEN_ITER],
					 1000000);

		bigint_t nodes = data_p->nodes();
		long data_part = IDIV_ROUND(nodes * Par[DIMEN_DATA_VAL],
					1000000);

		bigint_t edges_log_edges = data_p->edges_log_edges();
		long loadx_part = IDIV_ROUND(edges_log_edges * Par[DIMEN_LOADX],
					1000000);

		return Par[DIMEN_STARTUP] +
			load_part + rule_part + result_part + iter_part +
			loadx_part + data_part;
	}

	//---------------------------------------------------------------------
	// Attributes:
	//---------------------------------------------------------------------

	private:
	long Par[DIMEN_LAST+1];
};

//-----------------------------------------------------------------------------
// Pointer Type for Formula Parameters:
//-----------------------------------------------------------------------------

typedef par_c *par_t;


//=============================================================================
// Evaluation of the Formula (How good is it?):
//=============================================================================

//-----------------------------------------------------------------------------
// check_formula:
//-----------------------------------------------------------------------------

int check_formula(par_c *par_p, sys_t sys)
{
	int hits = 0;

	data_c *data_p = sys->data();
	for(int i = 0; i < NUM_GRAPHS; i++, data_p++) {
		long estimate = par_p->formula(data_p);
		long limit = estimate / 5;
		long runtime = data_p->real_time();
		if(runtime >= estimate-limit && runtime <= estimate+limit)
			hits++;
	}

	return hits;
}

//-----------------------------------------------------------------------------
// Check Formula with Verbose Output:
//-----------------------------------------------------------------------------

int check_formula_with_output(par_c *par_p, sys_t sys)
{
	int hits = 0;

	std::cout << "Detailed Formula Evaluation:\n";
	std::cout << "---------------------------\n";
	par_p->print("\t");
	std::cout << "\n";

	data_c *data_p = sys->data();
	for(int i = 0; i < NUM_GRAPHS; i++, data_p++) {
		long estimate = par_p->formula(data_p);
		long limit = estimate / 5;
		long runtime = data_p->real_time();
		std::cout << "Graph: " << data_p->graph() <<
			", Estimate: " << estimate <<
			", Runtime: " << runtime;
		if(estimate-limit > runtime)
			std::cout << ", miss (+).\n";
		else if(estimate+limit < runtime)
			std::cout << ", miss (-).\n";
		else {
			hits++;
			if(estimate > runtime)
				std::cout << ", hit (+).\n";
			else if(estimate < runtime)
				std::cout << ", hit (-).\n";
			else
				std::cout << ", hit (!).\n";
		}
	}

	std::cout << "Hits: " << hits << "\n";
	return hits;
}

//-----------------------------------------------------------------------------
// Distance measure for local optimization:
//-----------------------------------------------------------------------------

bigint_t distance(par_c *par_p, sys_t sys)
{
	bigint_t sum_dist = 0;

	data_c *data_p = sys->data();
	for(int i = 0; i < NUM_GRAPHS; i++, data_p++) {
		long estimate = par_p->formula(data_p);
		long limit = estimate / 5;
		if(estimate <= 0)
			estimate = 1;
		long runtime = data_p->real_time();
		long rel_dist;

		// Compute distance of estimate from real value in promille:
		if(estimate > runtime)
			rel_dist = ((estimate - runtime) * 1000) / estimate;
		else if(estimate == runtime)
			rel_dist = 0;
		else // runtime > estimate
			rel_dist = ((runtime - estimate) * 1000) / estimate;

		// Limit the relative distance to make sure that hits have prio:
		if(rel_dist > 100000/NUM_GRAPHS)
			rel_dist = 100000/NUM_GRAPHS;

		// Add relative distance to sum (to optimize within same hits):
		sum_dist += rel_dist;

		// Emphasize distance outside the 20% limit:
		if(rel_dist > 200) {
			//sum_dist += 1000000;
			sum_dist += 100 * (rel_dist - 200);
		}

		// Because of rounding problems, it happened that something
		// was not counted as hit, although the relative distance
		// was 200.
		// To be sure, we use the exact condition for hits (negated):
		if(runtime < estimate-limit || runtime > estimate+limit)
			sum_dist += 10000000;

	}

	return sum_dist;
}

//=============================================================================
// Auxiliary Functions to Determine the Possible Range of Parameters:
//=============================================================================


//-----------------------------------------------------------------------------
// Compute Maximal Value for the Rule Application Cost Parameter:
//-----------------------------------------------------------------------------

// If the runtime has to be explained entirely by the rule application cost
// (i.e. all other parameters are 0), what would be the value?
// This function computes the maximal such value over all graphs:

long compute_max_rule(sys_t sys)
{
	long max_rule = 0;
	long sum_rule = 0;

	data_c *data_p = sys->data();
	for(int i = 0; i < NUM_GRAPHS; i++, data_p++) {
		bigint_t cost = data_p->cost();
		bigint_t real_time = data_p->real_time();
		if(cost > 0) {
			long rule = (real_time*1000000 + (cost-1)) / cost;
			if(rule > max_rule)
				max_rule = rule;
			sum_rule += rule;
		}
	}

	// Return reasonable upper limit:
	long double_avg = (2*sum_rule) / NUM_GRAPHS;
	return MIN(double_avg, max_rule);
}

//-----------------------------------------------------------------------------
// Compute Maximal Value for the Result Cost Parameter:
//-----------------------------------------------------------------------------

long compute_max_result(sys_t sys)
{
	long max_result = 0;
	long sum_result = 0;

	data_c *data_p = sys->data();
	for(int i = 0; i < NUM_GRAPHS; i++, data_p++) {
		bigint_t tc_size = data_p->tc_size();
		bigint_t real_time = data_p->real_time();
		if(tc_size > 0) {
			long result = (real_time*1000000 + (tc_size-1)) / 
								tc_size;
			if(result > max_result)
				max_result = result;

			sum_result += result;
		}
	}

	// Return reasonable upper limit:
	long double_avg = (2*sum_result) / NUM_GRAPHS;
	return MIN(double_avg, max_result);
}

//-----------------------------------------------------------------------------
// Compute Maximal Value for the Load Parameter:
//-----------------------------------------------------------------------------

long compute_max_load(sys_t sys)
{
	long max_load = 0;
	long sum_load = 0;

	// Compute maximum value over all graphs:
	data_c *data_p = sys->data();
	for(int i = 0; i < NUM_GRAPHS; i++, data_p++) {
		bigint_t edges = data_p->edges();
		bigint_t real_time = data_p->real_time();
		if(edges > 0) {
			long load = (real_time*1000000 + (edges-1)) / edges;
			if(load > max_load)
				max_load = load;

			sum_load += load;
		}
	}

	// Limit it with a somewhat reasonable value:
	// if(max_load > 10000)
	//	max_load = 10000;

	// Return reasonable upper limit:
	long double_avg = (2*sum_load) / NUM_GRAPHS;
	return MIN(double_avg, max_load);
}


//-----------------------------------------------------------------------------
// Compute Step Size for a Parameter:
//-----------------------------------------------------------------------------

long step_size(long max_value, long num_values)
{
	// Avoid division by 0:
	if(num_values == 0)
		return 1;

	// One value is always the minimal value:
	if(num_values == 1)
		return max_value + 1;

	// Compute step size (first value is always the minimal value):
	long step = max_value / (num_values-1);

	// Ensure that at least 1:
	if(step < 1)
		step = 1;

	// Return step size:
	return step;
}

//-----------------------------------------------------------------------------
// Compute Maximal Value for the Data Values Cost Parameter:
//-----------------------------------------------------------------------------

long compute_max_dataval(sys_t sys)
{
	long max_dataval = 0;
	long sum_dataval = 0;

	data_c *data_p = sys->data();
	for(int i = 0; i < NUM_GRAPHS; i++, data_p++) {
		bigint_t nodes = data_p->nodes();
		bigint_t real_time = data_p->real_time();
		if(nodes > 0) {
			long dataval = (real_time*1000000 + (nodes-1)) / nodes;
			if(dataval > max_dataval)
				max_dataval = dataval;
			sum_dataval += dataval;
		}
	}

	// Return reasonable upper limit:
	long double_avg = (2*sum_dataval) / NUM_GRAPHS;
	return MIN(double_avg, max_dataval);
}


//-----------------------------------------------------------------------------
// Compute Maximal Value for the Iter Cost Parameter:
//-----------------------------------------------------------------------------

long compute_max_iter(sys_t sys)
{
	long max_val = 0;
	long sum_val = 0;

	data_c *data_p = sys->data();
	for(int i = 0; i < NUM_GRAPHS; i++, data_p++) {
		bigint_t cost = data_p->iter_edges();
		bigint_t real_time = data_p->real_time();
		if(cost > 0) {
			long val = IDIV_CEILING(real_time*1000000, cost);
			if(val > max_val)
				max_val = val;
			sum_val += val;
		}
	}

	// Return reasonable upper limit:
	long double_avg = (2*sum_val) / NUM_GRAPHS;
	return MIN(double_avg, max_val);
}

//-----------------------------------------------------------------------------
// Compute Maximal Value for the Index Insertion Cost Parameter (LOADX):
//-----------------------------------------------------------------------------

long compute_max_loadx(sys_t sys)
{
	long max_val = 0;
	long sum_val = 0;

	data_c *data_p = sys->data();
	for(int i = 0; i < NUM_GRAPHS; i++, data_p++) {
		bigint_t cost = data_p->edges_log_edges();
		bigint_t real_time = data_p->real_time();
		if(cost > 0) {
			long val = IDIV_CEILING(real_time*1000000, cost);
			if(val > max_val)
				max_val = val;
			sum_val += val;
		}
	}

	// Return reasonable upper limit:
	long double_avg = (2*sum_val) / NUM_GRAPHS;
	return MIN(double_avg, max_val);
}

//-----------------------------------------------------------------------------
// Determine Maximal Reasonable Values for All Parameters:
//-----------------------------------------------------------------------------

void compute_max_par(par_t max_par, sys_t sys)
{
	// Determine maximal value for load efficiency:
	long max_load = compute_max_load(sys);
	// The file contains about 8-9 bytes per edge.
	// I.e. it should be possible to read about 10mio edges per second
	// (with a fast disc and a sequentially stored/non-fragmented file).
	// However, the edges must be processed. i.e. entered in an index.
	// For some systems, we measured the load time.
	// xsb uses up to 6s load time per million edges. The average is 2.2s.
	// long max_load = 6000;

	// Limit the load time to a somewhat reasonable value:
	if(max_load > 100000)
		max_load = 100000;

	// Determine maximal value for speed:
	long max_rule = compute_max_rule(sys);

	// Reasonable maximal value for startup time:
	long max_startup = 1000;

	// Compute maximal value for time per one million result tuples:
	long max_result = compute_max_result(sys);

	// Compute maximal value for time per one million scans (iter*edges):
	long max_iter = compute_max_iter(sys);

	// Compute maximal value for time per one million index insertion cost:
	long max_loadx = compute_max_loadx(sys);

	// Compute maximal value for time per one million data values:
	long max_dataval = compute_max_dataval(sys);

	// Set the maximal values:
	max_par->set(par_c::DIMEN_STARTUP,  max_startup);
	max_par->set(par_c::DIMEN_LOAD,     max_load);
	max_par->set(par_c::DIMEN_RULE_APP, max_rule);
	max_par->set(par_c::DIMEN_RESULT,   max_result);
	max_par->set(par_c::DIMEN_ITER,     max_iter);
	max_par->set(par_c::DIMEN_LOADX,    max_loadx);
	max_par->set(par_c::DIMEN_DATA_VAL, max_dataval);
}

//=============================================================================
// Function for local optimization:
//=============================================================================

void optimize_local(par_t par, sys_t sys, int max_steps)
{
	int best_dimen = 1;
	bigint_t best_dist = distance(par, sys);
	while(best_dimen != 0 && max_steps-- > 0) {

		//std::cout << "DIST: " << best_dist <<
		//	", HITS: " << check_formula(par, sys) << ": ";
		//par->print_short();

		// Stop loop if we cannot improve the distance:
		best_dimen = 0;

		// Try all dimensions, move in direction with smallest distance:
		for(int dimen = 1; dimen < par_c::DIMENSIONS; dimen++) {
			par->move(dimen);
			bigint_t new_dist = distance(par, sys);
			if(new_dist < best_dist) {
				best_dist = new_dist;
				best_dimen = dimen;
			}
			// Move back:
			par->move(-dimen);

			// Now do the same for the negative direction:
			par->move(-dimen);
			new_dist = distance(par, sys);
			if(new_dist < best_dist) {
				best_dist = new_dist;
				best_dimen = -dimen;
			}
			// Move back:
			par->move(dimen);
		}
		//std::cout << "\tBest Dimen: " << best_dimen << "\n";

		// Now do the best move:
		if(best_dimen != 0)
			par->move(best_dimen);

	// End of while loop over optimization steps:
	}
}

//=============================================================================
// Function for global optimization with relative optimization at each point:
//=============================================================================

int optimize(par_t best_par, sys_t sys,
		par_t min_par, par_t max_par, par_t inc_par)
{
	int best_hits = 0;

	// Compute total number of steps:
	bigint_t total_steps = 1;
	for(int d = par_c::DIMEN_FIRST; d <= par_c::DIMEN_LAST; d++) {
		long diff = max_par->get(d) - min_par->get(d);
		total_steps *= IDIV_CEILING((diff+1), inc_par->get(d));
	}

	// For output of progress:
	bigint_t one_percent = IDIV_CEILING(total_steps, 100);
	bigint_t countdown = 1;
	int pct = 0;

	// Optimization:
	par_c par;
	while(1) {

		// Print control output:
		if(--countdown <= 0) {
			std::cout << pct << "%\n";
			countdown = one_percent;
			pct++;
		}

		par_c opt_par = par;
		optimize_local(&opt_par, sys, 1000);
		//par.print_short();

		int hits = check_formula(&opt_par, sys);
		if(hits > best_hits) {
			*best_par = opt_par;
			best_hits = hits;
		}

		// Determine next parameter value:
		int dimen = par_c::DIMEN_LAST;
		while(dimen >= par_c::DIMEN_FIRST) {
			par.add(dimen, inc_par->get(dimen));
			if(par.get(dimen) <= max_par->get(dimen))
				break;
			par.set(dimen, min_par->get(dimen));
			dimen--;
			
		}
		if(dimen < par_c::DIMEN_FIRST)
			break;
	}

	return best_hits;
}

//=============================================================================
// Class for Optimization Job (Can be executed in a thread):
//=============================================================================

class job_c {

	public:

	//---------------------------------------------------------------------
	// Default Constructor (needed to declare array of jobs):
	//---------------------------------------------------------------------

	job_c() :
		MinPar(), MaxPar(), IncPar(), BestPar()
	{
		Sys   = nullptr;
		JobNo = 0;
		BestHits = -1;
	}

	//---------------------------------------------------------------------
	// This does the real initialization:
	//---------------------------------------------------------------------

	void init(sys_t job_sys, const par_c& job_min_par,
			const par_c& job_max_par, const par_c& job_inc_par,
			int split_dimen, int split_jobs, int job_no)
	{
		// Store given data:
		Sys      = job_sys;
		MinPar   = job_min_par;
		MaxPar   = job_max_par;
		IncPar   = job_inc_par;
		JobNo    = job_no;
		BestHits = -1;

		// Check splitting parameters:
#if DO_CHECK
		if(split_jobs < 1) {
			std::cerr << "Invalid number of jobs: " <<
					split_jobs << "\n";
			exit(6);
		}
		if(job_no < 0 || job_no >= split_jobs) {
			std::cerr << "Invalid job no: " << job_no << "\n";
			exit(6);
		}
#endif

		// Stop here if only one job (serial execution):
		if(split_dimen == par_c::DIMEN_NULL || split_jobs <= 1)
			return;

		// Split parameter range in given dimension (for parallel exec):
		long min = MinPar.get(split_dimen);
		long max = MaxPar.get(split_dimen);
		long inc = IncPar.get(split_dimen);
		long diff = max - min;
		long steps = IDIV_CEILING((diff+1), inc);
		if(steps < split_jobs) {
			std::cerr << split_jobs << " jobs requested, " <<
				"but only " << steps << " steps " <<
				"in dimension " << split_dimen << ".\n";
				exit(7);
		}
		long steps_per_job = steps / split_jobs;
		long remaining_steps = steps % split_jobs;
		long min_in_this_job = min + job_no * steps_per_job * inc;
		long max_in_this_job = min + (job_no+1) * steps_per_job * inc
			- 1;
		// We distribute the remaining_steps over the first jobs
		// (one additional step per job):
		if(job_no < remaining_steps) {
			min_in_this_job += (job_no * inc);
			// The first job has job_no==0, thus the first minimum
			// will not be increased.
			max_in_this_job += ((job_no+1) * inc);
			// This does not change the property that the intervals
			// are consecutive, because the next minimum will be
			// increased by the next job_no, which is job_no+1.
		}
		else {
			min_in_this_job += (remaining_steps * inc);
			max_in_this_job += (remaining_steps * inc);
		}

		// Now set minimum and maximum for the split_dimenion:
		MinPar.set(split_dimen, min_in_this_job);
		MaxPar.set(split_dimen, max_in_this_job);

		// Make sure that the interval of the last job ends at max:
		// (This is probably unnecessary)
		if(job_no == split_jobs - 1)
			MaxPar.set(split_dimen, max);

		// Test Output:
		// std::cout << "Job " << JobNo << ":\n";
		// std::cout << "\tMin:   "; MinPar.print_short();
		// std::cout << "\tMax:   "; MaxPar.print_short();
		// std::cout << "\tInc:   "; IncPar.print_short();
		// std::cout << "\tSteps: " << TotalSteps << std::endl;
	}

	//---------------------------------------------------------------------
	// Do Optimization (without relative optimization):
	//---------------------------------------------------------------------

	int optimize_simple()
	{
		int max_hits = 0;

		// For output of progress:
		bigint_t output_steps = IDIV_CEILING(total_steps() * OUT_PCT,
							100);
		bigint_t countdown = 1;
		int pct = 0;

		// Optimization:
		par_c par = MinPar;
		while(1) {
			// Print control output:
			if(--countdown <= 0) {
#if NUM_JOBS > 1
				std::cout << "[Job " << JobNo << ": " <<
					pct << "%, Best: " << max_hits <<
					"]: ";
#else
				std::cout << pct << "%, Best:" << max_hits;
#endif
				BestPar.print_short();
				std::cout << std::flush;
				countdown = output_steps;
				pct += OUT_PCT;
			}

			//par.print_short();

			int hits = check_formula(&par, Sys);
			if(hits > max_hits) {
				BestPar = par;
				max_hits = hits;
			}

			// Determine next parameter value:
			int dimen = par_c::DIMEN_LAST;
			while(dimen >= par_c::DIMEN_FIRST) {
				par.add(dimen, IncPar.get(dimen));
				if(par.get(dimen) <= MaxPar.get(dimen))
					break;
				par.set(dimen, MinPar.get(dimen));
				dimen--;
				
			}
			if(dimen < par_c::DIMEN_FIRST)
				break;
		}

		BestHits = max_hits;
		return max_hits;
	}

	//---------------------------------------------------------------------
	// Attribute Accessor Functions:
	//---------------------------------------------------------------------

	inline const par_c *best_par() const { return &BestPar; }

	inline int best_hits() const { return BestHits; }

	//---------------------------------------------------------------------
	// Compute total number of steps:
	//---------------------------------------------------------------------

	bigint_t total_steps() const
	{
		bigint_t total = 1;
		for(int d = par_c::DIMEN_FIRST; d <= par_c::DIMEN_LAST; d++) {
			long diff = MaxPar.get(d) - MinPar.get(d);
			long steps = IDIV_CEILING((diff+1), IncPar.get(d));
			total *= steps;
		}
		return total;
	}

	//---------------------------------------------------------------------
	// Attributes:
	//---------------------------------------------------------------------

	private:

	sys_t Sys;
	par_c MinPar;
	par_c MaxPar;
	par_c IncPar;
	par_c BestPar;
	int   BestHits;
	int   JobNo;
};

//-----------------------------------------------------------------------------
// Pointer Type for Job:
//-----------------------------------------------------------------------------

typedef job_c *job_t;

//=============================================================================
// Functions For Setting Up and Executing Optimization Jobs:
//=============================================================================

//-----------------------------------------------------------------------------
// Function to be Executed by a Separate Thread:
//-----------------------------------------------------------------------------

void run_job(job_t job) {
	job->optimize_simple();
}

//-----------------------------------------------------------------------------
// Parallel Execution of Simple Optimization:
//-----------------------------------------------------------------------------

#if NUM_JOBS > 1

int optimize_simple(sys_t sys, const par_c& min_par, const par_c&max_par,
			const par_c&inc_par, par_t best_par)
{
	// Optimization Jobs:
	job_c jobs[NUM_JOBS];
	for(int i = 0; i < NUM_JOBS; i++) {
		jobs[i].init(sys, min_par, max_par, inc_par,
				par_c::DIMEN_STARTUP, NUM_JOBS, i);
	}

	// Compute total number of parameter tuples:
	bigint_t total_steps = 0;
	for(int i = 0; i < NUM_JOBS; i++)
		total_steps += jobs[i].total_steps();
	std::cout << "Number of parameter tuples: " << total_steps << "\n";
	std::cout << std::endl;

	// Start threads for parallel execution of all jobs:
	std::cout << "Starting " << NUM_JOBS << " threads ..." << std::endl;
	std::thread threads[NUM_JOBS];
	for(int i = 0; i < NUM_JOBS; i++) {
		// Move-assign threads:
		threads[i] = std::thread(run_job, &(jobs[i]));
	}

	// Join threads:
	for(int i = 0; i < NUM_JOBS; i++)
		threads[i].join();

	// Output from main thread:
	std::cout << "Threads have finished.\n";

	// Find the best solution among the job results:
	*best_par = *(jobs[0].best_par());
	int best_hits = jobs[0].best_hits();
	for(int i = 1; i < NUM_JOBS; i++) {
		int job_best_hits = jobs[i].best_hits();
		if(jobs[i].best_hits() > best_hits) {
			best_hits = job_best_hits;
			*best_par = *(jobs[i].best_par());
		}
	}

	// Return hits of best parameter:
	return best_hits;
}

#endif

//-----------------------------------------------------------------------------
// Serial Version of Simple Optimization (without threads):
//-----------------------------------------------------------------------------

#if NUM_JOBS <= 1

int optimize_simple(sys_t sys, const par_c& min_par, const par_c&max_par,
			const par_c&inc_par, par_t best_par)
{
	// Optimization Job (parameter range is not splitted, only one job):
	job_c job;
	job.init(sys, min_par, max_par, inc_par, par_c::DIMEN_NULL, 1, 0);

	// Print total number of parameter tuples:
	std::cout << "Number of parameter tuples: " << job.total_steps() <<
			"\n";
	std::cout << std::endl;

	// Do optimization:
	//best_hits = optimize_simple(&best_par, sys,
	//		&min_par, &max_par, &inc_par);
	job.optimize_simple();

	// Find the best solution among the job results:
	*best_par = *(job.best_par());
	int best_hits = job.best_hits();

	// Return hits of best parameter:
	return best_hits;
}

#endif

//=============================================================================
// Function to Check the Best Known Parameter Values:
//=============================================================================


void best_known()
{
	std::cout << "\n";

	// XSB:
	std::cout << "*** XSB ***\n\n";
	  sys_t xsb = get_sys("xsb");
	//           |Startup|   Load|   Rule| Result|   Iter|  LoadX|DataVal|
	  par_c xsb_par(   92,    999,     97,    300,      0,      0,      0);
	  check_formula_with_output(&xsb_par, xsb);
	std::cout << "\n\n";

	//BAM:
	std::cout << "*** BAM ***\n\n";
	  sys_t bam = get_sys("bam");
	//           |Startup|   Load|   Rule| Result|   Iter|  LoadX|DataVal|
	  par_c bam_par(   53,    320,     45,     36,      0,      0,   2000);
	  check_formula_with_output(&bam_par, bam);
	std::cout << "\n\n";

	//BAM New:
	std::cout << "*** BAM NEW ***\n\n";
	//           |Startup|   Load|   Rule| Result|   Iter|  LoadX|DataVal|
	  par_c bam_par2(   0,      0,     50,     48,      0,      0,      0);
	  check_formula_with_output(&bam_par2, bam);
	std::cout << "\n\n";

	//YAP:
	std::cout << "*** YAP ***\n\n";
	  sys_t yap = get_sys("yap");
	//           |Startup|   Load|   Rule| Result|   Iter|  LoadX|DataVal|
	  par_c yap_par(   33,    670,    242,    300,      0,      0,      0);
	  check_formula_with_output(&yap_par, yap);
	std::cout << "\n\n";

	//SQLite:
	std::cout << "*** SQLite ***\n\n";
	  sys_t sqlite = get_sys("sqlite");
	//           |Startup|   Load|   Rule| Result|   Iter|  LoadX|DataVal|
	  par_c sql_par(   55,   6000,    692,   1000,      0,      0,      0);
	  check_formula_with_output(&sql_par, sqlite);
	std::cout << "\n\n";

	//Jena:
	std::cout << "*** Jena ***\n\n";
	  sys_t jena = get_sys("jena");
	//           |Startup|   Load|   Rule| Result|   Iter|  LoadX|DataVal|
	  par_c jen_par( 1650,      0,    150,   1500,      0,   1710,      0);
	  check_formula_with_output(&jen_par, jena);
	std::cout << "\n\n";

	// Print number of graphs:
	std::cout << "Number of Graphs: " << NUM_GRAPHS << "\n\n";
}

//=============================================================================
// The main Function (Does the Optimization):
//=============================================================================


int main(int argc, str_t argv[])
{
	int best_hits = 0;
	par_c best_par;

	// List evaluation of best known parameter settings:
	if(argc == 2 and str_eq(argv[1],"best")) {
		best_known();
		return 0;
	}

	// The program should be called with the system name as cmdline arg:
	str_t sys_name = "xsb";
	if(argc == 2)
		sys_name = argv[1];
	else if(argc > 2) {
		std::cerr << "Usage: opt <SystemName>, e.g. \"opt xsb\".\n";
		exit(1);
	}

	// Get data for system:
	sys_t sys = get_sys(sys_name);
	if(sys == nullptr) {
		std::cerr << "Unkown system: " << sys_name << ".\n";
		exit(2);
	}

	// Print welcome:
	std::cout << "Welcome to the TC runtime prediction optimizer.\n";
	std::cout << "\n";
	std::cout << "Number of defined systems: " << NUM_SYSTEMS << "\n";
	std::cout << "Doing optimization for:    " << sys->name() << "\n";
	std::cout << "Number of graphs:          " << NUM_GRAPHS << "\n";
	std::cout << "Number of Threads:         " << NUM_JOBS << "\n";

	// Define number of steps for search:
	//         |Startup|   Load|   Rule| Result|DataVal|
	//par_c steps(     10,     50,    100,     10,     50);
	// This runs 5040.07user 0.08system 1:24:00 on bach
	// The total number of steps is 25mio, i.e. 200s/mio steps
	// Another test used 240s/mio steps

	// Small test search:
	//         |Startup|   Load|   Rule| Result|DataVal|
	//par_c steps(     10,     10,    100,     10,     10);
	//std::cout << "Number of parameter tuples:     " << steps.prod()<<"\n";
	//std::cout << "\n";
	// 240.99user 0.00system 4:00.97elapsed
	// Finds for Jena 19/35
	// Startup:  1346
	// Load:     0
	// Rule App: 224
	// Speed:    4464
	// Result:   30
	// DataVal:  48830

	// Small test search:
	//           |Startup|   Load|   Rule| Result|   Iter|  LoadX|DataVal|
	//par_c steps(     10,     50,    100,     50,      1,      50,     1);
	// This uses about 30s of CPU time

	// Larger Search:
	//           |Startup|   Load|   Rule| Result|   Iter|  LoadX|DataVal|
	//par_c steps(    50,    100,   1000,    1000,    100,      1,      );
	// This took about 32h CPU-time (8h with four threads).

	// Current search:
	//           |Startup|   Load|   Rule| Result|   Iter|  LoadX|DataVal|
	//par_c steps(    600,      1,   1000,    500,      1,    500,     1);
	  par_c steps(     60,      1,    101,   1001,      1,   2000,     1);

	//par_c steps(     60,    100,    100,    100,     10,    100,     5);
	// This needed 50h CPU-time on work128, 2h with 30 threads

	// Compute maximal reasonable parameter values:
	//par_c max_par;
	//compute_max_par(&max_par, sys);

	// Manual setting of reasonable upper limits:
	//           |Startup|   Load|   Rule| Result|   Iter|  LoadX|DataVal|
	//par_c max_par( 3000,   5000,   1000,   3000,   3000,   5000,   3000);
	//par_c max_par( 3000,      0,   3000,   3000,      0,   5000,      0);
	//par_c max_par(  300,      0,   1000,   1000,      0,   8000,      0);
	  par_c max_par(   60,      0,    100,   1000,      0,   2000,      0);

	// Print search range:
	std::cout << "Maximal Parameter Values:\n";
	std::cout << "------------------------\n";
	max_par.print("\t");
	std::cout << "\n";

	// Currently the minimal parameters are 0:
	par_c min_par;

	// Print number of steps:
	std::cout << "Requested Number of Steps " <<
		"(Might Change Due to Rounded Step Size):\n";
	std::cout << "--------------------------" <<
		"---------------------------------------\n";
	steps.print("\t");
	std::cout << "\n";

	// Compute step size:
	par_c inc_par;
	for(int dimen = par_c::DIMEN_FIRST; dimen <= par_c::DIMEN_LAST;
			dimen++) {
		inc_par.set(dimen,
			step_size(max_par.get(dimen), steps.get(dimen)));
	}

	// Print step size:
	std::cout << "Step Size:\n";
	std::cout << "---------\n";
	inc_par.print("\t");

	// Local Optimization form good starting points:
	//          Start   Load  RuleApp Result
	//par_c opt(   90,  1000,      97,   300); // xsb
	//par_c opt(   70,  1130,      50,     0); // bam
	//par_c opt(   30,   670,     253,   300); // yap
	//par_c opt(   40,  6000,     694,  1000); // sqlite
	//par_c opt(  970, 27300,     168,  1558); // jena
	//optimize_local(&opt, sys, 1000);
	//best_par = opt;

	// Do optimization:
	best_hits = optimize_simple(sys, min_par, max_par, inc_par, &best_par);

	// Show runtimes for best formula:
	std::cout << "\n";
	std::cout << "\n";
	best_hits = check_formula_with_output(&best_par, sys);
	std::cout << "\n";
	std::cout << "\n";
	std::cout << "Best Parameter Values (" << best_hits << "/" <<
					NUM_GRAPHS << "):\n";
	best_par.print("\t");
	std::cout << "\n";
}

